/* arch/arm/mach-s3c2410/tomtomgo-type.c
 *
 * Definitions for different TomTom GO types.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 * Authors:
 * Jeroen Taverne <jeroen.taverne@tomtom.com>
 * Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <asm/arch/irqs.h>
#include <barcelona/gotype.h>

#ifdef __KERNEL__
#include <asm/arch/map.h>	/* for S3C2410_VA_GPIO */
#include <asm/system.h>		/* for local_irq_save/restore */
#include <linux/module.h>	/* for EXPORT_SYMBOL */
#define GP_BASE		S3C2410_VA_GPIO
#define DBG(fmt, arg...) printk(KERN_DEBUG "%s: " fmt, __func__ ,##arg)
#else /* __KERNEL__ */
#define GP_BASE		0x56000000
#define DBG(fmt, arg...)
#define local_irq_save(x)
#define local_irq_restore(x)
#define EXPORT_SYMBOL(x)
#endif  /* __KERNEL__ */

#define rGSTATUS1	(*(volatile unsigned *)(GP_BASE + 0xb0))
#define rGPCDAT		(*(volatile unsigned *)(GP_BASE + 0x24))

#define rGP_CON(p)	(*(volatile unsigned*)(GP_BASE + ((p) << 4) + 0x0))
#define rGP_DAT(p)	(*(volatile unsigned*)(GP_BASE + ((p) << 4) + 0x4))
#define rGP_UP(p)	(*(volatile unsigned*)(GP_BASE + ((p) << 4) + 0x8))

/* Hardware type lookup table */
static const struct gotype gotypes[] =
{
#include "tomtomgo-ioclassic.h"
#include "tomtomgo-iohdproto.h"
#include "tomtomgo-iom100.h"
#include "tomtomgo-iom300.h"
#include "tomtomgo-iom500.h"
#include "tomtomgo-iosmdk2440.h"
};

#define NUM_GOTYPES (sizeof gotypes / sizeof gotypes[0])

// Use an array so the defines can change without problems
const unsigned irqList[] = 
{
	IRQ_EINT0,
	IRQ_EINT1,
	IRQ_EINT2,
	IRQ_EINT3,
	IRQ_EINT4,
	IRQ_EINT5,
	IRQ_EINT6,
	IRQ_EINT7,
	IRQ_EINT8,
	IRQ_EINT9,
	IRQ_EINT10,
	IRQ_EINT11,
	IRQ_EINT12,
	IRQ_EINT13,
	IRQ_EINT14,
	IRQ_EINT15,
	IRQ_EINT16,
	IRQ_EINT17,
	IRQ_EINT18,
	IRQ_EINT19,
	IRQ_EINT20,
	IRQ_EINT21,
	IRQ_EINT22,
	IRQ_EINT23
};

/* Currently detected GO type */
struct gotype gotype_current = 
{
#include "tomtomgo-iodefault.h"
};
EXPORT_SYMBOL(gotype_current);

int IO_Init(void)
{
	unsigned id;
	unsigned i;
	
	DBG("Starting detection\n");
#ifdef CONFIG_SMDK2440_BOARD
	id = GOTYPE_SMDK2440;
#else /* CONFIG_SMDK2440_BOARD */
	id = (rGPCDAT >> 2) & 3;
	if (rGSTATUS1 != 0x32440001)
		id |= 4;
#endif /* CONFIG_SMDK2440_BOARD */
	
	/* Lookup ID in table */
	for (i = 0; i < NUM_GOTYPES; ++i) {
		if (gotypes[i].id == id) {
			DBG("Detected type %d, name %s\n", gotypes[i].id, gotypes[i].name);
			gotype_current = gotypes[i];
			/* Power down all devices on init, for unknown reasons */
			IO_PowerOn();
			return 0;
		}
	}
	
	/* ID not found, remember it anyway */
	gotype_current.id = id;
	return -1;
}
EXPORT_SYMBOL(IO_Init);

void IO_PowerOn(void)
{
	// This should not be needed because the drivers will manage this, check if the system works without it!!!
	
	/* SD Card Interface */
	IO_Deactivate(SD_PWR_ON);
	IO_Deactivate(SDCLK);
	IO_Deactivate(SDCMD);
	IO_Deactivate(SDDATA0);
	IO_Deactivate(SDDATA1);
	IO_Deactivate(SDDATA2);
	IO_Deactivate(SDDATA3);
	IO_Deactivate(WP_SD);
	IO_SetInput(CD_SD);

	/* GPS interface */
	IO_Activate(GPS_RESET);
	IO_Deactivate(GPS_ON);
	IO_Deactivate(GPS_1PPS);
	IO_Deactivate(GPS_REPRO);

	/* Hardware detection */
	IO_SetInput(HW_TYPE_0);
	IO_SetInput(HW_TYPE_1);

	/* Harddisk interface */
	IO_Activate(HD_LED);
	IO_Deactivate(HDD_PWR_ON);
	IO_Activate(AT_RST);
	IO_Deactivate(HDD_BUF_EN);
	IO_Deactivate(HDD_IRQ);
	IO_Deactivate(HDD_DRQ);
	IO_Deactivate(HDD_DACK);

	/* Battery */
	IO_SetInput(CHARGEFAULT);
	IO_Deactivate(CHARGE_OUT);
	IO_Deactivate(AIN4_PWR);

	/* Docking */
	IO_Deactivate(DOCK_GP1);
	IO_Deactivate(DOCK_GP2);
	IO_Activate(DOCK_PWREN);
	IO_SetInput(DOCKSNS0);
	IO_SetInput(DOCKSNS1);

	/* Remote */
	IO_SetInput(FSK_DATA_FFS);
	IO_SetInput(FSK_IRQ);
	IO_SetInput(SPICLK);
	IO_SetInput(SPIMSI);
	IO_SetInput(SPIMSO);
	IO_SetInput(FSK_EN);
	IO_SetInput(FSK_DCLK_FFE);
	IO_SetInput(FSK_CLKOUT);

	/* Accelerometer */
	IO_Deactivate(ACC_PWR_ON);

	/* UARTs */
	IO_SetFunction(CTS_DOCK);
	IO_SetFunction(RTS_DOCK);
	IO_SetFunction(TXD_DOCK);
	IO_SetFunction(RXD_DOCK);
	IO_Deactivate(TXD_GPS);
	IO_Deactivate(RXD_GPS);
	IO_SetFunction(TXD_BT);
	IO_SetFunction(RXD_BT);
	IO_SetFunction(RTS_BT);
	IO_SetFunction(CTS_BT);

	/* USB */
	IO_SetInput(USB_HOST_DETECT);
	IO_Deactivate(USB_PWR_EN);
	IO_Deactivate(USB_PULL_EN);
	IO_Deactivate(SUSPEND);
	IO_Deactivate(USB_CLKOUT);
	IO_Activate(USB_RST);

	/* Power management */
	IO_SetInput(ON_OFF);
	IO_SetInput(IGNITION);
	IO_SetInput(ACPWR);

	/* Touchscreen */
	IO_Deactivate(XMON);
	IO_Deactivate(XPON);
	IO_Deactivate(YMON);
	IO_Deactivate(YPON);
}
EXPORT_SYMBOL(IO_PowerOn);

void IO_PowerOff(void)
{
	// This should be the same as in the bootloader!!!
	
	/* SD Card Interface */
	IO_Deactivate(SD_PWR_ON);
	IO_Deactivate(SDCLK);
	IO_Deactivate(SDCMD);
	IO_Deactivate(SDDATA0);
	IO_Deactivate(SDDATA1);
	IO_Deactivate(SDDATA2);
	IO_Deactivate(SDDATA3);
	IO_Deactivate(WP_SD);
	IO_SetInput(CD_SD);

	/* GPS interface */
	IO_Activate(GPS_RESET);
	IO_Deactivate(GPS_ON);
	IO_Deactivate(GPS_1PPS);
	IO_Deactivate(GPS_REPRO);

	/* Hardware detection */
	IO_SetInput(HW_TYPE_0);
	IO_SetInput(HW_TYPE_1);

	/* Harddisk interface */
	IO_Activate(HD_LED);
	IO_Deactivate(HDD_PWR_ON);
	IO_Activate(AT_RST);
	IO_Deactivate(HDD_BUF_EN);
	IO_Deactivate(HDD_IRQ);
	IO_Deactivate(HDD_DRQ);
	IO_Deactivate(HDD_DACK);

	/* Battery */
	IO_SetInput(CHARGEFAULT);
	IO_Deactivate(CHARGE_OUT);
	IO_Deactivate(AIN4_PWR);

	/* Docking */
	IO_Deactivate(DOCK_GP1);
	IO_Deactivate(DOCK_GP2);
	IO_Deactivate(DOCK_PWREN);
	IO_SetInput(DOCKSNS0);
	IO_SetInput(DOCKSNS1);

	/* Remote */
	IO_SetInput(FSK_DATA_FFS);
	IO_SetInput(FSK_IRQ);
	IO_SetInput(SPICLK);
	IO_SetInput(SPIMSI);
	IO_SetInput(SPIMSO);
	IO_SetInput(FSK_EN);
	IO_SetInput(FSK_DCLK_FFE);
	IO_SetInput(FSK_CLKOUT);

	/* Accelerometer */
	IO_Deactivate(ACC_PWR_ON);

	/* UARTs */
	IO_Deactivate(CTS_DOCK);
	IO_Deactivate(RTS_DOCK);
	IO_Deactivate(TXD_DOCK);
	IO_Deactivate(RXD_DOCK);
	IO_Deactivate(TXD_GPS);
	IO_Deactivate(RXD_GPS);
	IO_SetFunction(TXD_BT);
	IO_SetFunction(RXD_BT);
	IO_SetFunction(RTS_BT);
	IO_SetFunction(CTS_BT);

	/* USB */
	IO_SetInput(USB_HOST_DETECT);
	IO_Deactivate(USB_PWR_EN);
	IO_Deactivate(USB_PULL_EN);
	IO_Deactivate(SUSPEND);
	IO_Deactivate(USB_CLKOUT);
	IO_Activate(USB_RST);

	/* Power management */
	IO_SetInput(ON_OFF);
	IO_SetInput(IGNITION);
	IO_SetInput(ACPWR);

	// Touchscreen
	if (IO_HaveTsFets())
	{
		// Disable all FETs
		IO_Deactivate(XMON);           
		IO_Deactivate(XPON);           
		IO_Deactivate(YMON);           
		IO_Deactivate(YPON);          
	}
	else
	{
		// Force all pins low by removing the flags using PIN_MASK
		IOP_Activate(gotype_current.pins.XMON & PIN_MASK);           
		IOP_Activate(gotype_current.pins.XPON & PIN_MASK);           
		IOP_Activate(gotype_current.pins.YMON & PIN_MASK);           
		IOP_Activate(gotype_current.pins.YPON & PIN_MASK);          
	}
}
EXPORT_SYMBOL(IO_PowerOff);

inline void IOP_SetBit(gopin_t pin, unsigned value)
{
	unsigned pinNr = pin & PIN_MASK;
	unsigned portNr;
	unsigned long flags;
	
	// Don't do anything if pin isn't used
	if (pin & PIN_NOT_USED) return;
	
	portNr = pinNr >> PIN_STEP_SIZE_BITS;
	pinNr &= PIN_STEP_SIZE_MASK;
	
	local_irq_save(flags);
	
	if (portNr == 0)
	{
		// Set config to output
		rGP_CON(portNr) &= ~(1 << pinNr);
	}
	else
	{
		// Set config to output
		rGP_CON(portNr) = (rGP_CON(portNr) & ~(3 << (pinNr*2))) | (1 << (pinNr*2));
		// Disable pullup
		rGP_UP(portNr) |= (1 << pinNr);
	}
	
	// Set bit
	if (value) rGP_DAT(portNr) |= (1 << pinNr); else rGP_DAT(portNr) &= ~(1 << pinNr);
	
	local_irq_restore(flags);
}

void IOP_Activate(gopin_t pin)
{
	if (pin & PIN_INVERTED)
	{
		IOP_SetBit(pin, 0);
	}
	else
	{
		IOP_SetBit(pin, 1);
	}
}
EXPORT_SYMBOL(IOP_Activate);

void IOP_Deactivate(gopin_t pin)
{
	if ((pin & PIN_FLOATING) == PIN_FLOATING)
	{
		IOP_SetInput(pin);
	}
	else if (pin & PIN_INVERTED)
	{
		IOP_SetBit(pin, 1);
	}
	else
	{
		IOP_SetBit(pin, 0);
	}
}
EXPORT_SYMBOL(IOP_Deactivate);

void IOP_SetFunction(gopin_t pin)
{
	unsigned pinNr = pin & PIN_MASK;
	unsigned portNr;
	unsigned long flags;
	
	// Don't do anything if pin isn't used
	if (pin & PIN_NOT_USED) return;
	
	portNr = pinNr >> PIN_STEP_SIZE_BITS;
	pinNr &= PIN_STEP_SIZE_MASK;
	
	local_irq_save(flags);
	
	if (portNr)
	{
		// Set config to output
		rGP_CON(portNr) = (rGP_CON(portNr) & ~(3 << (pinNr*2))) | (2 << (pinNr*2));
		// Disable pullup
		rGP_UP(portNr) |= (1 << pinNr);
	}
	
	local_irq_restore(flags);
}
EXPORT_SYMBOL(IOP_SetFunction);

unsigned IOP_SetInterrupt(gopin_t pin)
{
	unsigned pinNr = pin & PIN_MASK;
	unsigned portNr;
	unsigned long flags;
	unsigned intNr;
	
	// Don't do anything if pin isn't used
	if (pin & PIN_NOT_USED) return -1;
	
	portNr = pinNr >> PIN_STEP_SIZE_BITS;
	pinNr &= PIN_STEP_SIZE_MASK;

	switch (portNr)
	{
	case PORT_GPF:
		intNr = irqList[pinNr];
		break;
	case PORT_GPG:
		intNr = irqList[pinNr + 8];
		break;
	default:
		return -1;
	}

#if 1
	local_irq_save(flags);
	
	if (portNr)
	{
		// Set config to interrupt
		rGP_CON(portNr) = (rGP_CON(portNr) & ~(3 << (pinNr*2))) | (2 << (pinNr*2));
		// Enable pullup for interrupt pin
		if (pin & PIN_NO_PULLUP) rGP_UP(portNr) |= (1 << pinNr); else rGP_UP(portNr) &= ~(1 << pinNr);
	}
	
	local_irq_restore(flags);
#endif

	return intNr;
}
EXPORT_SYMBOL(IOP_SetInterrupt);

unsigned IOP_SetInterruptOnActivation(gopin_t pin)
{
	unsigned intNr;
	
	intNr = IOP_SetInterrupt(pin);
	if (pin & PIN_INVERTED) set_irq_type(intNr, IRQT_FALLING); else set_irq_type(intNr, IRQT_RISING);
	return intNr;
}
EXPORT_SYMBOL(IOP_SetInterruptOnActivation);

unsigned IOP_SetInterruptOnDeactivation(gopin_t pin)
{
	unsigned intNr;

	intNr = IOP_SetInterrupt(pin);
	if (pin & PIN_INVERTED) set_irq_type(intNr, IRQT_RISING); else set_irq_type(intNr, IRQT_FALLING);
	return intNr;
}
EXPORT_SYMBOL(IOP_SetInterruptOnDeactivation);

unsigned IOP_SetInterruptOnActivated(gopin_t pin)
{
	unsigned intNr;
	
	intNr = IOP_SetInterrupt(pin);
	if (pin & PIN_INVERTED) set_irq_type(intNr, IRQT_LOW); else set_irq_type(intNr, IRQT_HIGH);
	return intNr;
}
EXPORT_SYMBOL(IOP_SetInterruptOnActivated);

unsigned IOP_SetInterruptOnDeactivated(gopin_t pin)
{
	unsigned intNr;

	intNr = IOP_SetInterrupt(pin);
	if (pin & PIN_INVERTED) set_irq_type(intNr, IRQT_HIGH); else set_irq_type(intNr, IRQT_LOW);
	return intNr;
}
EXPORT_SYMBOL(IOP_SetInterruptOnDeactivated);

unsigned IOP_SetInterruptOnToggle(gopin_t pin)
{
	unsigned intNr;

	intNr = IOP_SetInterrupt(pin);
	set_irq_type(intNr, IRQT_BOTHEDGE);
	return intNr;
}
EXPORT_SYMBOL(IOP_SetInterruptOnToggle);

void IOP_SetInput(gopin_t pin)
{
	unsigned pinNr = pin & PIN_MASK;
	unsigned portNr;
	unsigned long flags;
	
	// Don't do anything if pin isn't used
	if (pin & PIN_NOT_USED) return;
	
	portNr = pinNr >> PIN_STEP_SIZE_BITS;
	pinNr &= PIN_STEP_SIZE_MASK;
	
	local_irq_save(flags);
	
	if (portNr)
	{
		// Set config to input
		rGP_CON(portNr) = (rGP_CON(portNr) & ~(3 << (pinNr*2))) | (0 << (pinNr*2));
		// Enable pullup for input pin
		if (pin & PIN_NO_PULLUP) rGP_UP(portNr) |= (1 << pinNr); else rGP_UP(portNr) &= ~(1 << pinNr);
	}
	
	local_irq_restore(flags);
}
EXPORT_SYMBOL(IOP_SetInput);

int IOP_GetInput(gopin_t pin)
{
	unsigned pinNr = pin & PIN_MASK;
	unsigned portNr;
	
	// Don't do anything if pin isn't used
	if (pin & PIN_NOT_USED) return 0;
	
	portNr = pinNr >> PIN_STEP_SIZE_BITS;
	pinNr &= PIN_STEP_SIZE_MASK;
	
	if (rGP_DAT(portNr) & (1 << pinNr))
	{
		if (pin & PIN_INVERTED) return 0; else return 1;
	}
	else
	{
		if (pin & PIN_INVERTED) return 1; else return 0;
	}
}
EXPORT_SYMBOL(IOP_GetInput);

/* EOF */
