/* linux/arch/arm/mach-s3c2410/wake.c
 *
 * Copyright (c) 2005 tomTom BV
 * Author: Thomas Kleffel <tk@maintech.de>
 *
 * S3C2410 Power Manager Advanced Wakeup Management
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/

#include <linux/config.h>
#include <linux/init.h>
#include <linux/suspend.h>
#include <linux/errno.h>
#include <linux/time.h>
#include <linux/interrupt.h>
#include <linux/crc32.h>
#include <linux/ioport.h>
#include <linux/delay.h>

#include <asm/hardware.h>
#include <asm/io.h>

#include <asm/arch/regs-serial.h>
#include <asm/arch/regs-clock.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-mem.h>
#include <asm/arch/regs-irq.h>

#include <asm/mach/time.h>

#include "pm.h"
#include "wake.h"


static struct wake_t wake;

#define to_wake_slot(obj) container_of(obj, struct wake_slot, kobj)
#define param_to_wake_slot(obj) container_of(obj, struct wake_slot, mparam)
#define to_wake_module(obj) container_of(obj, struct wake_module, kobj)

ssize_t	wake_mparam_show(struct kobject *kobj, struct attribute *attr, char *value) {
	struct wake_slot *slot = param_to_wake_slot(kobj);

	if(slot->module && slot->module->ops.get_param)
		return slot->module->ops.get_param(slot->mdata, attr, value);

	return -1;
}

ssize_t	wake_mparam_store(struct kobject *kobj, struct attribute *attr, const char *value, size_t size) {
	struct wake_slot *slot = param_to_wake_slot(kobj);

	if(slot->module && slot->module->ops.set_param)
		return slot->module->ops.set_param(slot->mdata, attr, value, size);

	return -1;
}

static struct sysfs_ops wake_module_sysfs_ops = {
	.show	= wake_mparam_show,
	.store	= wake_mparam_store,
};

static void wake_module_release(struct kobject * kobj)
{

}

static struct kobj_type ktype_wake_module = {
	.sysfs_ops	= &wake_module_sysfs_ops,
	.release	= wake_module_release,
};


static struct sysfs_ops wake_mparam_sysfs_ops = {
	.show	= wake_mparam_show,
	.store	= wake_mparam_store,
};

static struct kobj_type ktype_wake_mparam = {
	.sysfs_ops	= &wake_mparam_sysfs_ops,
};


int s3c2410_wake_module_register(struct wake_module *module) {
	int error;

	error = kobject_set_name(&module->kobj, "%s", module->name);
	if (error) return error;

	module->kobj.kset  = &wake.moduleset;

	if ((error = kobject_register(&module->kobj))) {
		return error;
	}

	return 0;
}

void s3c2410_wake_module_unregister(struct wake_module *module) {
	//TODO: !!!!

	return;
}


static struct attribute wake_slot_attr_module = {
	.name	= "module",
	.mode	= 0644,
};

static struct attribute *wake_slot_attributes[] = {
	&wake_slot_attr_module,
	NULL,
};

ssize_t	wake_slot_show(struct kobject *kobj, struct attribute *attr, char *value) {
	struct wake_slot *slot = to_wake_slot(kobj);

	if(attr == (&wake_slot_attr_module)) {
		if(slot->module) 	return sprintf(value, "%s",slot->module->name);
		else			return sprintf(value, "<empty>");
	}

	return 0;
}

ssize_t	wake_slot_store(struct kobject *kobj, struct attribute *attr,const char *value, size_t size) {
	struct wake_slot *slot = to_wake_slot(kobj);

	if(attr == (&wake_slot_attr_module)) {
		struct kobject *module_object;
		char mname[KOBJ_NAME_LEN];
		int i,l;

		//Remove old module
		if(slot->module) {

			if(slot->module->attributes)
				sysfs_remove_group(&slot->mparam, slot->module->attributes);

			if(slot->module->ops.free_context) {
				slot->module->ops.free_context(slot->mdata);
				slot->mdata = NULL;
			} else {
				if(slot->mdata) {
					printk(KERN_ERR "wake_slot_store: module [%s] has instance data and provides no free_context function. Memory leak!\n", slot->module->name);
				}
				slot->mdata = NULL;
			}

			kobject_put(&slot->module->kobj);
			slot->module=NULL;
		}

		//Find and insert new module
		strncpy(mname, value, min((size_t)(KOBJ_NAME_LEN-1), size));
		mname[min((size_t)(KOBJ_NAME_LEN-1), size)] = 0;
		l = strlen(mname);
		for(i=0;i<l;i++) if(mname[i] == '\n') mname[i] = 0;

		module_object = kset_find_obj(&wake.moduleset, mname);
		if(module_object) {
			slot->module = to_wake_module(module_object);
			if(slot->module->ops.create_context) slot->mdata = slot->module->ops.create_context();
			else slot->mdata = NULL;

			if(slot->module->attributes)
				sysfs_create_group(&slot->mparam, slot->module->attributes);

		} else {
			if(size) printk(KERN_ERR "wake_slot_store: module [%s] not found.\n", mname);
		}

		return size;
	}

	return 0;
}



static struct sysfs_ops wake_slot_sysfs_ops = {
	.show	= wake_slot_show,
	.store	= wake_slot_store,
};


static void wake_slot_release(struct kobject * kobj)
{

}

static struct kobj_type ktype_wake_slot = {
	.sysfs_ops	= &wake_slot_sysfs_ops,
	.release	= wake_slot_release,
	.default_attrs	= wake_slot_attributes,
};

void s3c2410_wake_before_sleep() {
	int tmp;
	for(tmp=0;tmp<WAKE_SLOTS;tmp++) {
		if((wake.slots[tmp].module) && (wake.slots[tmp].module->ops.before_sleep))  {
			wake.slots[tmp].module->ops.before_sleep(wake.slots[tmp].mdata);
		}
	}
}

void s3c2410_wake_after_wakeup() {
	int tmp;
	for(tmp=0;tmp<WAKE_SLOTS;tmp++) {
		if((wake.slots[tmp].module) && (wake.slots[tmp].module->ops.after_wakeup))  {
			wake.slots[tmp].module->ops.after_wakeup(wake.slots[tmp].mdata);
		}
	}

}

int s3c2410_wake_decide_wakeup() {
	int tmp;
	for(tmp=0;tmp<WAKE_SLOTS;tmp++) {
		if((wake.slots[tmp].module) && (wake.slots[tmp].module->ops.decide_wakeup))  {
			int res = wake.slots[tmp].module->ops.decide_wakeup(wake.slots[tmp].mdata);
			if(res != WAKEUP_NONE) return res;
		}
	}

	return WAKEUP_NONE;
}

int __init s3c2410_wake_init(void)
{
	int i;
	int retval=0;

	memzero(&wake, sizeof(wake));

	retval = kobject_set_name(&wake.subsys.kset.kobj, "wake");
	if (retval) goto out;

	subsys_set_kset(&wake, power_subsys);
	retval = subsystem_register(&wake.subsys);
	if (retval) goto out;

	wake.slotset.subsys = &wake.subsys;
	wake.slotset.ktype  = &ktype_wake_slot;

	retval = kobject_set_name(&wake.slotset.kobj, "slots");
	if (retval) goto out;

	retval = kset_register(&wake.slotset);
	if (retval) goto out;

	wake.moduleset.subsys = &wake.subsys;
	wake.moduleset.ktype  = &ktype_wake_module;

	retval = kobject_set_name(&wake.moduleset.kobj, "modules");
	if (retval) goto out;

	retval = kset_register(&wake.moduleset);
	if (retval) goto out;

	wake.mparamset.subsys = &wake.subsys;
	wake.mparamset.ktype  = &ktype_wake_mparam;

	retval = kobject_set_name(&wake.mparamset.kobj, "mparam");
	if (retval) goto out;

	kset_init(&wake.mparamset);

	for(i=0;i<WAKE_SLOTS;i++) {
		kobject_set_name(&wake.slots[i].kobj, "%i", i);
		wake.slots[i].kobj.kset = &wake.slotset;
		kobject_register(&wake.slots[i].kobj);

		kobject_set_name(&wake.slots[i].mparam, "parameters");
		wake.slots[i].mparam.parent	= &wake.slots[i].kobj;
		wake.slots[i].mparam.kset 	= &wake.mparamset;
		kobject_register(&wake.slots[i].mparam);
	}

	printk("S3C24XX Advanced Wakeup Management, Author Thomas Kleffel, (c) 2005 TomTom BV\n");

out:
	return retval;
}

subsys_initcall(s3c2410_wake_init);
