/* linux/arch/arm/mach-s3c2410/wake_switch.c
 *
 * Copyright (c) 2005 TomTom BV
 * Author: Thomas Kleffel <tk@maintech.de>
 *
 * S3C2410 Advanced Wakeup Management - Check ADC channel(s) before waking up
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/

#include <linux/module.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/suspend.h>
#include <linux/errno.h>
#include <linux/time.h>
#include <linux/interrupt.h>
#include <linux/crc32.h>
#include <linux/ioport.h>
#include <linux/delay.h>

#include <asm/hardware.h>
#include <asm/io.h>

#include <asm/arch/regs-clock.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-mem.h>
#include <asm/arch/regs-irq.h>
#include <asm/arch/regs-adc.h>

#include <asm/mach/time.h>
#include <asm/hardware/clock.h>

#include "pm.h"
#include "wake.h"

struct ch_description {
	char		name[10];
	unsigned int	ch;
};

#define NR_OF_CHANNELS (sizeof(channels) / sizeof(struct ch_description))
static struct ch_description channels[] = {
	{.name = "AIN0", .ch = 0},
	{.name = "AIN1", .ch = 1},
	{.name = "AIN2", .ch = 2},
	{.name = "AIN3", .ch = 3},
	{.name = "AIN4", .ch = 4},
	{.name = "AIN5", .ch = 5},
	{.name = "AIN6", .ch = 6},
	{.name = "AIN7", .ch = 7},
};

struct adcwake_context {
	struct ch_description *ch;
	unsigned int min_value;
	unsigned int last_value;
	int have_last_value;
};


static struct attribute adcwake_ch = {
	.name = "channel",
	.mode = 0644,
};

static struct attribute adcwake_mval = {
	.name = "min_value",
	.mode = 0644,
};

static struct attribute adcwake_lval = {
	.name = "last_value",
	.mode = 0644,
};


static struct attribute * adcwake_attr[] = {
	&adcwake_ch,
	&adcwake_mval,
	&adcwake_lval,
	NULL,
};

static struct attribute_group adcwake_params = {
	.attrs		= adcwake_attr,
};

ssize_t adcwake_param_get(void *context, struct attribute *attr, char *value) {
	struct adcwake_context *ctx = (struct adcwake_context *)context;
	if(!ctx) return sprintf(value, "no context");

	if(attr == &adcwake_ch) {
		if(ctx->ch) 	return sprintf(value, ctx->ch->name);
		else		return sprintf(value, "<none>");
	}

	if(attr == &adcwake_mval) return sprintf(value, "%u",ctx->min_value);

	if(attr == &adcwake_lval) {
		if(ctx->have_last_value) return sprintf(value, "%u", ctx->last_value);
		else			 return sprintf(value, "<not available>");
	}

	return -1;
}

ssize_t adcwake_param_set(void *context, struct attribute *attr, const char *value, size_t size) {
	struct adcwake_context *ctx = (struct adcwake_context *)context;
	if(!ctx) return -1;

	if(attr == &adcwake_ch) {
		int c;
		ctx->ch = NULL;
		for(c=0;c<NR_OF_CHANNELS;c++) {
			if(0 == strncmp(value, channels[c].name, min(size, strlen(channels[c].name)))) {
				ctx->ch = &channels[c];
				break;
			}
		}
	}

	if(attr == &adcwake_mval) {
		if(1 != sscanf(value, "%u", &ctx->min_value)) ctx->min_value = 0;
	}

	return size;
}

int adcwake_decide_wakeup(void *context) {
	struct clk *clk;
	int i;
	unsigned long adctsc, adccon;
	unsigned long my_adctsc, my_adccon;
	unsigned long cvalue;
	struct adcwake_context *ctx = (struct adcwake_context *)context;
	if(!(ctx && ctx->ch)) return WAKEUP_NONE;
	clk = clk_get(NULL, "adc");
	if(!clk) return WAKEUP_NONE;

	//Save touch screen registers
	adctsc = __raw_readl(S3C2410_ADCTSC);
	adccon = __raw_readl(S3C2410_ADCCON);

	//Set them up to what we want
	my_adccon = (0xff << 6) | (1 << 14) | (ctx->ch->ch << 3);
	my_adctsc = (1 << 3) | (1 << 4) | (1 << 6);

	//Enable clock and let it settle a bit
	clk_enable(clk);
	mdelay(10);

	//Write my values
	__raw_writel(my_adctsc, S3C2410_ADCTSC);
	__raw_writel(my_adccon, S3C2410_ADCCON);

	for(i=0;i<10;i++) {
		//Start conversion
		__raw_writel(my_adccon | 1, S3C2410_ADCCON );

		//Wait for start of conversion
		//TODO: Timeout?
		while (__raw_readl(S3C2410_ADCCON) & 1);

		//Wait for end of conversion
		//TODO: Timeout?
		while(1) if(__raw_readl(S3C2410_ADCCON) & (1<<15)) break;
	}

	//Read result
	cvalue = __raw_readl(S3C2410_ADCDAT0) & 0x3FF;

	ctx->last_value = cvalue;
	ctx->have_last_value = 1;

	//disable clock and restore original register contents
	__raw_writel(adccon, S3C2410_ADCCON);
	__raw_writel(adctsc, S3C2410_ADCTSC);

	clk_disable(clk);
	clk_put(clk);


	if(cvalue>=ctx->min_value) return WAKEUP_NONE;
	else return WAKEUP_SLEEP;
}

static void * adcwake_create_context(void) {
	struct adcwake_context *context = kmalloc(sizeof(struct adcwake_context), GFP_KERNEL);
	memzero(context, sizeof(struct adcwake_context));

	return context;
}

static void adcwake_free_context(void *context) {
	kfree(context);
}

static struct wake_module adcwake = {
	.name 			= "adc",
	.attributes		= &adcwake_params,
	.ops.set_param		=  adcwake_param_set,
	.ops.get_param		=  adcwake_param_get,
	.ops.decide_wakeup	=  adcwake_decide_wakeup,
	.ops.create_context	=  adcwake_create_context,
	.ops.free_context	=  adcwake_free_context,
};

static int __init adcwake_init(void)
{
	return s3c2410_wake_module_register(&adcwake);
}

static void __exit adcwake_exit(void)
{
	s3c2410_wake_module_unregister(&adcwake);
}

module_init(adcwake_init);
module_exit(adcwake_exit);

MODULE_AUTHOR("Thomas Kleffel <tk@maintech.de> ((c) 2005 TomTom BV)");
MODULE_DESCRIPTION("ADC wake up module for S3C24XX Advanced Wakeup Management");
MODULE_LICENSE("GPL");

