/* linux/arch/arm/mach-s3c2410/wake_switch.c
 *
 * Copyright (c) 2005 TomTom BV
 * Author: Thomas Kleffel <tk@maintech.de>
 *
 * S3C2410 Advanced Wakeup Management - Wakeup from Switch/Button
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/

#include <linux/module.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/suspend.h>
#include <linux/errno.h>
#include <linux/time.h>
#include <linux/interrupt.h>
#include <linux/crc32.h>
#include <linux/ioport.h>
#include <linux/delay.h>

#include <asm/hardware.h>
#include <asm/io.h>

#include <asm/arch/regs-serial.h>
#include <asm/arch/regs-clock.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-mem.h>
#include <asm/arch/regs-irq.h>

#include <asm/mach/time.h>

#include "pm.h"
#include "wake.h"


struct pin_description {
	char		name[10];
	unsigned int	pin;
	unsigned int	cfg;
};

#define NR_OF_PINS (sizeof(pins) / sizeof(struct pin_description))
static struct pin_description pins[] = {
	{.name = "GPF0", .pin = S3C2410_GPF0, .cfg=S3C2410_GPF0_INP},
	{.name = "GPF1", .pin = S3C2410_GPF1, .cfg=S3C2410_GPF1_INP},
	{.name = "GPF2", .pin = S3C2410_GPF2, .cfg=S3C2410_GPF2_INP},
	{.name = "GPF3", .pin = S3C2410_GPF3, .cfg=S3C2410_GPF3_INP},
	{.name = "GPF4", .pin = S3C2410_GPF4, .cfg=S3C2410_GPF4_INP},
	{.name = "GPF5", .pin = S3C2410_GPF5, .cfg=S3C2410_GPF5_INP},
	{.name = "GPF6", .pin = S3C2410_GPF6, .cfg=S3C2410_GPF6_INP},
	{.name = "GPF7", .pin = S3C2410_GPF7, .cfg=S3C2410_GPF7_INP},
	{.name = "GPG0", .pin = S3C2410_GPG0, .cfg=S3C2410_GPG0_INP},
	{.name = "GPG1", .pin = S3C2410_GPG1, .cfg=S3C2410_GPG1_INP},
	{.name = "GPG2", .pin = S3C2410_GPG2, .cfg=S3C2410_GPG2_INP},
	{.name = "GPG3", .pin = S3C2410_GPG3, .cfg=S3C2410_GPG3_INP},
	{.name = "GPG4", .pin = S3C2410_GPG4, .cfg=S3C2410_GPG4_INP},
	{.name = "GPG5", .pin = S3C2410_GPG5, .cfg=S3C2410_GPG5_INP},
	{.name = "GPG6", .pin = S3C2410_GPG6, .cfg=S3C2410_GPG6_INP},
	{.name = "GPG7", .pin = S3C2410_GPG7, .cfg=S3C2410_GPG7_INP},
	{.name = "GPG8", .pin = S3C2410_GPG8, .cfg=S3C2410_GPG8_INP},
	{.name = "GPG9", .pin = S3C2410_GPG9, .cfg=S3C2410_GPG9_INP},
	{.name = "GPG10", .pin = S3C2410_GPG10, .cfg=S3C2410_GPG10_INP},
	{.name = "GPG11", .pin = S3C2410_GPG11, .cfg=S3C2410_GPG11_INP},
	{.name = "GPG12", .pin = S3C2410_GPG12, .cfg=S3C2410_GPG12_INP},
	{.name = "GPG13", .pin = S3C2410_GPG13, .cfg=S3C2410_GPG13_INP},
	{.name = "GPG14", .pin = S3C2410_GPG14, .cfg=S3C2410_GPG14_INP},
	{.name = "GPG15", .pin = S3C2410_GPG15, .cfg=S3C2410_GPG15_INP},
};

struct switchwake_context {
	struct pin_description *pin;
	unsigned int delay;
};


static struct attribute switchwake_pin = {
	.name = "pin",
	.mode = 0644,
};

static struct attribute switchwake_delay = {
	.name = "delay",
	.mode = 0644,
};

static struct attribute * switchwake_attr[] = {
	&switchwake_pin,
	&switchwake_delay,
	NULL,
};

static struct attribute_group switchwake_params = {
	.attrs		= switchwake_attr,
};

ssize_t switchwake_param_get(void *context, struct attribute *attr, char *value) {
	struct switchwake_context *ctx = (struct switchwake_context *)context;
	if(!ctx) return sprintf(value, "no context");

	if(attr == &switchwake_pin) {
		if(ctx->pin) 	return sprintf(value, ctx->pin->name);
		else		return sprintf(value, "<none>");
	}

	if(attr == &switchwake_delay) return sprintf(value, "%u",ctx->delay);

	return -1;
}

ssize_t switchwake_param_set(void *context, struct attribute *attr, const char *value, size_t size) {
	struct switchwake_context *ctx = (struct switchwake_context *)context;
	if(!ctx) return -1;

	if(attr == &switchwake_pin) {
		int p;
		ctx->pin = NULL;
		for(p=0;p<NR_OF_PINS;p++) {
			if(0 == strncmp(value, pins[p].name, min(size, strlen(pins[p].name)))) {
				ctx->pin = &pins[p];
				break;
			}
		}
	}

	if(attr == &switchwake_delay) {
		if(1 != sscanf(value, "%u", &ctx->delay)) ctx->delay = 0;
	}

	return size;
}

void switchwake_before_sleep(void *context) {
	struct switchwake_context *ctx = (struct switchwake_context *)context;

	if(ctx && ctx->pin)
		enable_irq_wake(s3c2410_gpio_getirq(ctx->pin->pin));
}

void switchwake_after_wakeup(void *context) {
	struct switchwake_context *ctx = (struct switchwake_context *)context;

	if(ctx && ctx->pin)
		enable_irq_wake(s3c2410_gpio_getirq(ctx->pin->pin));
}

int switchwake_decide_wakeup(void *context) {
	int rval;
	int i;
	unsigned int pup, cfg;
	struct switchwake_context *ctx = (struct switchwake_context *)context;

	return WAKEUP_WAKE;

	if(!(ctx && ctx->pin)) return WAKEUP_NONE;

	//Check our pin is the wakeup source...
	if(!irq_asserted(s3c2410_gpio_getirq(ctx->pin->pin))) return WAKEUP_NONE;

	//Check if we have a delay...
	if(!(ctx->delay)) return WAKEUP_WAKE;

	//We have to leave hardware _exactly_ as we found it...
	pup = s3c2410_gpio_get_pullup(ctx->pin->pin);
	cfg = s3c2410_gpio_getcfg(ctx->pin->pin);

	s3c2410_gpio_pullup(ctx->pin->pin, 1);
	s3c2410_gpio_cfgpin(ctx->pin->pin, ctx->pin->cfg);

	for(i=0;i<ctx->delay;i++) {
		mdelay(1);
		if(s3c2410_gpio_getpin(ctx->pin->pin)) {
			rval = WAKEUP_SLEEP;
			goto nullwake_dw_out;
		}
	}

	rval = WAKEUP_WAKE;

nullwake_dw_out:
	s3c2410_gpio_pullup(ctx->pin->pin, pup);
	s3c2410_gpio_cfgpin(ctx->pin->pin, cfg);

	return rval;
}

static void * switchwake_create_context(void) {
	struct switchwake_context *context = kmalloc(sizeof(struct switchwake_context), GFP_KERNEL);
	memzero(context, sizeof(struct switchwake_context));

	context->pin = &pins[0];

	return context;
}

static void switchwake_free_context(void *context) {
	kfree(context);
}

static struct wake_module switchwake = {
	.name 			= "switch",
	.attributes		= &switchwake_params,
	.ops.set_param		=  switchwake_param_set,
	.ops.get_param		=  switchwake_param_get,
	.ops.before_sleep	=  switchwake_before_sleep,
	.ops.after_wakeup	=  switchwake_after_wakeup,
	.ops.decide_wakeup	=  switchwake_decide_wakeup,
	.ops.create_context	=  switchwake_create_context,
	.ops.free_context	=  switchwake_free_context,
};

static int __init switchwake_init(void)
{
	return s3c2410_wake_module_register(&switchwake);
}

static void __exit switchwake_exit(void)
{
	s3c2410_wake_module_unregister(&switchwake);
}

module_init(switchwake_init);
module_exit(switchwake_exit);

MODULE_AUTHOR("Thomas Kleffel <tk@maintech.de> ((c) 2005 TomTom BV)");
MODULE_DESCRIPTION("Switch wake up module for S3C24XX Advanced Wakeup Management");
MODULE_LICENSE("GPL");

