/* linux/arch/arm/mach-s3c2410/wake_ttg_bat.c
 *
 * Copyright (c) 2005 TomTom BV
 * Author: Thomas Kleffel <tk@maintech.de>
 *
 * S3C2410 Advanced Wakeup Management - TomTom Go's battery status
 *	before waking up.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/

#include <linux/module.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/suspend.h>
#include <linux/errno.h>
#include <linux/time.h>
#include <linux/interrupt.h>
#include <linux/crc32.h>
#include <linux/ioport.h>
#include <linux/delay.h>

#include <asm/hardware.h>
#include <asm/io.h>

#include <asm/arch/regs-clock.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-mem.h>
#include <asm/arch/regs-irq.h>
#include <asm/arch/regs-adc.h>

#include <asm/mach/time.h>
#include <asm/hardware/clock.h>

#include "pm.h"
#include "wake.h"

struct ttgb_context {
	unsigned int min_value;
	unsigned int last_value;
	int have_last_value;
};

static struct attribute ttgb_mval = {
	.name = "min_value",
	.mode = 0644,
};

static struct attribute ttgb_lval = {
	.name = "last_value",
	.mode = 0644,
};


static struct attribute * ttgb_attr[] = {
	&ttgb_mval,
	&ttgb_lval,
	NULL,
};

static struct attribute_group ttgb_params = {
	.attrs		= ttgb_attr,
};

ssize_t ttgb_param_get(void *context, struct attribute *attr, char *value) {
	struct ttgb_context *ctx = (struct ttgb_context *)context;
	if(!ctx) return sprintf(value, "no context");

	if(attr == &ttgb_mval) return sprintf(value, "%u",ctx->min_value);

	if(attr == &ttgb_lval) {
		if(ctx->have_last_value) return sprintf(value, "%u", ctx->last_value);
		else			 return sprintf(value, "<not available>");
	}

	return -1;
}

ssize_t ttgb_param_set(void *context, struct attribute *attr, const char *value, size_t size) {
	struct ttgb_context *ctx = (struct ttgb_context *)context;
	if(!ctx) return -1;

	if(attr == &ttgb_mval) {
		if(1 != sscanf(value, "%u", &ctx->min_value)) ctx->min_value = 0;
	}

	return size;
}

int ttgb_decide_wakeup(void *context) {
	struct clk *clk;
	int i;
	unsigned long adctsc, adccon;
	unsigned int pup, cfg, val;
	unsigned long my_adctsc, my_adccon;
	unsigned long battery, reference, voltage;
	struct ttgb_context *ctx = (struct ttgb_context *)context;
	if(!ctx) return WAKEUP_NONE;

	clk = clk_get(NULL, "adc");
	if(!clk) return WAKEUP_NONE;

	//Save touch screen & ADC registers
	adctsc = __raw_readl(S3C2410_ADCTSC);
	adccon = __raw_readl(S3C2410_ADCCON);

	//Save GPIO registers
	pup = s3c2410_gpio_get_pullup(S3C2410_GPA18);
	cfg = s3c2410_gpio_getcfg(S3C2410_GPA18);
	val = s3c2410_gpio_getpin(S3C2410_GPA18);

	//Activate reference voltage source
	s3c2410_gpio_setpin(S3C2410_GPA18, 1);
	s3c2410_gpio_cfgpin(S3C2410_GPA18, S3C2410_GPA18_OUT);

	//Setup adc
	my_adccon = (0xff << 6) | (1 << 14);
	my_adctsc = (1 << 3) | (1 << 4) | (1 << 6);

	//Enable clock and let it settle a bit
	clk_enable(clk);
	mdelay(10);

	//Write my values
	__raw_writel(my_adctsc, S3C2410_ADCTSC);
	__raw_writel(my_adccon, S3C2410_ADCCON);

	//Battery and reference voltages are averaged over 10 samples.
	battery = reference = 0;

	for(i=0;i<10;i++) {
		//Select battery voltage channel and wait a while
		__raw_writel(my_adccon | 0 | (0 << 3), S3C2410_ADCCON );
		mdelay(5);
		//Start conversion of battery voltage
		__raw_writel(my_adccon | 1 | (0 << 3), S3C2410_ADCCON );

		//Wait for start of and end of conversion
		//TODO: Timeout?
		while (__raw_readl(S3C2410_ADCCON) & 1);
		while (!(__raw_readl(S3C2410_ADCCON) & (1<<15)));
		//Read result;
		battery+= __raw_readl(S3C2410_ADCDAT0) & 0x3FF;

		//Select reference voltage and wait a while
		__raw_writel(my_adccon | 0 | (4 << 3), S3C2410_ADCCON );
		mdelay(5);
		//Start conversion of reference voltage
		__raw_writel(my_adccon | 1 | (4 << 3), S3C2410_ADCCON );

		//Wait for start of and end of conversion
		//TODO: Timeout?
		while (__raw_readl(S3C2410_ADCCON) & 1);
		while (!(__raw_readl(S3C2410_ADCCON) & (1<<15)));
		//Read result;
		reference+= __raw_readl(S3C2410_ADCDAT0) & 0x3FF;
	}

	battery/=10;
	reference/=10;

	//Calculate real battery voltage
	voltage=2450 * battery;
	voltage/=reference;


	ctx->last_value = voltage;
	ctx->have_last_value = 1;

	//disable clock and restore original register contents
	__raw_writel(adccon, S3C2410_ADCCON);
	__raw_writel(adctsc, S3C2410_ADCTSC);

	clk_disable(clk);
	clk_put(clk);

	//restore GPIO pin status
	//Save GPIO registers
	s3c2410_gpio_pullup(S3C2410_GPA18, pup);
	s3c2410_gpio_cfgpin(S3C2410_GPA18, cfg);
	s3c2410_gpio_setpin(S3C2410_GPA18, val);

	if(voltage>=ctx->min_value) return WAKEUP_NONE;
	else return WAKEUP_SLEEP;
}

static void * ttgb_create_context(void) {
	struct ttgb_context *context = kmalloc(sizeof(struct ttgb_context), GFP_KERNEL);
	memzero(context, sizeof(struct ttgb_context));

	return context;
}

static void ttgb_free_context(void *context) {
	kfree(context);
}

static struct wake_module ttgb = {
	.name 			= "ttgo_bat",
	.attributes		= &ttgb_params,
	.ops.set_param		=  ttgb_param_set,
	.ops.get_param		=  ttgb_param_get,
	.ops.decide_wakeup	=  ttgb_decide_wakeup,
	.ops.create_context	=  ttgb_create_context,
	.ops.free_context	=  ttgb_free_context,
};

static int __init ttgb_init(void)
{
	return s3c2410_wake_module_register(&ttgb);
}

static void __exit ttgb_exit(void)
{
	s3c2410_wake_module_unregister(&ttgb);
}

module_init(ttgb_init);
module_exit(ttgb_exit);

MODULE_AUTHOR("Thomas Kleffel <tk@maintech.de> ((c) 2005 TomTom BV)");
MODULE_DESCRIPTION("TTGo battery check module for Advanced Wakeup Management");
MODULE_LICENSE("GPL");

