/* drivers/barcelona/adc/adc.c
 *
 * Implementation of the ADC driver.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 * Author: Koen Martens <kmartens@sonologic.nl>
 * Author: Jeroen Taverne <jeroen.taverne@tomtom.com>
 * Author: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Description:
 *
 * The ADC driver, used by other drivers.  Starts a timer upon initialisation,
 * and takes care of ADC sampling.  Other drivers access these values through
 * a public array.  Previously part of the accelerometer driver.
 */

/* Includes */
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/hardware/clock.h>
#include <asm/arch/regs-adc.h>
#include <asm/arch/regs-timer.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-dsc.h>
#include <barcelona/debug.h>
#include <barcelona/timer.h>
#include <barcelona/Barc_adc.h>
#include <barcelona/gotype.h>

/* Defines */
#define PFX "adc: "
#define PK_DBG PK_DBG_FUNC
#undef ADC_SPEED_TEST

#define ADC_RAWRATE (2 * ADC_CHANNELS * ADC_RATE)
#define ADC_DIVIDER ((SYS_TIMER01_PRESCALER + 1) * SYS_TIMER1_DIVIDER * ADC_RAWRATE)

/* Kernel resources */
static int adc_irq;
static struct clk *adc_clock;

/* ADC measured values buffer */
static short adc_buffer[ADC_CHANNELS];
static rwlock_t adc_buffer_lock = RW_LOCK_UNLOCKED;

/* Poll function list */
struct adc_pollfunc_node {
	struct list_head node;
	adc_pollfunc_t func;
};
static LIST_HEAD(adc_pollfunc_list);
static spinlock_t adc_pollfunc_list_lock = SPIN_LOCK_UNLOCKED;

static int cycle;
static int channel;
static char tsConfigXY = (0 << 7) | (1 << 6) | (0 << 5) | (1 << 4);

void adc_get_buffer(short *buf)
{
	unsigned long flags;

	read_lock_irqsave(&adc_buffer_lock, flags);
	memcpy(buf, adc_buffer, ADC_BUFSIZE);
	read_unlock_irqrestore(&adc_buffer_lock, flags);
}
EXPORT_SYMBOL(adc_get_buffer);

short adc_get_channel(int channel)
{
	short val;
	unsigned long flags;

	if (channel < 0 || channel >= ADC_CHANNELS)
		return -EINVAL;

	read_lock_irqsave(&adc_buffer_lock, flags);
	val = adc_buffer[channel];
	read_unlock_irqrestore(&adc_buffer_lock, flags);

	return val;
}
EXPORT_SYMBOL(adc_get_channel);

int adc_register_poll(adc_pollfunc_t func)
{
	struct adc_pollfunc_node *buf;
	unsigned long flags;

	buf = kmalloc(sizeof(struct adc_pollfunc_node), GFP_KERNEL);
	if (buf == NULL) {
		PK_ERR("Unable to allocate pollfunc node\n");
		return -ENOMEM;
	}

	/* Quite a bit of overhead for a simple pointer... */
	buf->func = func;

	spin_lock_irqsave(&adc_pollfunc_list_lock, flags);
	list_add_tail(&buf->node, &adc_pollfunc_list);
	spin_unlock_irqrestore(&adc_pollfunc_list_lock, flags);

	return 0;
}
EXPORT_SYMBOL(adc_register_poll);

int adc_unregister_poll(adc_pollfunc_t func)
{
	unsigned long flags;
	struct adc_pollfunc_node *p;

	spin_lock_irqsave(&adc_pollfunc_list_lock, flags);
	list_for_each_entry(p, &adc_pollfunc_list, node) {
		if (p->func == func) {
			list_del(&p->node);
			kfree(p);
			spin_unlock_irqrestore(&adc_pollfunc_list_lock, flags);
			return 0;
		}
	}
	spin_unlock_irqrestore(&adc_pollfunc_list_lock, flags);

	PK_WARN("Pollfunc %p not found\n", func);
	return -EINVAL;
}
EXPORT_SYMBOL(adc_unregister_poll);

static inline void TouchscreenSelectChannel(int channel)
{
	switch (channel)
	{
		case 0:
		// Set pen down measurement
		__raw_writel((0 << 7) | (1 << 6) | (0 << 5) | (1 << 4), S3C2410_ADCTSC);

		// Enable TS Y plane for reading
		IO_Deactivate(XMON);
		IO_Deactivate(XPON);
		IO_Activate(YMON);
		IO_Deactivate(YPON);
		break;

		case 3:
		// Set XY measurement
		__raw_writel(tsConfigXY, S3C2410_ADCTSC);

		// Enable TS X plane for reading
		IO_Deactivate(YMON);
		IO_Deactivate(YPON);
		IO_Activate(XMON);
		IO_Activate(XPON);
		break;

		case 6:
		// Set XY measurement
		__raw_writel(tsConfigXY, S3C2410_ADCTSC);
		
		// Enable TS Y plane for reading
		IO_Deactivate(XMON);
		IO_Deactivate(XPON);
		IO_Activate(YMON);
		IO_Activate(YPON);
		break;
	}
}

static irqreturn_t adc_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	static short tmpbuf[ADC_CHANNELS];
	static const char chanlist[] = { 0, 1, 7, 2, 3, 5, 4, 6, 7 };
	unsigned long flags;
#ifdef ADC_SPEED_TEST
	static unsigned testcount = 0;

	if (testcount == 0) {
#ifdef ADC_SPEED_TEST_TIME
		struct timeval tv;
		do_gettimeofday(&tv);
		PK_DBG("TIME = %ld.%06ld\n", tv.tv_sec, tv.tv_usec);
#else
		PK_DBG("jiffies = %llu, tcnto1 = %u, tcnto4 = %u\n", get_jiffies_64(), __raw_readl(S3C2410_TCNTO(1)), __raw_readl(S3C2410_TCNTO(4)));
#endif
		testcount = ADC_RAWRATE - 1;
	} else {
		--testcount;
	}
#endif /* ADC_SPEED_TEST */

	if (cycle == 0) {
		/* This wait is needed, else the conversion will not take place */
		while (__raw_readl(S3C2410_ADCCON) & 1)
			;

		/* At even interrupt read and store the value and select the next channel */
		tmpbuf[channel] = __raw_readl(S3C2410_ADCDAT0) & 0x3ff;

		/* Jump to next channel */
		if (++channel >= ADC_CHANNELS)
			channel = 0;

		TouchscreenSelectChannel(channel);

		/* Select channel */
		__raw_writel((1 << 14) | (0xff << 6) | (chanlist[channel] << 3), S3C2410_ADCCON);

#ifdef CONFIG_SMDK2440_BOARD
		/* Set values for not connected channels
		 * {0,1,7,2,3,5,4,6,7} */
		tmpbuf[ADC_BAT]      = 1023;
		tmpbuf[ADC_ACC_TEMP] = 512;
		tmpbuf[ADC_ACC_Y]    = 512;
		tmpbuf[ADC_ACC_X]    = 512;
		tmpbuf[ADC_REF]      = 380;
		tmpbuf[ADC_CHRG]     = 0;
#endif /* CONFIG_SMDK2440_BOARD */

		cycle = 1;
	} else {
		/* At odd interrupt start next conversion */
		__raw_writel(__raw_readl(S3C2410_ADCCON) | 0x1, S3C2410_ADCCON);

		if (channel == 0) {
			struct adc_pollfunc_node *p;

			/* When all channels have been converted, update the buffer atomically. */
			write_lock_irqsave(&adc_buffer_lock, flags);
			memcpy((void *) adc_buffer, (const void *) tmpbuf, ADC_BUFSIZE);
			write_unlock_irqrestore(&adc_buffer_lock, flags);

			/* Then call all the registered poll functions. */
			spin_lock_irqsave(&adc_pollfunc_list_lock, flags);
			list_for_each_entry(p, &adc_pollfunc_list, node) {
				p->func(tmpbuf);
			}
			spin_unlock_irqrestore(&adc_pollfunc_list_lock, flags);
		}

		cycle = 0;
	}

	return IRQ_HANDLED;
}

static ssize_t adc_read(struct file *file, char __user *data, size_t len, loff_t *ppos)
{
	int ret;
	unsigned long flags;

	/* validate the size count passed */
	if (len < ADC_BUFSIZE)
		return -EINVAL;

	/* check user buffer sanity */
	if (!access_ok(VERIFY_WRITE, data, ADC_BUFSIZE))
		return -EFAULT;

	read_lock_irqsave(&adc_buffer_lock, flags);
	ret = __copy_to_user(data, adc_buffer, ADC_BUFSIZE) ? -EFAULT : ADC_BUFSIZE;
	read_unlock_irqrestore(&adc_buffer_lock, flags);

	return ret;
}

static int adc_open(struct inode *inode, struct file *file)
{
	return nonseekable_open(inode, file);
}

static int adc_release(struct inode *inode, struct file *file)
{
	return 0;
}

/* Kernel interface */
static struct file_operations adc_fops = {
	.owner		= THIS_MODULE,
	.read		= adc_read,
	.open		= adc_open,
	.release	= adc_release,
};

static int adc_hw_init(struct device *dev)
{
	unsigned rate;
	unsigned tcntb1;
	unsigned tcntb1_mod;
	unsigned tcon;

	cycle = channel = 0;
	
	// Without external FETs there is no need for a pullup during measurement
	if (!IO_HaveTsFets()) tsConfigXY |= (1 << 3);

	PK_DBG("Getting ADC clock\n");
	adc_clock = clk_get(dev, "adc");
	if (adc_clock == NULL) {
		PK_ERR("Failed to get ADC clock\n");
		return -ENOENT;
	}

	PK_DBG("Using and enabling ADC clock\n");
	clk_use(adc_clock);
	clk_enable(adc_clock);

	rate = clk_get_rate(adc_clock);
	PK_DBG("ADC clock rate: %u\n", rate);

	tcntb1 = rate / ADC_DIVIDER - 1;
	tcntb1_mod = rate % ADC_DIVIDER;
	PK_DBG("tcntb1 = %#x, tcntb1_mod = %#x\n", tcntb1, tcntb1_mod);
	__raw_writel(tcntb1, S3C2410_TCNTB(1));
	__raw_writel(0, S3C2410_TCMPB(1));

	PK_DBG("Starting timer 1\n");
	tcon = __raw_readl(S3C2410_TCON);
	tcon &= ~(0xf << 8);
	tcon |= S3C2410_TCON_T1RELOAD;
	tcon |= S3C2410_TCON_T1MANUALUPD;
	__raw_writel(tcon, S3C2410_TCON); /* Stop and manual load */

	tcon &= ~(0xf << 8);
	tcon |= S3C2410_TCON_T1RELOAD;
	tcon |= S3C2410_TCON_T1START;
	__raw_writel(tcon, S3C2410_TCON); /* Start */

	PK_DBG("Turning on AIN4_PWR\n");
	IO_Activate(AIN4_PWR);

	PK_DBG("Done\n");
	return 0;
}

static void adc_hw_exit(void)
{
	PK_DBG("Turning off AIN4_PWR\n");
	IO_Deactivate(AIN4_PWR);

	PK_DBG("Stopping timer 1\n");
	__raw_writel((__raw_readl(S3C2410_TCON) & ~(0xf << 8)) | (0xa << 8), S3C2410_TCON); /* Stop and manual load */

	if (adc_clock != NULL) {
		PK_DBG("Removing clock\n");
		clk_disable(adc_clock);
		clk_unuse(adc_clock);
		clk_put(adc_clock);
		adc_clock = NULL;
	}

	// Put ADC in standby mode
	__raw_writel(0x3fc4, S3C2410_ADCCON);
	__raw_writel(0x0058, S3C2410_ADCTSC);

	// Turn off TS supply, but leave negative power supply enabled to avoid floating CPU pins
	IO_Activate(XMON);
	IO_Deactivate(XPON);
	IO_Activate(YMON);
	IO_Deactivate(YPON);

	PK_DBG("Done\n");
}

static int adc_probe(struct device *dev)
{
	int ret;

	disable_irq(IRQ_TC);
	disable_irq(IRQ_ADC);

	PK_DBG("Requesting timer IRQ\n");
	ret = request_irq(IRQ_TIMER1, adc_irq_handler, 0, "ADC poll timer", dev);
	if (ret < 0) {
		PK_ERR("Failed to request timer IRQ (%d)\n", ret);
		return ret;
	}
	adc_irq = 1;

	ret = adc_hw_init(dev);
	if (ret < 0) {
		PK_ERR("Failed to initialize hardware (%d)\n", ret);
		return ret;
	}

	PK_DBG("Registering chardev\n");
	ret = register_chrdev(ADC_MAJOR, ADC_DEVNAME, &adc_fops);
	if (ret < 0) {
		PK_ERR("Unable to register chardev on major=%d (%d)\n", ADC_MAJOR, ret);
		return ret;
	}

	PK_DBG("Done\n");
	return 0;
}

static int adc_remove(struct device *dev)
{
	PK_DBG("Unregistering chardev\n");
	unregister_chrdev(ADC_MAJOR, ADC_DEVNAME);

	adc_hw_exit();

	if (adc_irq != 0) {
		PK_DBG("Freeing IRQ\n");
		free_irq(IRQ_TIMER1, dev);
		adc_irq = 0;
	}

	if (!list_empty(&adc_pollfunc_list)) {
		struct adc_pollfunc_node *p;
		struct adc_pollfunc_node *q;

		PK_WARN("Pollfunc list is not empty\n");
		list_for_each_entry_safe(p, q, &adc_pollfunc_list, node) {
			list_del(&p->node);
			kfree(p);
		}
		INIT_LIST_HEAD(&adc_pollfunc_list);
	}

	PK_DBG("Done\n");
	return 0;
}

static void adc_shutdown(struct device *dev)
{
	PK_DBG("Shutting down\n");
	adc_hw_exit();
}

#ifdef CONFIG_PM

static int adc_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("dev = %p, state = %u, level = %u\n", dev, state, level);
	if (level == SUSPEND_POWER_DOWN) {
		adc_hw_exit();
	}
	return 0;
}

static int adc_resume(struct device *dev, u32 level)
{
	PK_DBG("dev = %p, level = %u\n", dev, level);
	if (level == RESUME_POWER_ON) {
		int ret = adc_hw_init(dev);
		if (ret < 0) {
			PK_ERR("Failed to initialize hardware (%d)\n", ret);
			return ret;
		}
	}
	return 0;
}

#else /* CONFIG_PM */
#define adc_suspend NULL
#define adc_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver adc_driver = {
	.name		= "tomtomgo-adc",
	.bus		= &platform_bus_type,
	.probe		= adc_probe,
	.remove		= adc_remove,
	.shutdown	= adc_shutdown,
	.suspend	= adc_suspend,
	.resume		= adc_resume,
};

static int __init adc_mod_init(void)
{
	int ret;

	printk(KERN_INFO "TomTom GO ADC Driver, (C) 2004,2005 TomTom BV\n");
	PK_DBG("Registering driver\n");
	ret = driver_register(&adc_driver);
	if (ret) {
		PK_ERR("Unable to register driver (%d)\n", ret);
		return ret;
	}
	PK_DBG("Done\n");
	return 0;
}

static void __exit adc_mod_exit(void)
{
	PK_DBG("Unregistering driver\n");
	driver_unregister(&adc_driver);
	PK_DBG("Done\n");
}

module_init(adc_mod_init);
module_exit(adc_mod_exit);

MODULE_AUTHOR("Koen Martens <kmartens@sonologic.nl>, Jeroen Taverne <jeroen.taverne@tomtom.com> and Dimitry Andric <dimitry.andric@tomtom.com>");
MODULE_DESCRIPTION("TomTom GO ADC Driver");
MODULE_LICENSE("GPL");

/* EOF */
