/* drivers/barcelona/gpio/gpio.c
 *
 * Implementation of the GPIO driver.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 * Author: Dimitry Andric <dimitry.andric@tomtom.com>
 * Author: Koen Martens <kmartens@sonologic.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* Includes */
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/stat.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/hardware/clock.h>
#include <asm/arch/regs-gpio.h>
#include <barcelona/debug.h>
#include <barcelona/Barc_Gpio.h>
#include <barcelona/gotype.h>

/* Defines */
#define PFX "gpio: "
#define PK_DBG PK_DBG_FUNC

#define TOMTOM_USB_VBUS_DETECT

#define GPIO_POLL_DELAY (HZ / 5)

/* Forward declarations */
static void gpio_status_poll(unsigned long data);

/* Globals */
static HARDWARE_STATUS gpio_hw_status;
static unsigned long gpio_pin_status = 0;
static unsigned int diskAccessStarted = 0;
static spinlock_t gpio_hw_status_lock = SPIN_LOCK_UNLOCKED;
static struct timer_list gpio_timer = TIMER_INITIALIZER(gpio_status_poll, 0, 0);
static int power_button_timer = -1;
static int ignition_timer = -1;
static int usb_timer = -1;
static int diskAccess = 0;

/* Functions */
static void gpio_update_hw_status(void)
{
	UINT8 input_status = 0;
	UINT8 charge_status;
	unsigned long flags;
	
	// USB active?
	if (IO_GetInput(USB_HOST_DETECT)) 
	{
#ifdef TOMTOM_USB_VBUS_DETECT
		// Count down timer when USB was deactivated
		if (usb_timer > 0)
		{
			usb_timer--;
		}
		else
		if (usb_timer == 0)
		{
			// Disable timer to wait for connection of external power
			usb_timer = -1;
			input_status |= USB_DETECT_MASK;
		}
#endif
	}
	else
	{
		// Enable running timer when USB is deactivated (wait 1 second)
		usb_timer = 5;
	}
	
	// Button pressed?
	if (IO_GetInput(ON_OFF)) 
	{
		// Count down timer when button was released
		if (power_button_timer > 0)
		{
			power_button_timer--;
		}
		else
		if (power_button_timer == 0)
		{
#if 0
			// Give direct response to the user by making TFT black, because the application sometimes responds too slow
			IO_Deactivate(VD2);
			IO_Deactivate(VD3);
			IO_Deactivate(VD4);
			IO_Deactivate(VD5);
			IO_Deactivate(VD6);
			IO_Deactivate(VD7);
			IO_Deactivate(VD10);
			IO_Deactivate(VD11);
			IO_Deactivate(VD12);
			IO_Deactivate(VD13);
			IO_Deactivate(VD14);
			IO_Deactivate(VD15);
			IO_Deactivate(VD18);
			IO_Deactivate(VD19);
			IO_Deactivate(VD20);
			IO_Deactivate(VD21);
			IO_Deactivate(VD22);
			IO_Deactivate(VD23);			
#endif
			// Disable timer to wait for button release
			power_button_timer = -1;
			// Send message to application to shut down
			input_status |= ONOFF_MASK;
		}
	}
	else
	{
		// Enable running timer when button is released (wait 1/5 second)
		power_button_timer = 1;
	}
	
	//auto-switch off if in dock and power has been off for more than five seconds
	{
		// Check if docked
		if ((IO_GetInput(DOCKSNS0)) || (IO_GetInput(DOCKSNS1))) 
		{
			// External power?
			if (IO_GetInput(ACPWR)) 
			{
				// Enable running timer when external power is removed (wait 5 seconds)
				ignition_timer = 25;
			}
			else
			{
				// Count down timer when external power was active while docked
				if (ignition_timer > 0)
				{
					ignition_timer--;
				}
				else
				if (ignition_timer == 0)
				{
					// Disable timer to wait for connection of external power
					ignition_timer = -1;
					// Send message to application to shut down
					input_status |= ONOFF_MASK;
				}
			}
		}
		else 
		{
			// Not docked, so disable timer
			ignition_timer = -1;
		}
	}

	// Power state
	if (IO_GetInput(ACPWR))
	{
		if (IO_GetInput(CHARGEFAULT))
		{
			charge_status = CHARGE_STATE_FAULT;
		}
		else if (IO_GetInput(CHARGING))
		{
			IO_Activate(CHARGE_OUT);
			if (IO_GetInput(CHARGING))
			{
				charge_status = CHARGE_STATE_TOPPINGUP;
			}
			else
			{
				charge_status = CHARGE_STATE_FAST_CHARGING;
			}
			IO_Deactivate(CHARGE_OUT);
		}
		else
		{
			charge_status = CHARGE_STATE_COMPLETE;
		}
	}
	else
	{
		charge_status = CHARGE_STATE_ACPWR_OFF;
	}

	// Store status now so it can be read by the ioctl
	spin_lock_irqsave(&gpio_hw_status_lock, flags);
	
	// Docking signals
	if (IO_GetInput(DOCKSNS0)) gpio_hw_status.u8InputStatus |= DOCK_SENSE0_MASK; else gpio_hw_status.u8InputStatus &= ~DOCK_SENSE0_MASK;
	if (IO_GetInput(DOCKSNS1)) gpio_hw_status.u8InputStatus |= DOCK_SENSE1_MASK; else gpio_hw_status.u8InputStatus &= ~DOCK_SENSE1_MASK;
	if (IO_GetInput(DOCK_GP1)) gpio_hw_status.u8InputStatus |= DOCK_LIGHTS_MASK; else gpio_hw_status.u8InputStatus &= ~DOCK_LIGHTS_MASK;
	
	// Power button and USB detection (this is or'ed, so it cannot be missed by the application at all)
	gpio_hw_status.u8InputStatus |= input_status;
	
	// Power state
	gpio_hw_status.u8ChargeStatus = charge_status;

	// Disk access?
	if (IO_GetInput(HD_LED))
	{
		// Transition to disk access started?
		if (diskAccess == 0)
		{
			diskAccess = 1;
			diskAccessStarted = 1;
		}
	} else diskAccess = 0;
	
	spin_unlock_irqrestore(&gpio_hw_status_lock, flags);
}

static void gpio_status_poll(unsigned long data)
{
	gpio_update_hw_status();
	mod_timer(&gpio_timer, jiffies + GPIO_POLL_DELAY);
}

static inline unsigned long gpio_read_pin_status(void)
{
	return gpio_pin_status;
}

static void gpio_mute_ext_speaker(int on)
{
	if (on)
	{
		IO_Activate(MUTE_EXT);
	}
	else
	{
		IO_Deactivate(MUTE_EXT);
	}
}

static inline void gpio_write_pin_status(unsigned long status)
{
	/* Store status for reading */
	gpio_pin_status = status;

	gpio_mute_ext_speaker(status & MUTE_EXT_SPEAKER_MASK);
}

#if 0

static inline void gpio_usb(int on)
{
	if (on)
	{
		IO_Activate(USB_PULL_EN);
	}
	else
	{
		IO_Deactivate(USB_PULL_EN);
	}
}

#endif

static int gpio_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int ret = 0;
	unsigned long flags;
	
	switch (cmd) {
	case IOR_HWSTATUS:
		spin_lock_irqsave(&gpio_hw_status_lock, flags);
		ret = copy_to_user((void __user *) arg, &gpio_hw_status, sizeof gpio_hw_status) ? -EFAULT : 0;
		spin_unlock_irqrestore(&gpio_hw_status_lock, flags);
		break;

	case IOW_MUTE_ON_EXTERNAL_SPEAKER:
		gpio_mute_ext_speaker(1);
		break;

	case IOW_MUTE_OFF_EXTERNAL_SPEAKER:
		gpio_mute_ext_speaker(0);
		break;

	case IOR_DISK_ACCESS:
		spin_lock_irqsave(&gpio_hw_status_lock, flags);
		ret = copy_to_user((void __user *) arg, &diskAccessStarted, sizeof diskAccessStarted) ? -EFAULT: 0;
		diskAccessStarted = 0;
		spin_unlock_irqrestore(&gpio_hw_status_lock, flags);
		break;

	default:
		PK_WARN("Invalid ioctl command %u\n", cmd);
		ret = -EINVAL;
		break;
	}

	return ret;
}

static int gpio_open(struct inode *inode, struct file *file)
{
	return nonseekable_open(inode, file);
}

static int gpio_release(struct inode *inode, struct file *file)
{
	return 0;
}

/* Kernel interface */
static struct file_operations gpio_fops = {
	.owner		= THIS_MODULE,
	.ioctl		= gpio_ioctl,
	.open		= gpio_open,
	.release	= gpio_release,
};

static void gpio_hw_init(void)
{
	// Stop timers
	power_button_timer = -1;
	ignition_timer = -1;
	usb_timer = -1;
	// Clear state from before suspend
	gpio_hw_status.u8InputStatus = 0;
	// Get new state
	gpio_update_hw_status();
	mod_timer(&gpio_timer, jiffies + GPIO_POLL_DELAY);
}

static void gpio_hw_exit(void)
{
	gpio_mute_ext_speaker(0);
	del_timer_sync(&gpio_timer);
}

static int gpio_probe(struct device *dev)
{
	int ret;

	gpio_hw_init();

	PK_DBG("Registering chardev\n");
	ret = register_chrdev(GPIO_MAJOR, GPIO_DEVNAME, &gpio_fops);
	if (ret < 0) {
		PK_ERR("Unable to register chardev on major=%d (%d)\n", GPIO_MAJOR, ret);
		return ret;
	}

	PK_DBG("Done\n");
	return 0;
}

static int gpio_remove(struct device *dev)
{
	PK_DBG("Unregistering chardev\n");
	unregister_chrdev(GPIO_MAJOR, GPIO_DEVNAME);

	gpio_hw_exit();

	PK_DBG("Done\n");
	return 0;
}

static void gpio_shutdown(struct device *dev)
{
	PK_DBG("Shutting down\n");
	gpio_hw_exit();
}

#ifdef CONFIG_PM

static int gpio_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("dev = %p, state = %u, level = %u\n", dev, state, level);
	if (level == SUSPEND_POWER_DOWN) {
		gpio_hw_exit();
	}
	return 0;
}

static int gpio_resume(struct device *dev, u32 level)
{
	PK_DBG("dev = %p, level = %u\n", dev, level);
	if (level == RESUME_POWER_ON) {
		gpio_hw_init();
	}
	return 0;
}

#else /* CONFIG_PM */
#define gpio_suspend NULL
#define gpio_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver gpio_driver = {
	.name		= "tomtomgo-gpio",
	.bus		= &platform_bus_type,
	.probe		= gpio_probe,
	.remove		= gpio_remove,
	.shutdown	= gpio_shutdown,
	.suspend	= gpio_suspend,
	.resume		= gpio_resume,
};

static int __init gpio_mod_init(void)
{
	int ret;

	printk(KERN_INFO "TomTom GO GPIO Driver, (C) 2004,2005 TomTom BV\n");
	PK_DBG("Registering driver\n");
	ret = driver_register(&gpio_driver);
	if (ret) {
		PK_ERR("Unable to register driver (%d)\n", ret);
		return ret;
	}
	PK_DBG("Done\n");
	return 0;
}

static void __exit gpio_mod_exit(void)
{
	PK_DBG("Unregistering driver\n");
	driver_unregister(&gpio_driver);
	PK_DBG("Done\n");
}

module_init(gpio_mod_init);
module_exit(gpio_mod_exit);

MODULE_AUTHOR("Dimitry Andric <dimitry.andric@tomtom.com> and Koen Martens <kmartens@sonologic.nl>");
MODULE_DESCRIPTION("TomTom GO GPIO Driver");
MODULE_LICENSE("GPL");

/* EOF */
