/* drivers/barcelona/hw/hwdrv.c
 *
 * Implementation of the hardware detection driver.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 * Author: Dimitry Andric <dimitry.andric@tomtom.com>
 * Author: Jeroen Taverne <jeroen.taverne@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* Includes */ 
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/hardware/clock.h>
#include <barcelona/debug.h>
#include <barcelona/gotype.h>

/* Defines */
#define PFX "hw: "
#define PK_DBG PK_DBG_FUNC

/* Local variables */
static const char hw_proc_root_name[] = "barcelona";
static struct proc_dir_entry *hw_proc_root;

/* Local types */
struct hw_proc_entry {
	const char *name;
	read_proc_t *func;
};

static int hw_proc_read(char *page, char **start, off_t offset, int count, int *eof, const char *msg)
{
	int avail = strlen(msg) - offset;

	if (count >= avail) {
		count = avail;
		*eof = 1;
	}
	if (count <= 0)
		return 0;

	*start = page;
	memcpy(page, msg + offset, count);

	return count;
}

static int hw_proc_read_int(char *page, char **start, off_t offset, int count, int *eof, int i)
{
	unsigned char msg[16];

	sprintf(msg, "%d\n", i);

	return hw_proc_read(page, start, offset, count, eof, msg);
}

int hw_proc_modelid(char *page, char **start, off_t offset, int count, int *eof, void *data)
{
	return hw_proc_read_int(page, start, offset, count, eof, IO_GetModelId());
}

int hw_proc_modelname(char *page, char **start, off_t offset, int count, int *eof, void *data)
{
	unsigned char msg[128];

	sprintf(msg, "%s\n", IO_GetModelName());

	return hw_proc_read(page, start, offset, count, eof, msg);
}

int hw_proc_cputype(char *page, char **start, off_t offset, int count, int *eof, void *data)
{
	return hw_proc_read_int(page, start, offset, count, eof, IO_GetCpuType());
}

int hw_proc_bluetooth(char *page, char **start, off_t offset, int count, int *eof, void *data)
{
	return hw_proc_read_int(page, start, offset, count, eof, IO_HaveBluetooth());
}

int hw_proc_handsfree(char *page, char **start, off_t offset, int count, int *eof, void *data)
{
	return hw_proc_read_int(page, start, offset, count, eof, IO_HaveHandsfree());
}

int hw_proc_battery(char *page, char **start, off_t offset, int count, int *eof, void *data)
{
	return hw_proc_read_int(page, start, offset, count, eof, IO_HaveBattery());
}

int hw_proc_sdcard(char *page, char **start, off_t offset, int count, int *eof, void *data)
{
	return hw_proc_read_int(page, start, offset, count, eof, IO_HaveSdCard());
}

int hw_proc_harddisk(char *page, char **start, off_t offset, int count, int *eof, void *data)
{
	return hw_proc_read_int(page, start, offset, count, eof, IO_HaveHarddisk());
}

int hw_proc_tsfets(char *page, char **start, off_t offset, int count, int *eof, void *data)
{
	return hw_proc_read_int(page, start, offset, count, eof, IO_HaveTsFets());
}

int hw_proc_capabilities(char *page, char **start, off_t offset, int count, int *eof, void *data)
{
	unsigned char msg[256];

	sprintf(msg,
		"model       : %d\n"
		"cputype     : %d\n"
		"bluetooth   : %d\n"
		"handsfree   : %d\n"
		"battery     : %d\n"
		"sdcard      : %d\n"
		"harddisk    : %d\n"
		"tsfets      : %d\n",
		IO_GetModelId(),
		IO_GetCpuType(),	
		IO_HaveBluetooth(),
		IO_HaveHandsfree(),
		IO_HaveBattery(),
		IO_HaveSdCard(),
		IO_HaveHarddisk(),
		IO_HaveTsFets());

	return hw_proc_read(page, start, offset, count, eof, msg);
}

#define HW_PROC_ENTRY(x) { .name = #x, .func = hw_proc_##x }

static struct hw_proc_entry hw_proc_entries[] = {
	HW_PROC_ENTRY(modelid),
	HW_PROC_ENTRY(modelname),
	HW_PROC_ENTRY(cputype),
	HW_PROC_ENTRY(bluetooth),
	HW_PROC_ENTRY(handsfree),
	HW_PROC_ENTRY(battery),
	HW_PROC_ENTRY(sdcard),
	HW_PROC_ENTRY(harddisk),
	HW_PROC_ENTRY(tsfets),
	HW_PROC_ENTRY(capabilities),
};

static int hw_proc_init(void)
{
	int i;

	hw_proc_root = proc_mkdir(hw_proc_root_name, NULL);
	if (hw_proc_root == NULL) {
		PK_ERR("Unable to create hw /proc root %s\n", hw_proc_root_name);
		return -1;
	}

	for (i = 0; i < ARRAY_SIZE(hw_proc_entries); ++i) {
		if (create_proc_read_entry(hw_proc_entries[i].name, S_IRUGO, hw_proc_root, hw_proc_entries[i].func, 0) == NULL) {
			PK_ERR("Unable to create hw /proc entry %s\n", hw_proc_entries[i].name);
			return -1;
		}
	}

	return 0;
}

static void hw_proc_exit(void)
{
	int i;

	if (hw_proc_root != NULL) {
		for (i = ARRAY_SIZE(hw_proc_entries) - 1; i >= 0; --i) {
			remove_proc_entry(hw_proc_entries[i].name, hw_proc_root);
		}
		remove_proc_entry(hw_proc_root_name, NULL);
		hw_proc_root = NULL;
	}
}

static int hw_probe(struct device *dev)
{
	int ret;

#if 0
	PK_DBG("Detecting hardware\n");
	ret = IO_Init();
	if (ret < 0) {
		PK_ERR("Unable to detect hardware\n");
		return -ENODEV;
	}
#endif

	PK_DBG("Setting up proc entries\n");
	ret = hw_proc_init();
	if (ret < 0) {
		PK_ERR("Unable to setup proc entries\n");
		return -ENODEV;
	}

	PK_DBG("Done\n");
	return 0;
}

static int hw_remove(struct device *dev)
{
	PK_DBG("Removing proc entries\n");
	hw_proc_exit();

	PK_DBG("Done\n");
	return 0;
}

static void hw_shutdown(struct device *dev)
{
	PK_DBG("Shutting down\n");
}

#ifdef CONFIG_PM

static int hw_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("dev = %p, state = %u, level = %u\n", dev, state, level);
	if (level == SUSPEND_POWER_DOWN) {
		hw_shutdown(dev);
	}
	return 0;
}

static int hw_resume(struct device *dev, u32 level)
{
	PK_DBG("dev = %p, level = %u\n", dev, level);
	return 0;
}

#else /* CONFIG_PM */
#define hw_suspend NULL
#define hw_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver hw_driver = {
	.name		= "tomtomgo-hw",
	.bus		= &platform_bus_type,
	.probe		= hw_probe,
	.remove		= hw_remove,
	.shutdown	= hw_shutdown,
	.suspend	= hw_suspend,
	.resume		= hw_resume,
};

static int __init hw_mod_init(void)
{
	int ret;

	printk(KERN_INFO "TomTom GO Hardware Detection Driver, (C) 2004,2005 TomTom BV\n");
	PK_DBG("Registering driver\n");
	ret = driver_register(&hw_driver);
	if (ret) {
		PK_ERR("Unable to register driver (%d)\n", ret);
		return ret;
	}
	PK_DBG("Done\n");
	return 0;
}

static void __exit hw_mod_exit(void)
{
	PK_DBG("Unregistering driver\n");
	driver_unregister(&hw_driver);
	PK_DBG("Done\n");
}

module_init(hw_mod_init);
module_exit(hw_mod_exit);

MODULE_AUTHOR("Dimitry Andric <dimitry.andric@tomtom.com> and Jeroen Taverne <jeroen.taverne@tomtom.com>");
MODULE_DESCRIPTION("TomTom GO Hardware Detection Driver");
MODULE_LICENSE("GPL");

/* EOF */
