/*
 * TomTom Go on/off button driver
 * Copyright by Christian Daniel 2004
 * Published under GPLv2 or newer license
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/input.h>
#include <linux/init.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/interrupt.h>

#include <asm/irq.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/tomtomgo-irq.h>
#include <barcelona/debug.h>
#include <barcelona/gotype.h>

#define PFX "onoff-tomtom: "
#define PK_DBG PK_DBG_FUNC

static struct input_dev onoff_dev;
static int irq;

static irqreturn_t onoff_tomtom_interrupt(int irq, void *dev_id,
	struct pt_regs *fp)
{
	input_report_key(&onoff_dev, KEY_POWER, IO_GetInput(ON_OFF));
	input_sync(&onoff_dev);
	return IRQ_HANDLED;
}

static int onoff_tomtom_probe(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);

	irq = platform_get_irq(pdev, 0);
	if (irq == 0) {
		PK_ERR("failed to get interrupt resource\n");
		return -EINVAL;
	}

	PK_DBG("on/off button using irq:%d\n", irq);

	IO_SetInterrupt(ON_OFF);
	set_irq_type(irq, IRQT_BOTHEDGE);
	if (request_irq(irq, onoff_tomtom_interrupt, 0, "tomtomgo-onoff",
		dev)) {
		PK_ERR("can't allocate irq %d\n", irq);
		return -EBUSY;
	}

	memset(&onoff_dev, 0, sizeof onoff_dev);
	onoff_dev.name = "TomTom On/Off Button";
	onoff_dev.evbit[0] = BIT(EV_KEY);
	onoff_dev.keybit[LONG(KEY_POWER)] = BIT(KEY_POWER);
	input_register_device(&onoff_dev);

	return 0;
}

static int onoff_tomtom_remove(struct device *dev)
{
	/* nothing to be deactivated as we want on/off still be a
	   wake up event */
	return 0;
}

static void onoff_tomtom_shutdown(struct device *dev)
{
	PK_DBG("Shutting down\n");
	/* during shutdown only falling edge is interesting */
	if (irq != 0)
		set_irq_type(irq, IRQT_FALLING);
}

static void onoff_tomtom_powerup(struct device *dev)
{
	PK_DBG("Powering up\n");
	/* make sure on/off works again like expected */
	IO_SetInterrupt(ON_OFF);
	if (irq != 0)
		set_irq_type(irq, IRQT_BOTHEDGE);
}

#ifdef CONFIG_PM

static int onoff_tomtom_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("dev = %p, state = %u, level = %u\n", dev, state, level);
	if(level == SUSPEND_POWER_DOWN) {
		onoff_tomtom_shutdown(dev);
	}
	return 0;
}

static int onoff_tomtom_resume(struct device *dev, u32 level)
{
	PK_DBG("dev = %p, level = %u\n", dev, level);
	if(level == RESUME_ENABLE) {
		onoff_tomtom_powerup(dev);
	}
	return 0;
}

#else /* CONFIG_PM */
#define onoff_tomtom_suspend NULL
#define onoff_tomtom_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver onoff_tomtom_driver =
{
	.name		= "tomtomgo-onoff",
	.bus		= &platform_bus_type,
	.probe		= onoff_tomtom_probe,
	.remove		= onoff_tomtom_remove,
	.shutdown	= onoff_tomtom_shutdown,
	.suspend	= onoff_tomtom_suspend,
	.resume		= onoff_tomtom_resume,
};

static int __init onoff_tomtom_init(void)
{
	int ret;

	printk(KERN_INFO "TomTom GO On/Off-Button Driver\n");
	PK_DBG("Registering driver\n");
	ret = driver_register(&onoff_tomtom_driver);
	if (ret) {
		PK_ERR("Unable to register driver (%d)\n", ret);
		return ret;
	}
	PK_DBG("Done\n");
	return 0;
}

static void __exit onoff_tomtom_exit(void)
{
	PK_DBG("Unregistering driver\n");
	driver_unregister(&onoff_tomtom_driver);
	PK_DBG("Done\n");
}

module_init(onoff_tomtom_init);
module_exit(onoff_tomtom_exit);

MODULE_AUTHOR("Christian Daniel <cd@cdaniel.de>, Dimitry Andric <dimitry.andric@tomtom.com>");
MODULE_DESCRIPTION("TomTom Go On/Off-Button Driver");
MODULE_LICENSE("GPL");

/* EOF */
