/*
 * TomTom GO remote control driver
 * Copyright by Christian Daniel 2004
 * Published under GPLv2 or newer license
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/input.h>
#include <linux/init.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/hardware/clock.h>
#include <asm/irq.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-spi.h>
#include <asm/arch/tomtomgo-irq.h>
#include <barcelona/debug.h>
#include <barcelona/gotype.h>

#define PFX "rc-tomtom: "
#define PK_DBG PK_DBG_FUNC

#define RESSIZE(ressource) (((ressource)->end - (ressource)->start) + 1)

#define SPPRE		(*((u32 *)(rc.spibase + S3C2410_SPPRE)))
#define SPCON		(*((u32 *)(rc.spibase + S3C2410_SPCON)))
#define SPPIN		(*((u32 *)(rc.spibase + S3C2410_SPPIN)))
#define SPSTA		(*((u32 *)(rc.spibase + S3C2410_SPSTA)))
#define SPTDAT		(*((u32 *)(rc.spibase + S3C2410_SPTDAT)))
#define SPRDAT		(*((u32 *)(rc.spibase + S3C2410_SPRDAT)))

#define FSK_FIFOIT	(1 << 15)
#define FSK_FFOV	(1 << 14)
#define FSK_WKUP	(1 << 13)
#define FSK_LBD		(1 << 12)
#define FSK_FFEM	(1 << 11)
#define FSK_DRSSI	(1 << 10)
#define FSK_DQD		(1 << 9)
#define FSK_CRL		(1 << 8)
#define FSK_ATGL	(1 << 7)
#define FSK_ASAME	(1 << 6)
#define FSK_OFFS_MSK	(0x1f)


static struct rc {
  struct input_dev rc_dev;
  int irq;
  unsigned char *spibase;
  struct clk *spiclk;
} rc;

/* functions for s3c24x0 sio interface ****************************************/

/* so far this driver uses the sio interface only in polled mode as the little
 * amount if data isn't worth the effort to maintain a command queue and
 * serveral state machines... */

/* wait until spi is not busy
 * returns 0 when spi is ready
 * returns 1 when timed out */
static int rc_tomtom_busy_spi(void)
{
	int i;

	for(i = 0; i < 100000; i++) {
		if((SPSTA & 0x01) == 0x01)
			return 0;
	}
	PK_ERR("spi timeout\n");
	return 1;
}

/* init the fsk receiver */
static int rc_tomtom_init_spi(struct device *dev, unsigned long pclk_rate)
{
	int i;

	rc.spiclk = clk_get(dev, "spi");
	clk_use(rc.spiclk);
	clk_enable(rc.spiclk);
	IO_SetFunction(SPICLK);
	IO_SetFunction(SPIMSI);
	IO_SetFunction(SPIMSO);
	/* baudrate = PCLK / 2 / (prescaler + 1)
	 * prescaler = (PCLK / 2 / baudrate) - 1 */
	SPPRE = (pclk_rate / 2 / 100000) - 1;
	SPCON = S3C2410_SPCON_SMOD_POLL |
		S3C2410_SPCON_ENSCK |
		S3C2410_SPCON_MSTR |
		S3C2410_SPCON_CPOL_HIGH |
		S3C2410_SPCON_CPHA_FMTB;
	SPPIN = S3C2410_SPPIN_KEEP;
	for(i = 0; i < 9; i++) {
		SPTDAT = 0xff;
		if(rc_tomtom_busy_spi())
			return 1;
	}
	return 0;
}

/* turn off fsk clock during driver unload */
static void rc_tomtom_done_spi(void)
{
	clk_disable(rc.spiclk);
	clk_unuse(rc.spiclk);
	clk_put(rc.spiclk);
}

/* send and receive one byte
 * (*timeout) is set to 1 if the action took too long */
static u8 rc_tomtom_txrx_spi(u8 value, int *timeout)
{
	*timeout = 0;

	if(rc_tomtom_busy_spi())
		goto timeout;

	SPTDAT = value;

	if(rc_tomtom_busy_spi())
		goto timeout;

	return SPRDAT;

timeout:
	*timeout = 1;
	return 0;
}

/* abstraction functions for ia4320 fsk receiver ******************************/

/* read the status register of the fsk receiver */
static u16 rc_tomtom_status_fsk(int *timeout)
{
	u16 status;

	IO_Activate(FSK_DATA_FFS);

	IO_Deactivate(FSK_EN);
	status = rc_tomtom_txrx_spi(0x00, timeout) << 8;
	if(*timeout)
		return 0;
	status |= rc_tomtom_txrx_spi(0x00, timeout);
	if(*timeout)
		return 0;
	IO_Activate(FSK_EN);

	return status;
}

/* send a 16 bit command to the fsk receiver */
static int rc_tomtom_cmd_fsk(u16 value)
{
	int timeout;

	IO_Activate(FSK_DATA_FFS);

	IO_Deactivate(FSK_EN);
	rc_tomtom_txrx_spi(value >> 8, &timeout);
	if(timeout)
		return 1;
	rc_tomtom_txrx_spi(value & 0xff, &timeout);
	if(timeout)
		return 1;
	IO_Activate(FSK_EN);
	return 0;
}

/* read 8 bits from the fifo */
static u8 rc_tomtom_readfifo_fsk(int *timeout)
{
	u8 value;

	IO_Deactivate(FSK_DATA_FFS);

	IO_Deactivate(FSK_EN);
	value = rc_tomtom_txrx_spi(0xff, timeout);
	if(*timeout)
		return 0x00;
	IO_Activate(FSK_EN);

	return value;
}

/* init the fsk receiver */
static int rc_tomtom_init_fsk(void)
{
	int timeout;
	u16 status;
	int i;
	u8 fifo;

	IO_Activate(FSK_EN);
	msleep(1);

	status = rc_tomtom_status_fsk(&timeout);
	if(timeout)
		return 1;

	printk(KERN_INFO PFX "FSK STATUS: ");
	for(i = 0; i <= 15; i++)
		printk("%c", (status & (1 << (15 - i))) ? '1' : '0');
	printk("\n");

	printk(KERN_INFO PFX "FSK FIFO  : ");
	for(i = 0; i <= 7; i++) {
		fifo = rc_tomtom_readfifo_fsk(&timeout);
		if(timeout)
			return 1;
		printk("%02x ", fifo);
	}
	printk("\n");
	return 0;
}

/* driver core ****************************************************************/

static irqreturn_t rc_tomtom_interrupt(int irq, void *dev_id,
	struct pt_regs *fp)
{
/*
	input_report_key(&rc.rc_dev, KEY_POWER, IO_GetInput(ON_OFF));
	input_sync(&rc.rc_dev);
*/
	return IRQ_HANDLED;
}

static int rc_tomtom_probe(struct device *dev)
{
	struct clk *clk;
	unsigned long pclk_rate;
	struct platform_device *pdev;
	struct resource *mem;

	clk = clk_get(dev, "pclk");
	if (clk == NULL) {
		PK_ERR("failed get read PCLK rate\n");
		goto fail;
	}
	pclk_rate = clk_get_rate(clk);
	clk_put(clk);

	pdev = to_platform_device(dev);

	/* allocate remote control irq handler */
	rc.irq = platform_get_irq(pdev, 0);
	if (rc.irq == 0) {
		PK_ERR("failed to get interrupt resource\n");
		goto fail;
	}
	PK_DBG("remote control button using irq:%d\n", rc.irq);
	IO_SetInterrupt(ON_OFF);
	set_irq_type(rc.irq, IRQT_BOTHEDGE);
	if (request_irq(rc.irq, rc_tomtom_interrupt, 0, "tomtomgo-rc",
		dev)) {
		PK_ERR("can't allocate irq %d\n", rc.irq);
		goto fail;
	}

	/* allocate spi controller io region */
	mem = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (mem == NULL) {
		PK_ERR("failed to get io region resource\n");
		goto fail_mem;
	}
	rc.spibase = ioremap(mem->start, RESSIZE(mem));
	if (rc.spibase == NULL) {
		PK_ERR("failed to remap io region\n");
		goto fail_mem;
	}

	/* init spi controller */
	if (rc_tomtom_init_spi(dev, pclk_rate)) {
		PK_ERR("failed to init spi controller\n");
		goto fail_spi;
	}
	if (rc_tomtom_init_fsk()) {
		PK_ERR("failed to init fsk receiver\n");
		goto fail_fsk;
	}
/*
	memset(&rc.rc_dev, 0x00, sizeof(rc.rc_dev));
	rc.rc_dev.evbit[0] = BIT(EV_KEY);
	rc.rc_dev.keybit[LONG(KEY_POWER)] = BIT(KEY_POWER);
	input_register_device(&rc.rc_dev);
*/
	return 0;

fail_fsk:
	rc_tomtom_done_spi();
fail_spi:
	iounmap(rc.spibase);
fail_mem:
	free_irq(rc.irq, dev);
fail:
	PK_ERR("failed to load driver\n");
	return -EINVAL;
}

static int rc_tomtom_remove(struct device *dev)
{
	rc_tomtom_done_spi();
	iounmap(rc.spibase);
	free_irq(rc.irq, dev);
	return 0;
}

static void rc_tomtom_shutdown(struct device *dev)
{
	PK_DBG("Shutting down\n");
	/* Nothing yet */
}

#ifdef CONFIG_PM

static int rc_tomtom_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("dev = %p, state = %u, level = %u\n", dev, state, level);
	if (state == SUSPEND_POWER_DOWN)
		rc_tomtom_shutdown(dev);
	return 0;
}

static int rc_tomtom_resume(struct device *dev, u32 level)
{
	PK_DBG("dev = %p, level = %u\n", dev, level);
	return 0;
}

#endif /* CONFIG_PM */

static char rc_tomtom_name[] = "tomtomgo-rc";

static struct device_driver rc_tomtom_driver =
{
	.name		= rc_tomtom_name,
	.bus		= &platform_bus_type,
	.probe		= rc_tomtom_probe,
	.remove		= rc_tomtom_remove,
	.shutdown	= rc_tomtom_shutdown,
#ifdef CONFIG_PM
	.suspend	= rc_tomtom_suspend,
	.resume		= rc_tomtom_resume,
#endif /* CONFIG_PM */
};

static struct resource rc_tomtom_resource[] = {
	[0] = {
		.start = S3C2410_PA_SPI,
		.end   = S3C2410_PA_SPI + 0x14,
		.flags = IORESOURCE_MEM,
	},
	[1] = {
		.start = TOMTOMGO_IRQ_RC,
		.end   = TOMTOMGO_IRQ_RC,
		.flags = IORESOURCE_IRQ,
	},
};

static struct platform_device *rc_tomtom_device;

static int __init rc_tomtom_init(void)
{
	int ret;

	printk(KERN_INFO "TomTom GO Remote Control Driver -- cd05\n");
	PK_DBG("Registering driver\n");
	ret = driver_register(&rc_tomtom_driver);
	if (ret) {
		PK_ERR("Unable to register driver (%d)\n", ret);
		return ret;
	}
	PK_DBG("Registering platform device\n");
	rc_tomtom_device = platform_device_register_simple(rc_tomtom_name, -1, rc_tomtom_resource, ARRAY_SIZE(rc_tomtom_resource));
	if (IS_ERR(rc_tomtom_device)) {
		ret = PTR_ERR(rc_tomtom_device);
		rc_tomtom_device = NULL;
		PK_ERR("Unable to register platform device (%d)\n", ret);
		driver_unregister(&rc_tomtom_driver);
		return ret;
	}
	PK_DBG("Done\n");
	return 0;
}

static void __exit rc_tomtom_exit(void)
{
	if (rc_tomtom_device != NULL) {
		PK_DBG("Unregistering platform device\n");
		platform_device_unregister(rc_tomtom_device);
	}
	PK_DBG("Unregistering driver\n");
	driver_unregister(&rc_tomtom_driver);
	PK_DBG("Done\n");
}

module_init(rc_tomtom_init);
module_exit(rc_tomtom_exit);

MODULE_AUTHOR("Christian Daniel <cd@cdaniel.de>, Dimitry Andric <dimitry.andric@tomtom.com>");
MODULE_DESCRIPTION("TomTom GO Remote Control Driver");
MODULE_LICENSE("GPL");
