#include <linux/kernel.h>

#include <asm/errno.h>
#include <asm/io.h>
#include <asm/arch/regs-iis.h>
#include <asm/hardware/clock.h>

#include <barcelona/gotype.h>

#include "driver.h"
#include "iis.h"

struct iis_prescaler_info {
        const unsigned long hz;
        const unsigned long cdclk_goal;
        unsigned int div;
	unsigned long res;
};

struct iis_regs {
	unsigned long iismod;
	unsigned long iisfcon;
	unsigned long iispsr;
	unsigned long iiscon;
};

static struct clk *iis_clock;
static struct iis_regs iis_regs;
static struct iis_prescaler_info *current_samplerate;
static unsigned int current_samplesize;

/* The iis_prescaler_info values comes from the s3c24xx iis manual. */
#ifdef USE_FS256
static struct iis_prescaler_info fs256[] = {
        { .hz = 8000, .cdclk_goal = 2048000 },
        { .hz = 11025, .cdclk_goal = 2822400 },
        { .hz = 16000, .cdclk_goal = 4096000 },
        { .hz = 22050, .cdclk_goal = 5644800 },
        { .hz = 32000, .cdclk_goal = 8192000 },
        { .hz = 44100, .cdclk_goal = 11289600 },
        { .hz = 48000, .cdclk_goal = 12288000 },
        { .hz = 64000, .cdclk_goal = 16384000 },
        { .hz = 88200, .cdclk_goal = 22579200 },
        { .hz = 96000, .cdclk_goal = 24576000 },
};
#else /* USE_FS256 */
static struct iis_prescaler_info fs384[] = {
        { .hz = 8000, .cdclk_goal = 3072000 },
        { .hz = 11025, .cdclk_goal = 4233600 },
        { .hz = 16000, .cdclk_goal = 6144000 },
        { .hz = 22050, .cdclk_goal = 8467200 },
        { .hz = 32000, .cdclk_goal = 12288000 },
        { .hz = 44100, .cdclk_goal = 16934400 },
        { .hz = 48000, .cdclk_goal = 18432000 },
        { .hz = 64000, .cdclk_goal = 24576000 },
        { .hz = 88200, .cdclk_goal = 33868800 },
        { .hz = 96000, .cdclk_goal = 36864000 },
};
#endif /* USE_FS256 */


static void calc_div(struct iis_prescaler_info *fs, unsigned int size, unsigned long current_clock)
{
	unsigned int i;

	trace();

	for (i = 0; i < size; i++) {
		unsigned int j;
		unsigned long diff;
		unsigned long res;
		unsigned long min = current_clock;
		for (j = 0; j < 32; j++) {
			res = current_clock/(j+1);
			diff = abs(fs[i].cdclk_goal - res);
			if (diff < min) {
				fs[i].res = res;
				min = diff;
				fs[i].div = j;
			}
		}
		printk("CoolSound: Primary buffers now on %ld Hz.\n"
		       "           CDCLK real:%ld Hz, CDCLK wanted:%ld Hz.\n", 
		       fs[i].hz, fs[i].res, fs[i].cdclk_goal);
	}
}

void iis_showbus(const char *message) {
	unsigned long mod,fcon,psr,con;

	mod 	= readl(S3C2410_IISMOD);
	fcon 	= readl(S3C2410_IISFCON);
	psr 	= readl(S3C2410_IISPSR);
	con 	= readl(S3C2410_IISCON);


	printk("(%s) iismod: %lx, iisfcon: %lx, iiscon: %lx, iispsr: %lx\n",
	       message, mod, fcon, con, psr);
}

void iis_suspend(void)
{
	iis_regs.iismod = readl(S3C2410_IISMOD);
	iis_regs.iisfcon = readl(S3C2410_IISFCON);
	iis_regs.iispsr = readl(S3C2410_IISPSR);
	iis_regs.iiscon = readl(S3C2410_IISCON);
}

void iis_resume(void)
{
	writel(iis_regs.iismod, S3C2410_IISMOD);
	writel(iis_regs.iisfcon, S3C2410_IISFCON);
	writel(iis_regs.iispsr, S3C2410_IISPSR);
	writel(iis_regs.iiscon, S3C2410_IISCON);
}

void iis_control_cmd(iis_cmd_t command)
{
	unsigned int iiscon, iisfcon;

	switch (command) {
	case IIS_IDLE_TX:
		IO_Deactivate(I2SSDO);
		break;

	case IIS_ACTIVE_TX:
		IO_SetFunction(I2SSDO);
		break;

	case IIS_ALIGN_TX:
		if(readl(S3C2410_IISCON) & S3C2410_IISCON_LRINDEX) {
			iisfcon = readl(S3C2410_IISFCON);
			writel(iisfcon & (~S3C2410_IISFCON_TXDMA), S3C2410_IISFCON);
			writew(0, S3C2410_IISFIFO);
			writel(iisfcon, S3C2410_IISFCON);
		}
		break;
		

	case IIS_FLUSH_TX:
		iisfcon = readl(S3C2410_IISFCON);
		iisfcon &= ~S3C2410_IISFCON_TXENABLE;
		writel(iisfcon, S3C2410_IISFCON);
		iisfcon |=  S3C2410_IISFCON_TXENABLE;
		writel(iisfcon, S3C2410_IISFCON);
		break;
		
	case IIS_FLUSH_RX:
		iisfcon = readl(S3C2410_IISFCON);
		iisfcon &= ~S3C2410_IISFCON_RXENABLE;
		writel(iisfcon, S3C2410_IISFCON);
		iisfcon |=  S3C2410_IISFCON_RXENABLE;
		writel(iisfcon, S3C2410_IISFCON);
		break;
				
	case IIS_START:
		iiscon = readl(S3C2410_IISCON);
		iiscon |= S3C2410_IISCON_IISEN;
		writel(iiscon, S3C2410_IISCON);
		break;

	case IIS_STOP:
		iiscon = readl(S3C2410_IISCON);
		iiscon &= ~S3C2410_IISCON_IISEN;
		writel(iiscon, S3C2410_IISCON);
		break;

	default:
		break;

	}
	
}

int iis_set_samplerate(unsigned int rate)
{
	unsigned int i;

#ifdef USE_FS256
	/* Currently only uses 256fs. */
	for (i = 0; i < sizeof fs256 / sizeof *fs256; i++) {
		if (fs256[i].hz == rate) {
			current_samplerate = &(fs256[i]);
			writel((fs256[i].div << 5) + fs256[i].div, S3C2410_IISPSR);
			return 0;
		}
	}
#else /* USE_FS256 */
	/* Currently only uses 384fs. */
	for (i = 0; i < sizeof fs384 / sizeof *fs384; i++) {
		if (fs384[i].hz == rate) {
			current_samplerate = &(fs384[i]);
			writel((fs384[i].div << 5) + fs384[i].div, S3C2410_IISPSR);
			return 0;
		}
	}
#endif /* USE_FS256 */

   return -EINVAL;
}

unsigned int iis_get_samplerate(void)
{
	return current_samplerate->hz;
}

int iis_set_samplesize(unsigned int size)
{
	unsigned long iismod = readl(S3C2410_IISMOD);

	if (size == 8) iismod &= ~S3C2410_IISMOD_16BIT;
	else if (size == 16) iismod |= S3C2410_IISMOD_16BIT;
	else return -EINVAL;

	current_samplesize = size;
	writel(iismod, S3C2410_IISMOD);
	iis_showbus("after setsmp");	
	return 0;
}

unsigned int iis_get_samplesize(void)
{
	return current_samplesize;
}

int iis_init(void)
{
	unsigned long iismod, iisfcon, iiscon;

 	IO_Deactivate(I2SSDO);
	IO_SetFunction(CDCLK);
	IO_SetFunction(I2SSCLK);
	IO_SetFunction(I2SLRCK);
	IO_SetFunction(I2SSDI);

	if ((iis_clock = clk_get(NULL, "iis")) == NULL)
		return -ENOENT;

	clk_use(iis_clock);
	clk_enable(iis_clock);

	printk("Current clock rate: %ld\n", clk_get_rate(iis_clock));
#ifdef USE_FS256
	calc_div(fs256, sizeof fs256 / sizeof *fs256, clk_get_rate(iis_clock));
#else /* USE_FS256 */
	calc_div(fs384, sizeof fs384 / sizeof *fs384, clk_get_rate(iis_clock));
#endif /* USE_FS256 */

	iiscon = S3C2410_IISCON_PSCEN |
	         S3C2410_IISCON_IISEN |
	         S3C2410_IISCON_TXDMAEN |
	         S3C2410_IISCON_RXDMAEN;

        iisfcon = S3C2410_IISFCON_RXDMA |
	          S3C2410_IISFCON_TXDMA |
	          S3C2410_IISFCON_TXENABLE |
	          S3C2410_IISFCON_RXENABLE;

        iismod = S3C2410_IISMOD_LR_RLOW |
	         S3C2410_IISMOD_IIS |
#ifdef USE_FS256
	         S3C2410_IISMOD_256FS |
#else /* USE_FS256 */
	         S3C2410_IISMOD_384FS |
#endif /* USE_FS256 */
	         S3C2410_IISMOD_32FS |
		 S3C2410_IISMOD_TXRXMODE;
		 
//	dump_iis_bus("Test1");
	writel(iismod, S3C2410_IISMOD);
	writel(iisfcon, S3C2410_IISFCON);
	writel(iiscon, S3C2410_IISCON);
//	dump_iis_bus("Test2");

	iis_set_samplerate(22050);
	iis_set_samplesize(16);
	return 0;
}
