#include <linux/delay.h>
#include "driver.h"
#include "spi.h"
#include "wm8971.h"

/* Keep all the register values in an array.
 * If a value is zero don't ever set it...
 */
static uint16_t wm8971[43] = {
0,		/* R0 Left input vol, don't touch... */
0x100 | 64,	/* R1 Right input vol, unmute, zerocross + VOL (0 - 63). Both needs to be updated. */
0x00  | 121,	/* R2 LOUT1 vol, update R3, zerocross + VOL (0 - 127). */
0x100 | 121,	/* R3 ROUT1 vol, don't touch, updated by R2. */
0,		/* R4 Reserved, don't touch. */
0,		/* R5 ADC and DAC Control, XXX - Needs to be zero to unmute... how?? */
0,		/* R6 Reserved, don't touch. */
0x2,		/* R7 Audio Interface, IIS Mode. */
#ifdef USE_FS256
0x20,		/* R8 Sample rate */
#else /* USE_FS256 */
0x22,		/* R8 Sample rate */
#endif /* USE_FS256 */
0,		/* R9 Reserverd, don't touch. */
0,		/* R10 LDAC vol, don't touch. */
0,		/* R11 RDAC vol, don't touch. */
0xf,		/* R12 Bass control, don't touch for now. XXX - Add for acoustics. */
0xf,		/* R13 Treble control, don't touch for now. XXX - Add for acoustics. */
0,		/* R14 Don't mentioned in the spec, don't touch. */
0,		/* R15 Reset, only write zero to this reg. */
0,		/* R16 Reserved, don't touch. */
0,		/* R17 ALC control... set to 0. XXX - need that later for record*/
0,		/* R18 AlC control... set to 0. XXX - ... */
0,		/* R19 ALC control... set to 0. XXX - ... */
0,		/* R20 Noise gate, XXX ... */
0x1FF,		/* R21 Left ADC control, don't touch... not used. */
0x1FF,		/* R22 Right ADC control, keep default value. */
0,		/* R23 Additional control 1, XXX - unused for now. Could enable mono mixing for power 			saving and some other things.... should be researched. */
0,		/* R24 Additional control 2, don't touch. */
0xd6,		/* R25 Power mgmt 1, 50kOhm enabled, VREF, right PGA and ADC, MICBIAS. */
0x1e4,		/* R26 Power mgmt 2, left DAC and OUT1, right DAC and OUT2, MONO out. */
0,		/* R27 Additional control 3, don't touch. */
0,		/* R28 Don't mentioned. */
0,		/* R29 Don't mentioned. */
0,		/* R30 Don't mentioned. */
0x11,		/* R31 ADC input mode, Digital MONO mix. */
0x0,		/* R32 Left ADC signal path, don't touch. */
0x70,		/* R33 Right ADC signal path, MIC and some boost. */
0x000,		/* R34 Left out mix 1, Left DAC. */
0x100,		/* R35 Left out mix 2, don't touch. */
0x100,		/* R36 Right out mix 1, don't touch. */
0x000,		/* R37 Right out mix 2, Right DAC. */
0x100,		/* R38 Mono out mix, Left Dac. */
0x100,		/* R39 Mono out mix, Right Dac. */
0x0,		/* R40 LOUT2 volume, unsued, don't touch. */
0x0,		/* R41 ROUT2 volume, unused, don't touch. */
0x80 | 127	/* R42 MONO volume, zero cross + vol (0 - 127). */
};

static inline void WM8971_WRREG(uint16_t reg, uint16_t value)
{
	wm8971[reg] = value;
	D(printk("R%d=0x%x\n", reg, value));
	spi_write_register(reg, value);
}

static inline uint16_t WM8971_RDREG(uint16_t reg)
{
	return wm8971[reg];
}

static void wm8971_init_registers(void)
{
	int i;

	WM8971_WRREG(15, 0);
	for (i = 0; i < sizeof wm8971 / sizeof *wm8971; i++)
		if (WM8971_RDREG(i)) WM8971_WRREG(i, WM8971_RDREG(i)); 

	/* Unmute. */
	WM8971_WRREG(5, 0); 
}

static int wm8971_set_volume_internal(unsigned int volume)
{
	if (volume > 127) volume = 127;

	WM8971_WRREG(42, 0x80 | volume);

	return 0;
}

static int wm8971_set_volume_external(unsigned int volume)
{
	if (volume > 127) volume = 127;

	WM8971_WRREG(2, 0x180 | volume);
	WM8971_WRREG(3, 0x180 | volume);

	return 0; 
}

static int wm8971_set_volume(unsigned int volume)
{
	wm8971_set_volume_internal(volume);
	wm8971_set_volume_external(volume);

	return 0; 
}

static int wm8971_set_treble(unsigned int treble)
{
	WM8971_WRREG(13, treble);

	return 0; 
}

static int wm8971_set_bass(unsigned int bass)
{
	WM8971_WRREG(12, bass ^ 0x80);

	return 0; 
}

static int wm8971_set_register(unsigned int data)
{
	unsigned int reg = data >> 9;
	unsigned int value = (0x1ff & data);

	printk("SetRegister: %u => 0x%03x\n",reg,value);
	WM8971_WRREG(reg, value);
	return 0;
}

static void wm8971_powerdown(void)
{
	spi_write_register(5, 8);
	mdelay(50);
	spi_write_register(25, 0);
	spi_write_register(26, 0);
	mdelay(50);
}

void wm8971_init(struct iis_attached_ops *ops)
{
	ops->set_volume = wm8971_set_volume;
	ops->set_volume_internal = wm8971_set_volume_internal;
	ops->set_volume_external = wm8971_set_volume_external;
	ops->set_register = wm8971_set_register;
	ops->resume = wm8971_init_registers;
	ops->suspend = wm8971_powerdown;
	ops->set_treble = wm8971_set_treble;
	ops->set_bass = wm8971_set_bass;

	wm8971_init_registers();
}

