/* drivers/block/ttg_sd/sd_block.c
*
* SD card block driver.
*
* Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
* Author: Koen Martens <kmartens@sonologic.nl>
* Author: Jeroen taverne <jeroen.taverne@tomtom.com>
* Author: Dimitry Andric <dimitry.andric@tomtom.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*/

#include <linux/completion.h>
#include <linux/mm.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/dma-mapping.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/hdreg.h>
#include <linux/device.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/hardware/clock.h>
#include <linux/vmalloc.h>
#include <linux/list.h>
#include <linux/blkdev.h>
#include <linux/blkpg.h>
#include <linux/types.h>
#include <linux/ioctl.h>
#include <linux/genhd.h>
#include <linux/moduleparam.h>
#include <barcelona/gotype.h>
#include <barcelona/debug.h>
#include <asm/irq.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/tomtomgo-map.h>
#include "sd.h"
#include "sd_card.h"

/* Defines */
#define PFX "sd: "
#define PK_DBG PK_DBG_FUNC
#define XFER_DBG PK_DBG_NONE

char *param_devs;	// module parameter

#define MAJOR_NR	250
#define SD_RAHEAD	16		// number of sectors to read ahead

static int sd_open (struct inode *, struct file *);
static int sd_release (struct inode *, struct file *);
static int sd_ioctl (struct inode *inode, struct file *file,unsigned int, unsigned long);
static int sd_check_media_change(struct gendisk *disk);
static int sd_revalidate(struct gendisk *disk);

static void sd_tasklet(unsigned long data);
DECLARE_TASKLET(sd_tasklet_handle, sd_tasklet, 0L);

static struct block_device_operations sd_bd_ops =
{
	.owner           = THIS_MODULE,
	.open            = sd_open,
	.release         = sd_release,
	.ioctl           = sd_ioctl,
	.media_changed   = sd_check_media_change,
	.revalidate_disk = sd_revalidate,
};

// internal driver structure
typedef struct SD_DEVICE
{
	int usage;
	unsigned long maxsectors;
	spinlock_t lock;
	int nrahead;			// number of sectors to read ahead
	char *rahead;		// buffer for read ahead sectors
	unsigned long raheadno;	// number of first sector in rahead
	struct request_queue *queue;
	struct gendisk *disk;
	int media_changed;
} SD_DEVICE;

SD_DEVICE sd_device;

// hotswapping notify
#define SD_MAX_NOTIFY	16

struct notify_struct
{
	int in_use;
	int device;
	pid_t pid;
};

static struct notify_struct notify[SD_MAX_NOTIFY];

// functions

// called from sd_card.c on EINT2
static int sd_initialize(void);

void sd_hotswap_handler(int device)
{
	int i;

	PK_DBG("sd_hotswap_handler(%i)\n",device);

	sd_device.media_changed = 1;
	sd_device.raheadno = -1;

	if (SDCARD_InSlot())
	{
		// card inserted
		PK_DBG("card was inserted, reinitializing\n");
		sd_initialize();
	}

	for(i=0;i<SD_MAX_NOTIFY;i++)
	{
		if(notify[i].device==device && notify[i].in_use)
		{
			PK_DBG("sending notification (SIGIO) to pid %i\n",notify[i].pid);
			if(kill_proc(notify[i].pid,SIGIO,1) != 0)
			{
				// process has died, free notify entry
				notify[i].in_use=0;
			}
		}
	}
}

/*
* Satisfy request from read ahead buffer if possible
* returns: 1 if satisfied, 0 if physical transfer needed
*/
int sd_read_ahead(unsigned long sector,unsigned long nsect, char *buffer)
{
	if(sd_device.raheadno!=-1)
	{
		if( (sector>=sd_device.raheadno) && ((sector+nsect)<=(sd_device.raheadno+sd_device.nrahead)) )
		{
			memcpy(buffer,sd_device.rahead+((sector-sd_device.raheadno)*512),512*nsect);
			return 1;
		}
	}
	return 0;
}

/*
* Handle an I/O request, if this is a read request, tries to satisfy the
* request from the read ahead buffer. If it is a write request or it couldn't
* be satisfied from read ahead, it will call the proper physical I/O function.
*
* Returns 1 if succesfull, 0 when the request failed.
*/
static int sd_transfer(unsigned long sector, unsigned long nsect, char *buffer, int write)
{
	int ec;

	// out of bounds check
	if ((sector + nsect) > sd_device.maxsectors)
	{
		PK_WARN("Beyond-end transfer (sector=%ld, nsect=%ld, max=%ld)\n", sector, nsect, sd_device.maxsectors);
		return 0;
	}

	// if request satisfied from read ahead buffer, we're done
	if( !write && sd_read_ahead(sector,nsect,buffer))
	{
		XFER_DBG("fromra, sector=%li, nsect=%li, write=%i, buffer=0x%x\n",sector,nsect,write,(unsigned int)buffer);
		return 1;
	}

	if(write)
	{
		XFER_DBG("Before sdcardDataTransfer(WRITE,\n");
		ec=SDCARD_WriteSectors(buffer,sector,nsect);
		XFER_DBG("After sdcardDataTransfer(WRITE,\n");
	}
	else if( (nsect<=sd_device.nrahead) && (((sector+nsect))<sd_device.maxsectors) )
	{
		// read ahead
		ec=SDCARD_ReadSectors(sd_device.rahead,sector,sd_device.nrahead);
		if(ec)
		{
			sd_device.raheadno=sector;
			ec=sd_read_ahead(sector,nsect,buffer);
		}
		else
		{
			sd_device.raheadno=-1;
		}
	}
	else
	{
		// physical read
		ec=SDCARD_ReadSectors(buffer,sector,nsect);
	}
	return (ec?1:0);
}

static request_queue_t *q2 = NULL;

static void sd_tasklet(unsigned long data)
{
	static struct request *req;
	static int ec;
	
	if (q2)
	{
		if ((req = elv_next_request(q2)) != NULL)
		{
			if (! blk_fs_request(req))
			{
				// ignore non-block I/O requests
				end_request(req, 0);
			}
			else
			{
				// perform the transfer and signal result to kernel
				ec=sd_transfer(req->sector, req->current_nr_sectors, req->buffer, rq_data_dir(req));
				end_request(req, ec);
			}
			tasklet_schedule(&sd_tasklet_handle);
		} else q2 = NULL;
	}
}

/*
* Handle device requests, called by kernel
*/
static void sd_request(request_queue_t *q)
{
	q2 = q;
	tasklet_schedule(&sd_tasklet_handle);
}

/*
* Initialize sd card device
*/
static int sd_initialize(void)
{
	if(SDCARD_InitDisk())
	{
		sd_device.maxsectors = SDCARD_GetAmountSectors();
		printk(KERN_INFO "SDx: sd card of %lu sectors (%lu MiB)\n", sd_device.maxsectors, (sd_device.maxsectors + 1024LU) / 2048LU);
		return 0;
	}
	else
	{
		sd_device.maxsectors=0;
		printk(KERN_WARNING "SDx: sd card initialisation failed\n");
		return -ENODEV;
	}
}

/*
* Dummy
*/
static int sd_open (struct inode *inode, struct file *file)
{
	sd_device.usage++;

	return 0;          /* success */
}

/*
* Dummy
*/
static int sd_release(struct inode *inode, struct file *file)
{
	sd_device.usage--;

	return 0;
}

/*
* Revalidate, ie. just initialize again, probably there's more to it than this
*/
static int sd_revalidate(struct gendisk *disk)
{
	sd_initialize();

	set_capacity(disk, sd_device.maxsectors);

	sd_device.media_changed=0;
	sd_device.raheadno=-1;

	return 0;
}

/*
* Well..
*/
static int sd_check_media_change(struct gendisk *disk)
{
	return sd_device.media_changed;
}

/*
* Ioctl's
*
* Only implements SDHOT_REG, SDHOT_UNREG, IOR_SD_STATUS and HDIO_GETGEO,
* the rest is handled by the block device layer
*/
static int sd_ioctl (struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	long size;
	struct hd_geometry geo;
	pid_t apid;
	u32 u32status;
	char sdcard_id[SDCARD_ID_LEN];
	int i;

	switch(cmd)
	{
		case SDHOT_REG:
		// registers given pid in the 'notify me on media change' array
		for(i=0;i<SD_MAX_NOTIFY;i++)
		{
			if(!notify[i].in_use)
			{
				notify[i].in_use=1;
				notify[i].device=0;
				copy_from_user((void *)&notify[i].pid, (void *)arg, sizeof(pid_t));
				// TODO check copy_from_user error code
				return 0;
			}
		}
		return -ENOMEM;

		case SDHOT_UNREG:
		// unregisters given pid in the 'notify me on media change' array
		copy_from_user((void *)&apid, (void *)arg, sizeof(pid_t));
		for(i=0;i<SD_MAX_NOTIFY;i++)
		{
			if(notify[i].pid==apid && notify[i].in_use)
			{
				notify[i].in_use=0;
			}
		}
		return 0;

		case IOR_SD_STATUS:
		/* return whether SD card is inserted/removed */
		{
			if (SDCARD_InSlot())
			u32status = SD_CARD_INSERTED;
			else
			u32status = SD_CARD_REMOVED;
		}
		copy_to_user((void *)arg,&u32status,sizeof(u32));
		return 0;

		case GET_SD_SERIAL_NUMBER:
		if(SDCARD_GetSerial(sdcard_id)!=16)
		return -EIO;
		if(copy_to_user((void *)arg,sdcard_id,SDCARD_ID_LEN))
		return -EFAULT;
		return 0;

		case HDIO_GETGEO:
		// make up something reasonable
		size = sd_device.maxsectors;
		geo.cylinders = (size & ~0x3f) >> 6;
		geo.heads = 4;
		geo.sectors = 16;
		geo.start = 4;
		if (copy_to_user((void *) arg, &geo, sizeof(geo)))
		return -EFAULT;
		return 0;
	}

	return -EINVAL; /* unknown command */
}

/*
* Probes for devices and initializes
*/
static int sd_probe(struct device *dev)
{
	int ret;
	int ec=0;

	PK_DBG("initializing!\n");

	// register block device
	ret=register_blkdev(MAJOR_NR, "sdcard");
	if (ret<0)
	{
		PK_WARN("unable to get major number\n");
		ec=ret;
		goto out;
	}

	SDCARD_Init();
	ret=sd_initialize();
	if (ret<0)
	{
		PK_WARN("unable to initialize sd\n");
		ec=ret;
		goto out;
	}

	sd_device.usage=0;
	sd_device.nrahead=SD_RAHEAD;
	sd_device.rahead=NULL;
	sd_device.raheadno=-1;
	spin_lock_init(&sd_device.lock);
	//			SDCARD_HotswapInit(sd_hotswap_handler,0);

	// if initialization succesfull, fill internal & kernel structs
	// allocate read ahead buffer
	sd_device.rahead=kmalloc(sd_device.nrahead * 512,GFP_KERNEL);
	if(sd_device.rahead==NULL)
	{
		ec=-ENOMEM;
		goto memout;
	}

	// make a gendisk entry
	sd_device.disk=alloc_disk(16);
	if(!sd_device.disk)
	{
		ec=-ENOMEM;
		goto memout;
	}
	// fill the gendisk entry
	sd_device.disk->major=MAJOR_NR;
	sd_device.disk->first_minor=0;
	sd_device.disk->fops=&sd_bd_ops;
	sd_device.disk->private_data=&sd_device;
	sprintf(sd_device.disk->disk_name,"sdcard");
	set_capacity(sd_device.disk,sd_device.maxsectors);

	// create the request queue
	sd_device.queue = blk_init_queue(sd_request, &sd_device.lock);
	if (sd_device.queue == NULL)
	{
		ec=-ENOMEM;
		goto memout;
	}
	blk_queue_hardsect_size(sd_device.queue, 512); 	// not needed, is standard 512
	sd_device.queue->queuedata=&sd_device;
	sd_device.disk->queue = sd_device.queue;

	// all went well, add disks to the system
	add_disk(sd_device.disk);

out:
	PK_DBG("Probe return with ec=%i\n",ec);
	return ec;

memout:
	PK_DBG("Probe error, ec=%i\n",ec);
	if(sd_device.rahead) kfree(sd_device.rahead);
	return ec;
}

/*
* De-init everything
*/
static int sd_remove(struct device *dev)
{
	int hotswap=0;

	del_gendisk(sd_device.disk);
	put_disk(sd_device.disk);

	// de-init hotswapping if necesarry
	if(hotswap) SDCARD_HotswapUninit();

	// unregister the block device
	unregister_blkdev(MAJOR_NR,"sdcard");

	// and clean up the request queues
	blk_cleanup_queue(sd_device.queue);

	return 0;
}

static void sd_powerup(struct device *dev)
{
	PK_DBG("Powering up\n");
	SDCARD_InitDisk();
	PK_DBG("Powerup complete\n");
}

static void sd_shutdown(struct device *dev)
{
	PK_DBG("Shutting down\n");
	SDCARD_Off();
	PK_DBG("Shutdown complete\n");
}

#ifdef CONFIG_PM

static int sd_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("dev = %p, state = %u, level = %u\n", dev, state, level);
	if (level == SUSPEND_POWER_DOWN)
	{
		sd_shutdown(dev);
	}
	return 0;
}

static int sd_resume(struct device *dev, u32 level)
{
	PK_DBG("dev = %p, level = %u\n", dev, level);
	if (level == RESUME_POWER_ON) {
		sd_powerup(dev);
	}
	return 0;
}

#else /* CONFIG_PM */
#define sd_suspend NULL
#define sd_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver sd_driver =
{
	.name           = "tomtomgo-sd",
	.bus            = &platform_bus_type,
	.probe          = sd_probe,
	.remove         = sd_remove,
	.shutdown       = sd_shutdown,
	.suspend        = sd_suspend,
	.resume         = sd_resume,
};

static int __init sd_mod_init(void)
{
	int ret;

	if (!IO_HaveSdCard()) {
		PK_DBG("No SD card in this model.\n");
		return -ENODEV;
	}

	printk(KERN_INFO "TomTom GO Storage Device Driver, (C) 2004,2005 TomTom BV\n");

	PK_DBG("Registering driver\n");
	ret = driver_register(&sd_driver);
	if (ret)
	{
		PK_ERR("Unable to register driver (%d)\n", ret);
		return ret;
	}
	PK_DBG("Done\n");
	return 0;
}

static void __exit sd_mod_exit(void)
{
	if (!IO_HaveSdCard())
		return;

	PK_DBG("Unregistering driver\n");
	driver_unregister(&sd_driver);
	PK_DBG("Done\n");
}

module_init(sd_mod_init);
module_exit(sd_mod_exit);

module_param_named(devs,param_devs,charp,0);

MODULE_AUTHOR("Koen Martens <kmartens@sonologic.nl>, Jeroen taverne <jeroen.taverne@tomtom.com>, Dimitry Andric <dimitry.andric@tomtom.com>");
MODULE_DESCRIPTION("TomTom GO Storage Device Driver");
MODULE_LICENSE("GPL");

/* EOF */
