/*
 *  linux/drivers/ide/arm/tomtom.c - TomTomGo IDE bus glue
 *
 *  Copyright (C) 2004 TomTom BV
 *  Author: Thomas Kleffel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/*
 * Major TODOs:
 *   - proper disconnecting from generic ide driver
 *   - DMA (will not work on current hardware)
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/blkdev.h>
#include <linux/delay.h>
#include <linux/err.h>
#include <linux/ide.h>

#include <asm/hardware.h>
#include <asm/io.h>

#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-mem.h>
#include <asm/arch/tomtomgo-wake.h>

#include <barcelona/gotype.h>
#include <barcelona/debug.h>

#define PFX "ide-tomtom: "
#define PK_DBG PK_DBG_FUNC

#define RESSIZE(ressource) (((ressource)->end - (ressource)->start)+1)

// IDE interface
unsigned long idebase;
#define rIDEDATAREAD     (*(volatile unsigned short*)(idebase+(0*4)))
#define rIDEDATAWRITE    (*(volatile unsigned short*)(idebase+(0*4)))
#define rIDEERROR        (*(volatile unsigned char*) (idebase+(1*4)))
#define rIDEFEATURES     (*(volatile unsigned char*) (idebase+(1*4)))
#define rIDESECTORCOUNT  (*(volatile unsigned char*) (idebase+(2*4)))
#define rIDESECTORNUMBER (*(volatile unsigned char*) (idebase+(3*4)))
#define rIDECYLINDERLOW  (*(volatile unsigned char*) (idebase+(4*4)))
#define rIDECYLINDERHIGH (*(volatile unsigned char*) (idebase+(5*4)))
#define rIDEDEVICEHEAD   (*(volatile unsigned char*) (idebase+(6*4)))
#define rIDESTATUS       (*(volatile unsigned char*) (idebase+(7*4)))
#define rIDECOMMAND      (*(volatile unsigned char*) (idebase+(7*4)))

/* bits in Error register: */
#define BBK   0x0080
#define UNC   0x0040
#define MC    0x0020
#define IDNF  0x0010
#define MCR   0x0008
#define ABRT  0x0004
#define TK0NF 0x0002
#define AMNF  0x0001

/* bits in Status register: */
#define BSY   0x0080
#define DRDY  0x0040
#define DWF   0x0020
#define DSC   0x0010
#define DRQ   0x0008
#define CORR  0x0004
#define IDX   0x0002
#define ERR   0x0001

static void IDE_WaitForBusyCleared(void)
{
	/* check for BSY to be cleared */
	while ((rIDESTATUS & BSY) > 0)
	{
		/* do nothing */
	}
}

static void IDE_DoCommand(unsigned short cCommand)
{
	IDE_WaitForBusyCleared();
	rIDECOMMAND = cCommand;
	IDE_WaitForBusyCleared();
}

static void IDE_SpinDown(void)
{
	IDE_DoCommand(0xE0);
	mdelay(1000);
}

int ide_tomtom_getslot(hw_regs_t *hw, ide_hwif_t **hwifp) {
	ide_hwif_t *hwif;
	int index;

	for (index = 0; index < MAX_HWIFS; ++index) {
		hwif = &ide_hwifs[index];
		if (hwif->hw.io_ports[IDE_DATA_OFFSET] == hw->io_ports[IDE_DATA_OFFSET]) {
			if(hwifp) (*hwifp)=hwif;
			return index;
		}
	}

	for (index = 0; index < MAX_HWIFS; ++index) {
		hwif = &ide_hwifs[index];
		if (hwif->hold)
			continue;

		if (!hwif->hw.io_ports[IDE_DATA_OFFSET]) {
			if(hwifp) (*hwifp)=hwif;
			return index;
		}
	}

	if(hwifp) (*hwifp) = 0;
	return -1;
}

int ide_tomtom_register(hw_regs_t *hw, ide_hwif_t **hwifp) {
	ide_hwif_t *hwif;
	int index;

	index = ide_tomtom_getslot(hw, &hwif);
	if(!hwif) return -2;

	memcpy(&hwif->hw, hw, sizeof(*hw));
	memcpy(hwif->io_ports, hwif->hw.io_ports, sizeof(hwif->hw.io_ports));

	hwif->irq 	= hw->irq;
	hwif->noprobe 	= 0;
	hwif->chipset 	= hw->chipset;

	//We have no access to the CONTROL register and thus cannot do lba48!
	hwif->no_lba48	= 1;

	default_hwif_mmiops(hwif);

	probe_hwif_init_with_fixup(hwif, NULL);
	create_proc_ide_interfaces();

	if (hwifp) *hwifp = hwif;

	return (hwif->present) ? index : -1;
}

static void ide_tomtom_hw_init(void)
{
	__raw_writel((__raw_readl(S3C2410_BWSCON) & (~(7 << 12))) | (1 << 12) | (0 << 14), S3C2410_BWSCON);
	__raw_writel((0 << 4) | (1 << 6) | (4 << 8) | (1 << 11) | (0 << 13), S3C2410_BANKCON3);

	IO_SetInput(HDD_DRQ);
	IO_SetInput(HDD_DACK);
	IO_Activate(HDD_PWR_ON);
	IO_SetInterrupt(HDD_IRQ);
	IO_SetInput(HD_LED);
	IO_Activate(AT_RST);
	mdelay(50);
	IO_Deactivate(AT_RST);
	IO_Activate(HDD_BUF_EN);
}

static void ide_tomtom_hw_exit(void)
{
	IDE_SpinDown();

	IO_Deactivate(HDD_BUF_EN);     
	IO_Activate(AT_RST);         
	IO_Deactivate(HDD_PWR_ON);    
	IO_Deactivate(HDD_IRQ);        
	IO_Deactivate(HD_LED);        
	IO_Deactivate(HDD_DRQ);
	IO_Deactivate(HDD_DACK);
}

static int ide_tomtom_probe(struct device *dev)
{
	struct resource *mem;
	void __iomem *base;
	int irq;
	hw_regs_t ide_hw;
	ide_hwif_t *ide_hwif;

	int index, p;

	struct platform_device *pdev = to_platform_device(dev);
	int ret;

	mem = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!mem) {
		PK_ERR("failed to get io memory region resource.\n");
		ret = -ENOENT;
		goto probe_out;
	}

	mem = request_mem_region(mem->start, RESSIZE(mem), pdev->name);
	if (!mem) {
		PK_ERR("failed to request io memory region.\n");
		ret = -ENOENT;
		goto probe_out;
	}

	base = ioremap(mem->start, RESSIZE(mem));
	if (base == 0) {
		PK_ERR("failed to ioremap() io memory region.\n");
		ret = -EINVAL;
		goto probe_free_mem_region;
	}

	irq = platform_get_irq(pdev, 0);
	if (irq == 0) {
		PK_ERR("failed to get interrupt resource.\n");
		ret = -EINVAL;
		goto probe_iounmap;
	}

	//We don't need to wait for hd powerup as linux's ide code
	//will wait up to 35 seconds.

	PK_DBG("mapped mem:[%p]=>[%p] irq:%u\n", (void *) mem->start, base, irq);

	//prepare ide_hw structure
	memzero(&ide_hw, sizeof(ide_hw));

	for(p=0;p<8;p++) {
		ide_hw.io_ports[p] = (unsigned long) (base + (4*p));
	}
	
	// Also set base address for HDD temperature check
	tomtom_hddtemp_setaddress((unsigned long) base);
	idebase = (unsigned long) base;

	ide_hw.chipset	= ide_generic;
	ide_hw.dma	= NO_DMA;
	ide_hw.irq	= irq;

	// Init the hardware
	ide_tomtom_hw_init();

	set_irq_type(irq, IRQT_RISING);

	// Register device with IDE driver
	index = ide_tomtom_register(&ide_hw, &ide_hwif);
	if(index < 0) {
		PK_ERR("failed to register ide interface (%i).\n",index);
		ret = -EINVAL;
		goto probe_shutdown;
	}

	PK_DBG("initialisation done. Index:%i\n",index);
	return 0;

probe_shutdown:
	// Deinit hardware
	ide_tomtom_hw_exit();

probe_iounmap:
	iounmap(base);

probe_free_mem_region:
	release_mem_region(mem->start, RESSIZE(mem));

probe_out:
	return ret;
}

static int ide_tomtom_remove(struct device *dev)
{
	ide_tomtom_hw_exit();
	return 0;
}

static void ide_tomtom_shutdown(struct device *dev)
{
	PK_DBG("Shutting down\n");
	ide_tomtom_hw_exit();
}

#ifdef CONFIG_PM

static int ide_tomtom_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("dev = %p, state = %u, level = %u\n", dev, state, level);
	if(level == SUSPEND_POWER_DOWN) {
		ide_tomtom_hw_exit();
	}
	return 0;
}

static int ide_tomtom_resume(struct device *dev, u32 level)
{
	PK_DBG("dev = %p, level = %u\n", dev, level);
	if(level == RESUME_POWER_ON) {
		ide_tomtom_hw_init();
	}
	return 0;
}

#else /* CONFIG_PM */
#define ide_tomtom_suspend NULL
#define ide_tomtom_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver tomtomide_driver =
{
	.name           = "tomtomgo-ide",
	.bus            = &platform_bus_type,
	.probe          = ide_tomtom_probe,
	.remove         = ide_tomtom_remove,
	.shutdown       = ide_tomtom_shutdown,
	.suspend        = ide_tomtom_suspend,
	.resume         = ide_tomtom_resume,
};

static int __init ide_tomtom_init(void)
{
	int ret;

	printk(KERN_INFO "TomTom GO IDE Driver, (C) 2004,2005 TomTom BV\n");

	if (!IO_HaveHarddisk()) {
		PK_DBG("No harddisk in this model.\n");
		return -ENODEV;
	}

	PK_DBG("Registering driver\n");
	ret = driver_register(&tomtomide_driver);
	if (ret)
	{
		PK_ERR("Unable to register driver (%d)\n", ret);
		return ret;
	}
	PK_DBG("Done\n");
	return 0;
}

static void __exit ide_tomtom_exit(void)
{
	if (!IO_HaveHarddisk())
		return;

	PK_DBG("Unregistering driver\n");
	driver_unregister(&tomtomide_driver);
	PK_DBG("Done\n");
}

module_init(ide_tomtom_init);
module_exit(ide_tomtom_exit);

MODULE_DESCRIPTION("TomTom Go IDE bus glue");
MODULE_LICENSE("GPL");
