/* drivers/video/tomtomgo.c
 *
 * TomTom Go LCD Controller Frame Buffer Driver
 * based on skeletonfb.c, sa1100fb.c, tomtomgofb.c
 *
 * Copyright (c) 2005 TomTom BV
 * Author: Christian Daniel <cd@cdaniel.de>
 * Author: Thomas Kleffel <tk@maintech.de>
 * Author: Jeroen Taverne <jeroen.taverne@tomtom.com>
 * Author: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive for
 * more details.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/dma-mapping.h>

#include <asm/arch/regs-lcd.h>
#include <asm/hardware/clock.h>

#include <barcelona/Barc_Gpio.h>
#include <barcelona/gotype.h>
#include <barcelona/debug.h>

#include "tomtomgo.h"

#define PFX "tomtomgo-lcd: "
#define PK_DBG PK_DBG_FUNC

#define TTGL_XRES   240
#define TTGL_YRES   320
#define TTGL_XRESV  240
#define TTGL_YRESV  640
#define TTGL_BPP    16
#define TTGL_BYPP   2
#define TTGL_WIDTH  53
#define TTGL_HEIGHT 71

static struct tomtomgolcd_fb_slot fbs;

static struct resource *lcd_mem;
static struct clk *lcd_clock;
static void __iomem *lcd_base;

//IRQ not used yet
//static struct resource *lcd_irq;

static const struct tomtomgolcd_screen_info tomtomgo_screen_info = {
	.type = TFT_LCD,
	.bpp  = TFT_16BPP_565,

	.hsize= TTGL_XRESV,
	.hfpd = 8,
	.hbpd = 16,
	.hspw = 24,

	.vsize= TTGL_YRESV,
	.vfpd = 5,
	.vbpd = 5,
	.vspw = 2,

	.refresh_max = 30,
	.refresh_min = 20,

	.hwswp = 1,
	.invvline = 1,
};

static inline unsigned readlcd(unsigned offset)
{
	return readl(lcd_base + offset);
}

static inline void writelcd(unsigned offset, unsigned value)
{
	writel(value, lcd_base + offset);
}

static void tomtomgolcd_dump_regs(void)
{
	PK_DBG("LCDCON1:[0x%08x] LCDCON2:[0x%08x] LCDCON3:[0x%08x] LCDCON4:[0x%08x] LCDCON5:[0x%08x]\n",
		readlcd(S3C2410_LCDCON1_OFFSET),
		readlcd(S3C2410_LCDCON2_OFFSET),
		readlcd(S3C2410_LCDCON3_OFFSET),
		readlcd(S3C2410_LCDCON4_OFFSET),
		readlcd(S3C2410_LCDCON5_OFFSET));

	PK_DBG("LCDSADDR1:[0x%08x] LCDSADDR3:[0x%08x] LCDSADDR3:[0x%08x]\n",
		readlcd(S3C2410_LCDSADDR1_OFFSET),
		readlcd(S3C2410_LCDSADDR2_OFFSET),
		readlcd(S3C2410_LCDSADDR3_OFFSET));
}

/*
 *	s3c2410fb_check_var():
 *	Get the video params out of 'var'. If a value doesn't fit, round it up,
 *	if it's too big, return -EINVAL.
 *
 *	Round up in the following order: bits_per_pixel, xres,
 *	yres, xres_virtual, yres_virtual, xoffset, yoffset, grayscale,
 *	bitfields, horizontal timing, vertical timing.
 */
static int tomtomgolcd_fb_check_var(struct fb_var_screeninfo *var, struct fb_info *info)
{
	PK_DBG("check_var(var=%p, info=%p)\n", var, info);

	if (var->xres>TTGL_XRES)			return -EINVAL;
	if (var->yres>TTGL_YRES)			return -EINVAL;

	if (var->xres_virtual>TTGL_XRESV)	return -EINVAL;
	if (var->yres_virtual>TTGL_YRESV)	return -EINVAL;

	if (var->xoffset!=0)				return -EINVAL;
	if (var->yoffset!=0)				return -EINVAL;

	if (var->bits_per_pixel!=16)		return -EINVAL;
	if (var->grayscale)					return -EINVAL;

	if (var->red.offset!=11)			return -EINVAL;
	if (var->green.offset!=5)			return -EINVAL;
	if (var->blue.offset!=0)			return -EINVAL;

	if (var->red.length!=5)				return -EINVAL;
	if (var->green.length!=6)			return -EINVAL;
	if (var->blue.length!=5)			return -EINVAL;

	return 0;
}

/**
 *      tomtomgofb_set_par - Optional function. Alters the hardware state.
 *      @info: frame buffer structure that represents a single frame buffer
 *
 */
static int tomtomgolcd_fb_set_par(struct fb_info *info)
{
	PK_DBG("set_par(info=%p)\n", info);
	return 0;
}


static int tomtomgolcd_fb_setcolreg(unsigned regno, unsigned red, unsigned green, unsigned blue, unsigned transp, struct fb_info *info)
{
	struct tomtomgolcd_fb_slot *fbs = (struct tomtomgolcd_fb_slot *)info;

	unsigned val;
	unsigned sr,sg,sb;

	sr = (red   >> (16 - 5)) & 0x001f;
	sg = (green >> (16 - 6)) & 0x003f;
	sb = (blue  >> (16 - 5)) & 0x001f;

	val = (sb << 0) + (sg << 5) + (sr << 11);
#if 0
	val  = ((red >> 5)    & 0xf800);
	val |= ((green >> 11) & 0x07e0);
	val |= ((blue >> 16)  & 0x001f);
#endif

	PK_DBG("setcolreg(reg=%d, RGBA=%d,%d,%d,%d val=%016x info=%p)\n", regno, red, green, blue, transp, val, info);
	fbs->pseudo_pal[regno] = val;

	return 0;
}

static inline void tomtomgolcd_set_base(unsigned start, unsigned stop)
{
	PK_DBG("Setting viewport from 0x%08x to 0x%08x\n", start, stop);
	writelcd(S3C2410_LCDSADDR1_OFFSET, S3C2410_LCDBANK(start >> 22) | S3C2410_LCDBASEU(start >> 1));
	writelcd(S3C2410_LCDSADDR2_OFFSET, stop >> 1);
}

/**
 *	tomtomgofb_pan_display - NOT a required function. Pans the display.
 *	@var: frame buffer variable screen structure
 *	@info: frame buffer structure that represents a single frame buffer
 *
 *	Pan (or wrap, depending on the `vmode' field) the display using the
 *	`xoffset' and `yoffset' fields of the `var' structure.
 *	If the values don't fit, return -EINVAL.
 *
 *	Returns negative errno on error, or zero on success.
 */
static int tomtomgolcd_fb_pan_display(struct fb_var_screeninfo *var, struct fb_info *info)
{
	unsigned long flags;
	unsigned start;
	unsigned stop;

	PK_DBG("xoffset=%u, yoffset=%u\n", var->xoffset, var->yoffset);
	if (var->xoffset != 0)
		return -EINVAL;

	start = info->fix.smem_start + var->yoffset * info->fix.line_length;
	stop = start + TTGL_XRES * TTGL_YRES * TTGL_BYPP;
	local_irq_save(flags);
	/* Wait for retrace, according to vague Samsung sample */
	while ((readlcd(S3C2410_LCDCON1_OFFSET) >> 18) == 0)
		;
	tomtomgolcd_set_base(start, stop);
	local_irq_restore(flags);

	return 0;
}

/**
 *      xxxfb_blank - NOT a required function. Blanks the display.
 *	@blank_mode: the blank mode we want.
 *	@info: frame buffer structure that represents a single frame buffer
 *
 *	Blank the screen if blank_mode != 0, else unblank. Return 0 if
 *	blanking succeeded, != 0 if un-/blanking failed due to e.g. a
 *	video mode which doesn't support it. Implements VESA suspend
 *	and powerdown modes on hardware that supports disabling hsync/vsync:
 *	blank_mode == 2: suspend vsync
 *	blank_mode == 3: suspend hsync
 *	blank_mode == 4: powerdown
 *
 *	Returns negative errno on error, or zero on success.
 *
 */
static int tomtomgolcd_fb_blank(int blank_mode, struct fb_info *info)
{
	PK_DBG("blank(mode=%d, info=%p)\n", blank_mode, info);
	return 0;
}


static struct fb_ops tomtomgolcd_fb_ops = {
	.owner			= THIS_MODULE,
	.fb_check_var	= tomtomgolcd_fb_check_var,
	.fb_set_par		= tomtomgolcd_fb_set_par,
	.fb_blank		= tomtomgolcd_fb_blank,
	.fb_pan_display	= tomtomgolcd_fb_pan_display,
	.fb_setcolreg	= tomtomgolcd_fb_setcolreg,
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
	.fb_cursor		= soft_cursor,
};


/* Fake monspecs to fill in fbinfo structure */
/* Don't know if the values are important    */
static struct fb_monspecs tomtomgolcd_fb_monspecs __initdata = {
	.hfmin	= 30000,
	.hfmax	= 70000,
	.vfmin	= 50,
	.vfmax	= 65,
};


static struct clk *lcd_clock;

static int tomtomgolcd_setup_screen(const struct tomtomgolcd_screen_info *info)
{
	u32 cols, lines, pixels;
	u32 clkval, real_refresh;
	u32 flags;
	u32 modemask = 0;
	u32 tft = 0;

	if(!info) return -EINVAL;

	//Sanity checks for modes
	switch (info->type) {
	case TFT_LCD:
		modemask |= S3C2410_LCDCON1_TFT; tft = 1;
		break;
	case STN_4BDS:
		modemask |= S3C2410_LCDCON1_DSCAN4;
		break;
	case STN_4BSS:
		modemask |= S3C2410_LCDCON1_STN4;
		break;
	case STN_8BSS:
		modemask |= S3C2410_LCDCON1_STN8;
		break;
	default:
		return -EINVAL;
	}

	switch (info->bpp) {
	case STN_1BPP:
		if(tft) return -EINVAL;
		modemask |= S3C2410_LCDCON1_STN1BPP;
		break;
	case STN_2BPP:
		if(tft) return -EINVAL;
		modemask |= S3C2410_LCDCON1_STN2GREY;
		break;
	case STN_4BPP:
		if(tft) return -EINVAL;
		modemask |= S3C2410_LCDCON1_STN4GREY;
		break;
	case STN_8BPP:
		if(tft) return -EINVAL;
		modemask |= S3C2410_LCDCON1_STN8BPP;
		break;
	case STN_12BPP:
		if(tft) return -EINVAL;
		modemask |= S3C2410_LCDCON1_STN12BPP;
		break;
	case TFT_1BPP:
		if(!tft) return -EINVAL;
		modemask |= S3C2410_LCDCON1_TFT1BPP;
		break;
	case TFT_2BPP:
		if(!tft) return -EINVAL;
		modemask |= S3C2410_LCDCON1_TFT2BPP;
		break;
	case TFT_4BPP:
		if(!tft) return -EINVAL;
		modemask |= S3C2410_LCDCON1_TFT4BPP;
		break;
	case TFT_8BPP:
		if(!tft) return -EINVAL;
		modemask |= S3C2410_LCDCON1_TFT8BPP;
		break;
	case TFT_16BPP_5551:
	case TFT_16BPP_565:
		if(!tft) return -EINVAL;
		modemask |= S3C2410_LCDCON1_TFT16BPP;
		break;
	case TFT_24BPP_MSB:
	case TFT_24BPP_LSB:
		if(!tft) return -EINVAL;
		modemask |= S3C2410_LCDCON1_TFT24BPP;
		break;
	default:
		return -EINVAL;
	}

	if(!tft) {
		PK_DBG("This driver does not support STN displays yet. Sorry.\n");
		return -EINVAL;
	}

	//Sanity checks
	if((info->vsize) == 0) return -EINVAL;
	if((info->vfpd) == 0) return -EINVAL;
	if((info->vbpd) == 0) return -EINVAL;
	if((info->vspw) == 0) return -EINVAL;

	if((info->hsize) == 0) return -EINVAL;
	if((info->hfpd) == 0) return -EINVAL;
	if((info->hbpd) == 0) return -EINVAL;
	if((info->hspw) == 0) return -EINVAL;

	if((info->vsize) > (1 + S3C2410_LCDCON2_LINEVAL_MAX)) return -EINVAL;
	if((info->vfpd)  > (1 + S3C2410_LCDCON2_VFPD_MAX)) return -EINVAL;
	if((info->vbpd)  > (1 + S3C2410_LCDCON2_VBPD_MAX)) return -EINVAL;
	if((info->vspw)  > (1 + S3C2410_LCDCON2_VSPW_MAX)) return -EINVAL;

	if((info->hsize) > (1 + S3C2410_LCDCON3_HOZVAL_MAX)) return -EINVAL;
	if((info->hfpd)  > (1 + S3C2410_LCDCON3_HFPD_MAX)) return -EINVAL;
	if((info->hbpd)  > (1 + S3C2410_LCDCON3_HBPD_MAX)) return -EINVAL;
	if((info->hspw)  > (1 + S3C2410_LCDCON4_HSPW_MAX)) return -EINVAL;

	//Calculate number of pixels sent to screen for each frame
	lines = (info->vsize + info->vfpd + info->vbpd + info->vspw);
	cols  = (info->hsize + info->hfpd + info->hbpd + info->hspw);
	pixels= lines * cols;

	//Calculate coarse clkval for maximum refresh rate
	clkval = clk_get_rate(lcd_clock);
	clkval/= (2 * pixels * info->refresh_max);

	//clkval 0 is not possible
	if(clkval == 0) clkval=1;

	//check if clkval is too high
	if(clkval > (1 + S3C2410_LCDCON1_CLKVAL_MAX)) clkval = (1 + S3C2410_LCDCON1_CLKVAL_MAX);

	//calculate real refesh rate
	real_refresh = clk_get_rate(lcd_clock);
	real_refresh/= (2 * pixels * clkval);

	//check for rounding errors
	if(real_refresh > info->refresh_max) {
		clkval++;
		real_refresh = clk_get_rate(lcd_clock);
		real_refresh/= (2 * pixels * clkval);
	}

	//check if real refresh rate is ok
	if(real_refresh < info->refresh_min) return -EBUSY;

	//flags
	flags = 0;
	if(info->bpp == TFT_24BPP_MSB) flags |= S3C2410_LCDCON5_BPP24BL;
	if(info->bpp == TFT_16BPP_565) flags |= S3C2410_LCDCON5_FRM565;

	if(info->invvclk)	flags |= S3C2410_LCDCON5_INVVCLK;
	if(info->invvline)	flags |= S3C2410_LCDCON5_INVVLINE;
	if(info->invvframe)	flags |= S3C2410_LCDCON5_INVVFRAME;
	if(info->invvd)		flags |= S3C2410_LCDCON5_INVVD;
	if(info->invvden)	flags |= S3C2410_LCDCON5_INVVDEN;
	if(info->invpwren)	flags |= S3C2410_LCDCON5_INVPWREN;
	if(info->invlend)	flags |= S3C2410_LCDCON5_INVLEND;
	if(info->enpwren)	flags |= S3C2410_LCDCON5_PWREN;
	if(info->enlend)	flags |= S3C2410_LCDCON5_ENLEND;
	if(info->bswp)		flags |= S3C2410_LCDCON5_BSWP;
	if(info->hwswp)		flags |= S3C2410_LCDCON5_HWSWP;

#if 0
	//HACK: real setup should probably be done elsewhere
	writelcd(S3C2410_LCDCON1_OFFSET,
		modemask |
		S3C2410_LCDCON1_CLKVAL(clkval-1) );

	writelcd(S3C2410_LCDCON2_OFFSET,
		S3C2410_LCDCON2_LINEVAL(info->vsize-1) |
		S3C2410_LCDCON2_VFPD(info->vfpd-1) |
		S3C2410_LCDCON2_VBPD(info->vbpd-1) |
		S3C2410_LCDCON2_VSPW(info->vspw-1) );

	writelcd(S3C2410_LCDCON3_OFFSET,
		S3C2410_LCDCON3_HOZVAL(info->hsize-1) |
		S3C2410_LCDCON3_HFPD(info->hfpd-1) |
		S3C2410_LCDCON3_HBPD(info->hbpd-1) );

	writelcd(S3C2410_LCDCON4_OFFSET,
		S3C2410_LCDCON4_HSPW(info->hspw-1) );

	writelcd(S3C2410_LCDCON5_OFFSET, flags);

	writelcd(S3C2410_LCDCON1_OFFSET,
		modemask | S3C2410_LCDCON1_ENVID |
		S3C2410_LCDCON1_CLKVAL(clkval-1) );
#else
	#define MVAL    (5)
	#define MVAL_USED   (0)
	#define VBPD_240320_GO      ((5-1)&0xff)
	#define VFPD_240320_GO      ((5-1)&0xff)
	#define VSPW_240320_GO      ((2-1)&0x3f)
	#define HBPD_240320_GO      ((16-1)&0x7f)
	#define HFPD_240320_GO      ((8*4-1)&0xff)
	#define HSPW_240320_GO      ((24-1)&0xff)
	#define CLKVAL_TFT_240320  (6)

	// Setup controller for 240x320 display in 16 bits
	writelcd(S3C2410_LPCSEL - S3C2410_VA_LCD, 2);
	writelcd(S3C2410_LCDCON1_OFFSET, (CLKVAL_TFT_240320 << 8) | (MVAL_USED << 7) | (3 << 5) | (12 << 1) );
	writelcd(S3C2410_LCDCON2_OFFSET, (VBPD_240320_GO << 24) | (319 << 14) | (VFPD_240320_GO << 6) | (VSPW_240320_GO) );
	writelcd(S3C2410_LCDCON3_OFFSET, (HBPD_240320_GO << 19) | (239 << 8) | (HFPD_240320_GO) );
	writelcd(S3C2410_LCDCON4_OFFSET, (MVAL << 8) | (HSPW_240320_GO) );
	writelcd(S3C2410_LCDCON5_OFFSET, (1 << 11) | (0 << 9) | (1 << 8) | (1 << 3) | (1 << 0) | (0 << 7) | (0 << 10) | (1 << 2) | (1 << 4) );
	// This should be done as last to enable video
	writelcd(S3C2410_LCDCON1_OFFSET, readlcd(S3C2410_LCDCON1_OFFSET) | 1);
#endif

	PK_DBG("setup_screen: Now running a refresh rate of %uHz.\n",real_refresh);

	return 0;
}

static void tomtomgolcd_disable_screen(void)
{
	writelcd(S3C2410_LCDCON1_OFFSET, 0);
}


static void tomtomgolcd_fb_show(void)
{
	unsigned start = fbs.fbi.fix.smem_start + fbs.fbi.var.yoffset * fbs.fbi.fix.line_length;
	unsigned stop = start + TTGL_XRES * TTGL_YRES * TTGL_BYPP;
	tomtomgolcd_set_base(start, stop);
	writelcd(S3C2410_LCDSADDR3_OFFSET, S3C2410_OFFSIZE(0) | S3C2410_PAGEWIDTH(240));
}

static void tomtomgolcd_fb_setup(struct tomtomgolcd_fb_slot *fbs)
{
	struct fb_info *fbi= &fbs->fbi;

	strcpy(fbi->fix.id, "s3c2410-lcd");

	fbi->fix.type           = FB_TYPE_PACKED_PIXELS;
	fbi->fix.type_aux       = 0;
	fbi->fix.visual         = FB_VISUAL_TRUECOLOR;
	fbi->fix.xpanstep       = 0;
	fbi->fix.ypanstep       = 1;
	fbi->fix.ywrapstep      = 0;
	fbi->fix.line_length    = TTGL_XRESV * TTGL_BYPP;
	fbi->fix.accel          = FB_ACCEL_NONE;

	fbi->var.nonstd         = 0;
	fbi->var.activate       = FB_ACTIVATE_NOW;
	fbi->var.height         = TTGL_HEIGHT;
	fbi->var.width          = TTGL_WIDTH;
	fbi->var.accel_flags    = 0;
	fbi->var.vmode          = FB_VMODE_NONINTERLACED;

	fbi->fbops              = &tomtomgolcd_fb_ops;
	fbi->flags              = FBINFO_FLAG_DEFAULT;
	fbi->monspecs           = tomtomgolcd_fb_monspecs;
	fbi->pseudo_palette     = fbs->pseudo_pal;

	fbi->var.xres           = TTGL_XRES;
	fbi->var.xres_virtual   = TTGL_XRESV;
	fbi->var.yres           = TTGL_YRES;
	fbi->var.yres_virtual   = TTGL_YRESV;
	fbi->var.bits_per_pixel = TTGL_BPP;

	fbi->var.red.offset     = 11;
	fbi->var.green.offset   = 5;
	fbi->var.blue.offset    = 0;
	fbi->var.transp.offset  = 0;
	fbi->var.red.length     = 5;
	fbi->var.green.length   = 6;
	fbi->var.blue.length    = 5;
	fbi->var.transp.length  = 0;

	fbi->fix.smem_len       = TTGL_XRESV * TTGL_YRESV * TTGL_BYPP;
}

static int tomtomgolcd_fb_alloc(struct tomtomgolcd_fb_slot *fbs)
{

	fbs->map_size = PAGE_ALIGN(fbs->fbi.fix.smem_len + PAGE_SIZE);
	fbs->map_cpu  = dma_alloc_writecombine(fbs->dev, fbs->map_size, &fbs->map_dma, GFP_KERNEL);

	if (!fbs->map_cpu)
		return -ENOMEM;

	fbs->fbi.fix.smem_start = fbs->map_dma + PAGE_SIZE;
	fbs->fbi.screen_base    = fbs->map_cpu + PAGE_SIZE;
	fbs->fbi.screen_size    = fbs->map_size;

	memset(fbs->fbi.screen_base, 0, fbs->fbi.fix.smem_len);

	return 0;
}

static void tomtomgolcd_fb_free(struct tomtomgolcd_fb_slot *fbs)
{
	if (fbs->map_cpu)
		dma_free_writecombine(fbs->dev, fbs->map_size, fbs->map_cpu, fbs->map_dma);
}

static void tomtomgo_lcdoff(void)
{
	// Turn backlight control off
	IO_Deactivate(BACKLIGHT_EN);

	// Show white picture to avoid screen noise after power down
	IO_Activate(VD2);
	IO_Activate(VD3);
	IO_Activate(VD4);
	IO_Activate(VD5);
	IO_Activate(VD6);
	IO_Activate(VD7);
	IO_Activate(VD10);
	IO_Activate(VD11);
	IO_Activate(VD12);
	IO_Activate(VD13);
	IO_Activate(VD14);
	IO_Activate(VD15);
	IO_Activate(VD18);
	IO_Activate(VD19);
	IO_Activate(VD20);
	IO_Activate(VD21);
	IO_Activate(VD22);
	IO_Activate(VD23);

	// Wait till white image appears
	mdelay(200);

	// Show white picture to avoid screen noise after power down

	// Stop driving LCD data
	IO_Deactivate(VD2);
	IO_Deactivate(VD3);
	IO_Deactivate(VD4);
	IO_Deactivate(VD5);
	IO_Deactivate(VD6);
	IO_Deactivate(VD7);
	IO_Deactivate(VD10);
	IO_Deactivate(VD11);
	IO_Deactivate(VD12);
	IO_Deactivate(VD13);
	IO_Deactivate(VD14);
	IO_Deactivate(VD15);
	IO_Deactivate(VD18);
	IO_Deactivate(VD19);
	IO_Deactivate(VD20);
	IO_Deactivate(VD21);
	IO_Deactivate(VD22);
	IO_Deactivate(VD23);

	// Disable logic signals
	IO_Deactivate(VDEN);
	IO_Deactivate(VCLK);

	// Disable LCD timing
	IO_Deactivate(LCD_OEN);

	// Wait 50 ms
	mdelay(50);

	// Turn charge pump off
	IO_Deactivate(LCD_BIAS_PWREN);

	// Wait 50 ms
	mdelay(50);

	// Disable LCD power
	IO_Deactivate(LCD_VCC_PWREN);
}

static void tomtomgo_lcdon(void)
{
	// Enable LCD power
	IO_Activate(LCD_VCC_PWREN);

	// Wait 50 ms
	mdelay(50);

	// Turn charge pump on
	IO_Activate(LCD_BIAS_PWREN);

	// Wait 50 ms
	mdelay(50);

	// Enable LCD timing
	IO_Activate(LCD_OEN);

	// Simulate black picture to avoid backlight flash
	IO_Deactivate(VD2);
	IO_Deactivate(VD3);
	IO_Deactivate(VD4);
	IO_Deactivate(VD5);
	IO_Deactivate(VD6);
	IO_Deactivate(VD7);
	IO_Deactivate(VD10);
	IO_Deactivate(VD11);
	IO_Deactivate(VD12);
	IO_Deactivate(VD13);
	IO_Deactivate(VD14);
	IO_Deactivate(VD15);
	IO_Deactivate(VD18);
	IO_Deactivate(VD19);
	IO_Deactivate(VD20);
	IO_Deactivate(VD21);
	IO_Deactivate(VD22);
	IO_Deactivate(VD23);

	// Enable logic signals
	IO_SetFunction(VDEN);
	IO_SetFunction(VCLK);

	// Wait till picture is black
	mdelay(50);

	// Enable normal video
	IO_SetFunction(VD2);
	IO_SetFunction(VD3);
	IO_SetFunction(VD4);
	IO_SetFunction(VD5);
	IO_SetFunction(VD6);
	IO_SetFunction(VD7);
	IO_SetFunction(VD10);
	IO_SetFunction(VD11);
	IO_SetFunction(VD12);
	IO_SetFunction(VD13);
	IO_SetFunction(VD14);
	IO_SetFunction(VD15);
	IO_SetFunction(VD18);
	IO_SetFunction(VD19);
	IO_SetFunction(VD20);
	IO_SetFunction(VD21);
	IO_SetFunction(VD22);
	IO_SetFunction(VD23);

	// Turn backlight control on
	IO_Activate(BACKLIGHT_EN);
}

static int tomtomgolcd_probe(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct resource *res;
	int size;

	PK_DBG("probe=%p, device=%p\n", pdev, dev);

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL) {
		PK_ERR("failed to get memory region resouce\n");
		return -ENOENT;
	}

	size = (res->end-res->start)+1;
	lcd_mem = request_mem_region(res->start, size, pdev->name);
	if (lcd_mem == NULL) {
		PK_ERR("failed to get memory region\n");
		return -ENOENT;
	}

	lcd_base = ioremap(res->start, size);
	if (lcd_base == NULL) {
		PK_ERR("failed to ioremap() region\n");
		return -EINVAL;
	}

	PK_DBG("probe: mapped lcd_base=%p\n", lcd_base);

	lcd_clock = clk_get(dev, "lcd");
	if (lcd_clock == NULL) {
		PK_ERR("failed to find lcd clock source\n");
		return -ENOENT;
	}

	clk_use(lcd_clock);

	clk_enable(lcd_clock);

	tomtomgolcd_fb_setup(&fbs);
	tomtomgolcd_fb_alloc(&fbs);
	register_framebuffer(&fbs.fbi);

	if (fb_prepare_logo(&fbs.fbi) == 0) {
		PK_ERR("no suitable logo found\n");
	} else {
		PK_DBG("showing TomTom logo\n");
		if (fb_show_logo(&fbs.fbi) == 0)
			PK_ERR("unable to display logo\n");
	}

	tomtomgolcd_fb_show();
	tomtomgolcd_setup_screen(&tomtomgo_screen_info);
//	tomtomgo_lcdon();
	tomtomgolcd_dump_regs();

	return 0;
}

static int tomtomgolcd_remove(struct device *dev)
{
	PK_DBG("device=%p\n", dev);

	tomtomgo_lcdoff();
	unregister_framebuffer(&fbs.fbi);
	tomtomgolcd_fb_free(&fbs);
	if (lcd_clock != NULL) {
		clk_disable(lcd_clock);
		clk_unuse(lcd_clock);
		clk_put(lcd_clock);
		lcd_clock = NULL;
	}
	if (lcd_base != NULL) {
		iounmap(lcd_base);
		lcd_base = NULL;
	}
	if (lcd_mem != NULL) {
		release_resource(lcd_mem);
		kfree(lcd_mem);
		lcd_mem = NULL;
	}

	return 0;
}

static void tomtomgolcd_shutdown(struct device *dev)
{
	tomtomgolcd_dump_regs();
	tomtomgo_lcdoff();
	tomtomgolcd_disable_screen();
	clk_disable(lcd_clock);
}

static void tomtomgolcd_powerup(struct device *dev)
{
	clk_enable(lcd_clock);
	tomtomgolcd_fb_show();
	tomtomgolcd_setup_screen(&tomtomgo_screen_info);
	tomtomgo_lcdon();
	tomtomgolcd_dump_regs();
}

#ifdef CONFIG_PM

static int tomtomgolcd_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("%s: dev = %p, state = %u, level = %u\n", __func__, dev, state, level);
	if (level == SUSPEND_POWER_DOWN) {
		tomtomgolcd_shutdown(dev);
	}
	return 0;
}

static int tomtomgolcd_resume(struct device *dev, u32 level)
{
	PK_DBG("%s: dev = %p, level = %u\n", __func__, dev, level);
	if (level == RESUME_POWER_ON) {
		tomtomgolcd_powerup(dev);
	}
	return 0;
}

#else /* CONFIG_PM */
#define tomtomgolcd_suspend NULL
#define tomtomgolcd_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver tomtomgolcd_driver = {
	.name		= "s3c2410-lcd",
	.bus		= &platform_bus_type,
	.probe		= tomtomgolcd_probe,
	.remove		= tomtomgolcd_remove,
	.shutdown	= tomtomgolcd_shutdown,
	.suspend	= tomtomgolcd_suspend,
	.resume		= tomtomgolcd_resume,
};

static int __init tomtomgolcd_init(void) {
	printk(KERN_INFO "TomTom GO LCD Driver, (C) 2005 TomTom BV\n");
	return driver_register(&tomtomgolcd_driver);
}

static void __exit tomtomgolcd_exit(void) {
	driver_unregister(&tomtomgolcd_driver);
}

module_init(tomtomgolcd_init);
module_exit(tomtomgolcd_exit);

MODULE_AUTHOR("Christian Daniel <cd@cdaniel.de>, Thomas Kleffel <tk@maintech.de>, Jeroen Taverne <jeroen.taverne@tomtom.com>, Dimitry Andric <dimitry.andric@tomtom.com> ((c) 2005 TomTom BV)");
MODULE_DESCRIPTION("TomTom GO LCD Driver");
MODULE_LICENSE("GPL");

/* EOF */
