/*
 *      NGFFS, next generation flash file system
 *      Copyright (C) 2004, TomTom BV
 *      Author: Koen Martens <kmartens@sonologic.nl>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or (at
 *      your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful, but
 *      WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *      General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software Foundation,
 *      Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *      $Id: chunk.c,v 1.9 2005/02/03 14:37:24 koen Exp $
 *
 *
 *	Chunks are the units in which access to the flash is organised,
 *	there exist four types of chunks: 
 *
 *	  - header chunks
 *		each erase-block of flash has as it's first chunk
 *	 	a header chunk which specifies activated and validity
 *		and more
 * 	  - directory chunks
 *		each (sub-)directory on the flash filesystem has a 
 *		directory chunk, which contains the number of items,
 *		the name and pointers to the file/dir chunks of each
 *		entry in this directory
 *	  - file chunks
 *	  	each file on the flash filesystem has a file chunk that
 *	  	defines name, size, permissions, etc...
 * 	  - data chunks
 * 	  	contains actual file data, all data chunks contain MAX_CSIZE
 * 	  	bytes, with the possible exception of the last chunk, which
 * 	  	may contain 512 or less bytes
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/blkdev.h>
#include <linux/mtd/mtd.h>
#include <asm/uaccess.h>
#include <barcelona/debug.h>
#include "ngffs.h"
#include "flash.h"
#include "garbage.h"

/* Defines */
#define PFX "ngffs: "
#define PK_DBG PK_DBG_FUNC

int ngffs_invalidate_chunk(struct super_block *sb,__u32 ofs);

/*
 * Function that calculates for any type of chunk the total size
 */
__u32 CHUNK_SIZE(void *chunk) {
  switch(((struct generic_chunk *)chunk)->type) {
      case NGFFS_HEADER_CHUNK:
	      	// header chunk has constant size
        return sizeof(struct header_chunk);
      case NGFFS_DIR_CHUNK:
		// size of dir chunk is constant + length of name +
		// length of array of entry_id's
        return sizeof(struct dir_chunk) +
                 ( ((struct generic_chunk *)chunk)->len*sizeof(__u32) ) +
                 ((struct generic_chunk *)chunk)->extra;  // namelen
      case NGFFS_FILE_CHUNK:
		// length of file chunk is constant + length of name
        return sizeof(struct file_chunk) +
                 ( ((struct generic_chunk *)chunk)->len/NGFFS_MAXCSIZE ) +
		   (NGFFS_NUMCHUNKS(((struct generic_chunk *)chunk)->len)*sizeof(__u32)) +
                 ((struct generic_chunk *)chunk)->extra;  // namelen
      case NGFFS_DATA_CHUNK:
		// length of data chunk is constant + data size
        return sizeof(struct data_chunk) +
                 ((struct generic_chunk *)chunk)->len;
      default:
		// this should never happen (Hi, Dimitry ;)
        PK_DBG("Invalid chunk type %d\n",((struct generic_chunk *)chunk)->type);
        return -1;
  }
  return -1;
}

/*
 * get the pointer to the cached chunk with the given entry_id, so this
 * is not the pointer in the flash mem, it's a pointer to the chunk in 
 * the dir chunk list or file chunk list
 */
struct generic_chunk *ngffs_get_chunk(struct super_block *sb,__u32 entry_id) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  struct list_head *curr;

  list_for_each(curr, &ngsb->files) {
    struct file_list_entry *file=list_entry(curr, struct file_list_entry, list);
    if(file->chunk->entry_id==entry_id) 
	    return (struct generic_chunk *)file->chunk;
  }
  list_for_each(curr, &ngsb->dirs) {
    struct dir_list_entry *dir=list_entry(curr, struct dir_list_entry, list);
    if(dir->chunk->entry_id==entry_id) 
	    return (struct generic_chunk *)dir->chunk;
  }
  return NULL;
}

/*
** tries to read the data chunk with given offset (relative to start of flash mem) and
** puts it in the data_chunk struct pointed at by the provided pointer
** 
** RETURNS
**   data	if success
**   NULL	if read failed
**
*/
struct data_chunk *ngffs_get_data_chunk(struct super_block *sb,__u32 ofs,struct data_chunk *data) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  int ec;

  PK_DBG("ngffs_get_data_chunk(0x%x,%ul,0x%x)\n",(unsigned int)sb,ofs,(unsigned int)data);

  	// check for out of bound reads 
  if(ofs>ngsb->size) return NULL;

  if((ec=ngffs_read(ngsb->mtd,ofs,(u_char *)data,sizeof(struct generic_chunk)))) {
    PK_DBG("ngffs_get_data_chunk(..,%i): ngffs_read returned error %i for generic chunk\n",ofs,ec);
    return NULL;
  }
  PK_DBG("got generic chunk datalen=%ul\n",data->len);
  if((ec=ngffs_read(ngsb->mtd,ofs,(u_char *)data,sizeof(struct data_chunk)+data->len))) {
    PK_DBG("ngffs_get_data_chunk(..,%i): ngffs_read returned error %i for actual chunk\n",ofs,ec);
    return NULL;
  }
  return data;
}

/*
** tries to write out the provided chunk into the specified block of flash mem,
**
** TODO: should probably return offset
**
** RETURNS
**  0	  on success
**  -EIO  on error
*/
int ngffs_put_chunk(struct super_block *sb,struct generic_chunk *chunk,int block) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  struct ngffs_block *blocks=ngsb->blocklist;
  __u32 size;

  size=CHUNK_SIZE(chunk);

  if(blocks[block].freespace<size) return -ENOSPC;

  if(ngffs_write(ngsb->mtd,
           (ngsb->erasesize*block)+(ngsb->erasesize-blocks[block].freespace),
           (u_char *)chunk,
           size)) {
    return -EIO;
  }
  blocks[block].freespace-=size;
  return 0;
}

/*
** reallocates memory for the chunk, the new area is of size CHUNK_SIZE+delta
**
** RETURNS
**  NULL	out of memory
**  pointer to new chunk in other cases
*/
struct generic_chunk *ngffs_resize_chunk(struct generic_chunk *chunk,int delta) {
  int size=CHUNK_SIZE(chunk);
  struct generic_chunk *new;
  PK_DBG("size %i, delta %i\n",size,delta);
  new=kmalloc(size+delta,GFP_KERNEL);
  PK_DBG("kmalloced %i bytes at 0x%p\n",size+delta,new);
  if(new==NULL) {
    PK_ERR("ngffs_resize_chunk(): out of mem\n");
    return NULL;
  }
  bcopy(chunk,new,(delta>0)?size:(size+delta));
  PK_DBG("copied %i bytes..\n",(delta>0)?size:(size+delta));
  return new;
}


/*
** Scans the flash memory, searching for the valid chunk with entry_id
**
**
** RETURNS
**  0 if the chunk could not be found, the offset else
**
*/
__u32 ngffs_get_chunk_ofs(struct super_block *sb,unsigned long entry_id) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  int i;
  __u32 ofs;
  struct generic_chunk chunk;
  __u32 max_modification=0;
  __u32 max_ofs=0;

  for(i=0;i<ngsb->blocks;i++) {			// loop over all blocks
    ofs= (i*ngsb->erasesize) + EMPTY_BYTES;
    while(ofs<((i+1)*ngsb->erasesize)) {	// scan block

      if(ngffs_read(ngsb->mtd,ofs,(u_char *)&chunk,sizeof(struct generic_chunk))) {
	return 0;				// read error
      }

      if(chunk.valid==NGFFS_CHUNK_VALID && (chunk.type==NGFFS_DIR_CHUNK || chunk.type==NGFFS_FILE_CHUNK) && chunk.entry_id==entry_id) {
        if(chunk.modification>max_modification) {
	  max_modification=chunk.modification;
	  max_ofs=ofs;
        }
      }
      	// reached end of block yet (type==0xff)?
      if(chunk.valid==NGFFS_CHUNK_VALID && chunk.type==NGFFS_UNUSED) break;
      else if(CHUNK_SIZE(&chunk)!=-1) ofs+=CHUNK_SIZE(&chunk);	// no=>next chunk
      else return 0;	// return error
    }
  }
  return max_ofs;
}

/*
** Enter data chunk info in the data chunk cache
** This is not actually caching the data, merely the offset relative to the start of flash
** Each entry in the list of file_chunks has it's own personal cache (limited by NGGFS_MAXFSIZE)
** discards data which is already cached with a higher modification counter
**
** RETURNS
**  1 for success
**  0 for error
**
** (does no checking for overflow or whatever)
*/
int ngffs_cache_data_chunk(struct ngffs_info *ngsb,__u32 entry_id,__u32 seq_id,__u32 modification,__u32 ofs) {
  struct list_head *curr;
  struct file_list_entry *file;

  PK_DBG("ngffs_cache_data_chunk(..,%ul,%ul,%ul,%ul)\n",entry_id,seq_id,modification,ofs);
  list_for_each(curr, &ngsb->files) {				// loop over all files
    file=list_entry(curr, struct file_list_entry, list);
    if(file->chunk->entry_id==entry_id) {			// is it our file?
      PK_DBG("cache-mod=%ud, cache-ofs=%ud\n",file->cache[seq_id].modification,file->cache[seq_id].offset);
      if(file->cache[seq_id].modification) {	// don't invalidate if there's nothing cached yet
	if(file->cache[seq_id].modification < modification) { // info is newer, invalidate old chunk (cleanup)
	  PK_DBG("invalidating old (cached) chunk\n");
	  ngffs_invalidate_chunk(ngsb->sb,file->cache[seq_id].offset);
	} else if(file->cache[seq_id].modification > modification) { // cached is newer, invalidate old chunk (cleanup)
	  PK_DBG("invalidating old (offered) chunk\n");
	  ngffs_invalidate_chunk(ngsb->sb,ofs);
	}
      }

      if(file->cache[seq_id].modification <= modification) {	// is the info newer?? (added for gc: the same)
	  file->cache[seq_id].offset=ofs;			// yes, store info
	  file->cache[seq_id].modification=modification;
      }
      return 1;
    }
  }
  return 0;
}

/*
**
** get the cache entry for a given entry/sequence id pair. the returned pointer is to a struct
** containing the modification no. and the offset relative to flash start
**
** RETURNS
**  NULL	on error
**  ptr		else
*/
struct ngffs_data_cache *ngffs_get_cached_data_chunk(struct super_block *sb,__u32 entry_id,__u32 seq_id) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  struct list_head *curr;
  struct file_list_entry *file;
  
  list_for_each(curr, &ngsb->files) {
    file=list_entry(curr, struct file_list_entry, list);
    if(file->chunk->entry_id==entry_id) {
      return &(file->cache[seq_id]);
    }
  }
  return NULL;
}

/*
** gets the entire cache, which is an array of ngffs_data_cache structs
**
** RETURNS
**  NULL	on failure
**  array(=ptr)	else
*/
struct ngffs_data_cache *ngffs_get_file_data_cache(struct super_block *sb,__u32 entry_id) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  struct list_head *curr;
  struct file_list_entry *file;
  
  list_for_each(curr, &ngsb->files) {
    file=list_entry(curr, struct file_list_entry, list);
    if(file->chunk->entry_id==entry_id) {
      return file->cache;
    }
  }
  return NULL;
}


/*
**  Returns the pointer to the file list entry or NULL if not found
*/
struct file_list_entry *ngffs_get_filele(struct super_block *sb,unsigned long entry_id) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  struct list_head *curr;

  list_for_each(curr, &ngsb->files) {
    struct file_list_entry *file;
    file=list_entry(curr, struct file_list_entry, list);
    if(file->chunk->entry_id==entry_id) {
	return file;
    }
  }
  return NULL;
}

/*
**  Returns the pointer to the file list entry or NULL if not found
*/
struct dir_list_entry *ngffs_get_dirle(struct super_block *sb,unsigned long entry_id) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  struct list_head *curr;

  list_for_each(curr, &ngsb->dirs) {
    struct dir_list_entry *dir;
    dir=list_entry(curr, struct dir_list_entry, list);
    if(dir->chunk->entry_id==entry_id) {
	return dir;
    }
  }
  return NULL;
}


/*
** hmmmmm
*/
int ngffs_update_chunk_ofs(struct super_block *sb,unsigned long entry_id,__u32 ofs) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  struct list_head *curr;

  PK_DBG("ngffs_update_chunk_ofs(..,%ld,%ul)\n",entry_id,ofs);

  list_for_each(curr, &ngsb->files) {
    struct file_list_entry *file;
    file=list_entry(curr, struct file_list_entry, list);
    if(file->chunk->entry_id==entry_id) {
	file->ofs=ofs;
	return 1;
    }
  }
  list_for_each(curr, &ngsb->dirs) {
    struct dir_list_entry *dir;
    dir=list_entry(curr, struct dir_list_entry, list);
    if(dir->chunk->entry_id==entry_id) {
	dir->ofs=ofs;
	return 1;
    }
  }
  return 0;
}


/*
** Invalidates the chunk at ofs by writing 0 in the valid field. Does no
** checking what-so-ever, so better only call this with a valid ofs
**
** RETURNS
**  0	 on success
**  -EIO on error
*/
int ngffs_invalidate_chunk(struct super_block *sb,__u32 ofs) {
  struct ngffs_info *ngsb;
  char invalid=NGFFS_CHUNK_INVALID;	// 0x00

  PK_DBG("sb=0x%p, ofs=%ud\n",sb,ofs);

  ngsb=NGFFS_INFO(sb);

  PK_DBG("ngsb=0x%p\n",ngsb);
  
  if(ngffs_write(ngsb->mtd , ofs + (int)(&(((struct generic_chunk *)0)->valid)) , &invalid , 1)) {
    PK_WARN("unable to invalidate chunk\n");
    return -EIO;
  }
  return 0;
}

/*
** Activates the chunk at ofs by writing 0 in the activated field. Does no
** checking what-so-ever, so better only call this with a valid ofs. Makes
** sense only for header chunks really..
**
** RETURNS
**  0	 on success
**  -EIO on error
*/
int ngffs_activate_chunk(struct super_block *sb,__u32 ofs) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  char activated=0;	// 0x00

  if(ngffs_write(ngsb->mtd , ofs + (int)(&(((struct header_chunk *)0)->activated)) , &activated , 1)) {
    PK_WARN("unable to activate chunk\n");
    return -EIO;
  }
  return 0;
}


/*
** Finds the parent of the given entry
**
** TODO: this actually does not work anymore ! Or does it now..?
*/
__u32 ngffs_parent_chunk_ofs(struct super_block *sb,__u32 entry_id) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  struct list_head *curr;
  struct dir_list_entry *dir;
  __u32 *entries;
  int i;

  list_for_each(curr, &ngsb->dirs) {
    dir=list_entry(curr, struct dir_list_entry, list);

    entries=(__u32*)(((char *)dir->chunk)+sizeof(struct dir_list_entry)+dir->chunk->namelen);
    for(i=0;i<dir->chunk->len;i++) {
      if(entries[i]==entry_id) return dir->chunk->entry_id;
    }
  }
  return -1;
}

/*
**
**
*/
int ngffs_replace_chunk(struct super_block *sb,struct generic_chunk *chunk) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  struct list_head *curr;

  PK_DBG("ngffs_replace_chunk(..,%i)\n",chunk->entry_id);

  list_for_each(curr, &ngsb->files) {
    struct file_list_entry *file;
    file=list_entry(curr, struct file_list_entry, list);
    if(file->chunk->entry_id==chunk->entry_id) {
      file->chunk=(struct file_chunk *)chunk;
      return 1;
    }
  }
  list_for_each(curr, &ngsb->dirs) {
    struct dir_list_entry *dir;
    dir=list_entry(curr, struct dir_list_entry, list);
    if(dir->chunk->entry_id==chunk->entry_id) {
      dir->chunk=(struct dir_chunk *)chunk;
      return 1;
    }
  }
  return 0;
}

  // returns block no of first block with at least size bytes free
int ngffs_contig_available(struct super_block * sb, __u32 size) {
  struct ngffs_info *ngsb;
  struct ngffs_block *blocks;
  int i,retry;

  ngsb=NGFFS_INFO(sb);

  for(retry=0;retry<ngsb->blocks;retry++) {
    blocks=ngsb->blocklist;

    for(i=0;i<ngsb->blocks;i++) {
      if( (blocks[i].header.valid) && (!blocks[i].header.activated) &&
	      (size<=blocks[i].freespace) ) {
        return i;
      }
    }

    if(ngffs_collect_garbage(sb)) {
      return -1;
    }
  }
  return -1; 
}


void ngffs_delete_file_chunk(struct super_block *sb,__u32 entry_id) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  struct list_head *curr;

  list_for_each(curr,&ngsb->files) {
    struct file_list_entry *file=list_entry(curr, struct file_list_entry, list);
    if(file->chunk->entry_id==entry_id) {
      kfree(file->chunk);
      kfree(file->cache);
      list_del(curr);
      return;
    }
  }
}

void ngffs_delete_dir_chunk(struct super_block *sb,__u32 entry_id) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  struct list_head *curr;

  list_for_each(curr,&ngsb->files) {
    struct dir_list_entry *dir=list_entry(curr, struct dir_list_entry, list);
    if(dir->chunk->entry_id==entry_id) {
      kfree(dir->chunk);
      list_del(curr);
      return;
    }
  }
}



  // clean up the mess we made 
void ngffs_cleanup(struct super_block *sb) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  struct list_head *curr;
  struct list_head *temp;

  PK_DBG("freeing blocklist\n");
  kfree(ngsb->blocklist);
 
  PK_DBG("freeing files\n");
  list_for_each_safe(curr, temp, &ngsb->files) {
    struct file_list_entry *file=list_entry(curr, struct file_list_entry, list);
    kfree(file->chunk);
    kfree(file->cache);
    list_del(curr);
  }
  PK_DBG("freeing dirs\n");
  list_for_each_safe(curr, temp, &ngsb->dirs) {
    struct dir_list_entry *dir=list_entry(curr, struct dir_list_entry, list);
    kfree(dir->chunk);
    list_del(curr);
  }
  PK_DBG("cleanup done\n");
}
