/*
 *      NGFFS, next generation flash file system
 *      Copyright (C) 2004, TomTom BV
 *	Author: Koen Martens <kmartens@sonologic.nl>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or (at
 *      your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful, but
 *      WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *      General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software Foundation,
 *      Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *      $Id: garbage.c,v 1.3 2005/01/27 15:18:46 koen Exp $
 *
 *
 * 	Garbage collection
 * 	 
 * 	Garbage collection collects (contrary to it's name) all valid
 * 	chunks from a given block and writes them out to the next
 * 	available block.
 * 	To provide wear leveling, ...
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/blkdev.h>
#include <linux/mtd/mtd.h>
#include <asm/uaccess.h>
#include <barcelona/debug.h>
#include "ngffs.h"
#include "flash.h"
#include "chunk.h"

/* Defines */
#define PFX "ngffs: "
#define PK_DBG PK_DBG_FUNC

/*
**
** This does garbage collection for one block:
**  - loop over the chunks in block 'from'
**  - copy valid chunks to the new block 'to' (increase modification!)
**
** Does not check whether there is enough free space in the to block, assumes
** this is already done by caller.
**
*/
__u32 ngffs_collect_garbage_block (struct ngffs_info *ngsb, int from, int to) {
  __u32 ofs=EMPTY_BYTES+sizeof(struct header_chunk);	// skip empty buffer + header
  __u32 newofs;
  struct generic_chunk chunk;
  u_char buffer[sizeof(struct data_chunk)+NGFFS_MAXCSIZE];	// see (i) in ngffs.h
  
 
  PK_DBG("collecting garbage for block %i (new block is %i)\n",from,to); 
  while(ofs<ngsb->blocklist[from].size) {
    if(ngffs_read(ngsb->mtd, ofs+ngsb->blocklist[from].offset, (u_char*)&chunk, sizeof(struct generic_chunk))) {
      PK_WARN("ngffs_block_garbage(sb,%i): unable to read at %ul \n",from,ofs+ngsb->blocklist[from].offset);
      return 0; // return 0 so garbage collection will leave this block alone
    }
    PK_DBG("ofs=%7ul (size=%7ul), type=%i, chunk size=%i\n",ofs,ngsb->blocklist[from].size,chunk.type,CHUNK_SIZE(&chunk));

    if(chunk.type==NGFFS_UNUSED) {
      return 0;
    }
    if(chunk.valid==NGFFS_CHUNK_VALID) {
      // chunk is valid -> copy
      //
      if(ngffs_read(ngsb->mtd, ofs+ngsb->blocklist[from].offset, buffer, CHUNK_SIZE(&chunk))) {
	PK_WARN("Unable to read during garbage collection at ofs %ul\n",ofs+ngsb->blocklist[from].offset);
	return -EIO;
      }
      ((struct generic_chunk *)buffer)->modification++;
      newofs=ngsb->blocklist[to].offset + ( ngsb->blocklist[to].size-ngsb->blocklist[to].freespace );
      if(ngffs_write(ngsb->mtd, newofs, buffer, CHUNK_SIZE(&chunk))) {
	PK_WARN("Unable to write during garbage collection at ofs %ul\n",newofs);
	return -EIO;
      }
      ngsb->blocklist[to].freespace-=CHUNK_SIZE(&chunk);
	PK_DBG("copied to %ul, %ul\n",newofs,ngsb->blocklist[to].freespace);
    }
    ofs+=CHUNK_SIZE(&chunk);
  }
  PK_WARN("WARNING chunk exceeds erase block boundary!!!\n");
  return -EIO;
}

/*
**
** Counts the collectable garbage in a given block
**
*/
__u32 ngffs_block_garbage(struct ngffs_info *ngsb,int blockidx) {
  __u32 ofs=EMPTY_BYTES+sizeof(struct header_chunk);	// skip empty buffer + header
  struct generic_chunk chunk;
  __u32 garbage=0;
 
 
 
  PK_DBG("counting garbage for block %i\n",blockidx); 
  while(ofs<ngsb->blocklist[blockidx].size) {
    if(ngffs_read(ngsb->mtd, ofs+ngsb->blocklist[blockidx].offset, (u_char*)&chunk, sizeof(struct generic_chunk))) {
      PK_WARN("ngffs_block_garbage(sb,%i): unable to read\n",blockidx);
      return 0; // return 0 so garbage collection will leave this block alone
    }
    PK_DBG("ofs=%7ul (size=%7ul), type=%i, chunk size=%i\n",ofs,ngsb->blocklist[blockidx].size,chunk.type,CHUNK_SIZE(&chunk));
{ int dd=1000000; while(--dd) { dd*=2; dd=dd/(4-2); } }
    if(chunk.type==NGFFS_UNUSED) {
      PK_DBG("garbage count: %ul\n",garbage);
      return garbage;
    }
    if(chunk.valid!=NGFFS_CHUNK_VALID) {
      garbage+=CHUNK_SIZE(&chunk);
    }
    ofs+=CHUNK_SIZE(&chunk);
  }
  PK_WARN("WARNING chunk exceeds erase block boundary!!!\n");
  PK_DBG("garbage count: %ul\n",garbage);
  return garbage;
}

// get amount of freeable space
__u32 ngffs_garbage_collectable(struct super_block *sb) {
   return 0;
}

// do actual garbage collection
int ngffs_collect_garbage(struct super_block *sb) {
  struct ngffs_info *ngsb=NGFFS_INFO(sb);
  int i;
  int last_inactive=-1;
  int curr=0;
  int ec;
  u_char empty[EMPTY_BYTES];	// all zeroes, needed to initialize a block

  // lock, hmm aren't we already locked??

  	// find out how much garbage is present in each block
	// and find last inactive block (should be only one 
	// inactive block really!)
  for(i=0;i<ngsb->blocks;i++) {
    if(ngsb->blocklist[i].header.valid && !(ngsb->blocklist[i].header.activated)) {
      ngsb->blocklist[i].garbage=NGFFS_GARBAGE_COLLECT;
      ngsb->blocklist[i].reserved=ngffs_block_garbage(ngsb,i);
    } else {
      ngsb->blocklist[i].garbage=NGFFS_GARBAGE_IGNORE;
      if(last_inactive!=-1) PK_DBG("ngffs_collect_garbage(): more than one inactive block!!!");
      last_inactive=i;
      ngsb->blocklist[i].reserved=0; 
    }
  }

  // now inactive contains the number of the last inactive block
  // and the ngsb->blocks[i].reserved contain the amount of collectable garbage
  PK_DBG("last_inactive=%i\n",last_inactive);

  if(last_inactive==-1) {
    PK_ERR("ngffs_collect_garbage(): no inactive blocks, garbage collection failed\n");
    return -ENOSPC;
  }

  	// find first garbage-collectable block
  i=0;	// use i to count iterations, just to be sure
  curr=(last_inactive+1)%ngsb->blocks;
  while( (i<ngsb->blocks) && 
	!( (ngsb->blocklist[curr].header.valid) && !(ngsb->blocklist[curr].header.activated) ) ) {

    curr++;
    curr=curr%ngsb->blocks;
    i++;
  }
  if(i==ngsb->blocks) { // no suitable block found
	  PK_ERR("ngffs_garbage_collect(): no valid+active block found!!\n");
	  return -ENOSPC;
  }

  // erase last_inactive & mark as 'to be activated'
  ngffs_erase_flash(ngsb->mtd,last_inactive,2);
  while(!ngffs_erase_ready(2)) {
    // schedule();
  }
  ngsb->blocklist[last_inactive].garbage=NGFFS_GARBAGE_ACTIVATE;
  ngsb->blocklist[last_inactive].freespace=ngsb->blocklist[curr].size - EMPTY_BYTES - sizeof(struct header_chunk);
  
  i=0;	// use i to count iterations again
  curr=(last_inactive+1)%ngsb->blocks;
  while(i<(ngsb->blocks-1)) {
    __u32 notgarbage=ngsb->blocklist[curr].size - ngsb->blocklist[curr].reserved - EMPTY_BYTES - sizeof(struct header_chunk);
    if( (ngsb->blocklist[curr].garbage==NGFFS_GARBAGE_COLLECT) && (notgarbage<=ngsb->blocklist[last_inactive].freespace)) {
      if((ec=ngffs_collect_garbage_block(ngsb,curr,last_inactive))) {
        PK_DBG("ngffs_collect_garbage(): copy block from %i to %i failed\n",curr,last_inactive);
	return ec;
      }
      	// mark block for invalidation 
      ngsb->blocklist[curr].garbage=NGFFS_GARBAGE_INVALIDATE;
    }
    i++; curr++;
  } 

  // now first activate new block (write new header), then invalidate blocks
  // if we lose power here: we're screwed --> no we're not.. we have the
  // 'modified' counter right, we always use last chunk.. just as long as
  // we activate first

/*
struct header_chunk {
  __u32         type;
  char          valid;          // 0xff=valid
  __u32         len;
  __u32         rootdir;        // entry-id of the root directory (0xffffffff=no root dir yet)
  char          activated;      // 0xff=not activated
  __u32         maxfsize;
  __u32         maxcsize;
  __u32         maxdirent;
};
*/

	// TODO: memset?
  for(i=0;i<EMPTY_BYTES;i++) empty[i]=0;

  for(i=0;i<ngsb->blocks;i++) {
    if(ngsb->blocklist[i].garbage==NGFFS_GARBAGE_ACTIVATE) {
      ngsb->blocklist[i].header.type=NGFFS_HEADER_CHUNK;
      ngsb->blocklist[i].header.valid=NGFFS_CHUNK_VALID;
      ngsb->blocklist[i].header.activated=0;
      ngsb->blocklist[i].header.maxfsize= NGFFS_MAXFSIZE;
      ngsb->blocklist[i].header.maxcsize= NGFFS_MAXCSIZE;
      ngsb->blocklist[i].header.maxdirent= NGFFS_MAXDIRENT;

      PK_DBG("activating block %i\n",i);

      if(ngffs_write(ngsb->mtd,ngsb->blocklist[i].offset,empty,EMPTY_BYTES)) {
		PK_ERR("PANIC! error writing empty bytes\n");
		return -EIO;
      }
      if(ngffs_write(ngsb->mtd,ngsb->blocklist[i].offset+EMPTY_BYTES,(u_char *)&(ngsb->blocklist[i].header),sizeof(struct header_chunk))) {
		PK_ERR("PANIC! error writing new header chunk\n");
		return -EIO;
      }
    }
  }
  for(i=0;i<ngsb->blocks;i++) {
    if(ngsb->blocklist[i].garbage==NGFFS_GARBAGE_INVALIDATE) {
      PK_DBG("invalidating block %i\n",i);

      ngsb->blocklist[i].header.valid=NGFFS_CHUNK_INVALID;

      if(ngffs_write(ngsb->mtd,ngsb->blocklist[i].offset+EMPTY_BYTES,(u_char *)&(ngsb->blocklist[i].header),sizeof(struct header_chunk))) {
	PK_ERR("Error invalidating block in garbage collection, PANIC!\n");
	return -EIO;
      }
    }
  }

  // refresh cached offsets
  ngffs_rescan_flash(sb);
  
  // unlock
  return 0;
}
