#include <linux/list.h>


// not defined elsewhere, is in lib/string.c
//extern void bcopy(const void * srcp, void * destp, size_t count);

#define bcopy(src,dst,cnt) memcpy(dst,src,cnt)


// static configuration
#define NGFFS_MAXCSIZE	512			// maximum chunk size
#define NGFFS_MAXFSIZE	(30*1024)		// maximum file size
#define NGFFS_MAXDIRENT 32			// maximum entries per directory
						// (i) we want to make sure that sizeof(struct dir_chunk)+(MAXDIRENT*sizeof(__u32))+maxnamelen
						//     does not exceed sizeof(struct data_chunk)+512 !!!
#define NGFFS_HAS_SYSFILE	1

#if ( ((PAGE_CACHE_SIZE / NGFFS_MAXCSIZE)*NGFFS_MAXCSIZE)!=PAGE_CACHE_SIZE )	
	// avoid this, as it breaks the reading/writing of files
#error "NGFFS_MAXCSIZE badly chosen"
#endif

#define NGFFS_MAX_NAME_LEN 256
#define EMPTY_BYTES	10

#define NGFFS_HEADER_CHUNK	1
#define NGFFS_DIR_CHUNK		2
#define NGFFS_FILE_CHUNK	3
#define NGFFS_DATA_CHUNK	4
#define NGFFS_UNUSED		(-1)

#define NGFFS_CHUNK_VALID	0xff
#define NGFFS_CHUNK_INVALID	0x00

#define NGFFS_NULL_ENTRY	0xffffffff	// if present in the file_chunk entry pointers, indicates
						// a block consisting of all zeroes of size NGFFS_MAXCSIZE


struct ngffs_metadata {
  __u32		link_count;
  umode_t	mode;
  uid_t		uid;
  gid_t		gid;
//  time_t	atime;	// last accessed not used, mtime returned instead
  time_t	mtime;
  time_t	ctime;
};


// TODO: fix field offsets
struct generic_chunk {
  __u32		type;		
  char		valid;		// 0xff=valid
  __u32		len;
  __u32		extra;		// length of dirname
  __u32		modification;	// if applicable
  __u32		entry_id;	// if applicable
};

struct header_chunk {
  __u32		type;		
  char		valid;		// 0xff=valid
  __u32		len;
  __u32		rootdir;	// entry-id of the root directory (0xffffffff=no root dir yet)
  char		activated;	// 0xff=not activated
  __u32		maxfsize;
  __u32		maxcsize;
  __u32		maxdirent;
};

struct file_chunk {
  __u32		type;		// ==NGFFS_FILE_CHUNK
  char		valid;		// 0xff=valid
  __u32		len;		// length of the file in bytes
  __u32		namelen;	// length of filename
  __u32		modification;	// number of modification (largest is last)
  __u32		entry_id;	// ino
  struct ngffs_metadata meta;
	// .. 
  //char name[namelen]		// variable length filename
  //__u32 chunklist[len]	// variable length data ** NOT USED ANYMORE, REPLACED BY SEQ_ID **
};

	// see (i) in regard to size
struct dir_chunk {
  __u32		type;		// ==NGFFS_DIR_CHUNK
  char		valid;		// 0xff=valid
  __u32		len;		// number of entries
  __u32		namelen;	// length of dirname
  __u32		modification;	// number of modification (largest is last)
  __u32		entry_id;	// ino
  struct ngffs_metadata meta;
  	// ...
  //char name[namelen]		// variable length filename
  //__u32 entries[len]		// variable length array of file/dir chunks
};

	// see (i) in regard to size
struct data_chunk {
  __u32		type;		// ==NGFFS_DATA_CHUNK
  char		valid;		// 0xff=valid
  __u32		len;		// length of this chunk in bytes
  __u32		seq_id;		// which part of file it defines
  __u32		modification;	// number of modification (largest is last)
  __u32		entry_id;	// which inode it is associated with
};

struct dir_list_entry {
  struct list_head list;
  struct dir_chunk *chunk;
  __u32 ofs;
};

struct ngffs_data_cache {
  __u32 offset;
  __u32 modification;
};

struct file_list_entry {
  struct list_head list;
  struct file_chunk *chunk;
  __u32 ofs;
  struct ngffs_data_cache *cache;
};

struct ngffs_block {
  struct header_chunk 	header;
  __u32			offset;
  __u32			size;
  __u32			freespace;
  __u32			reserved;	// used when locked, to determine if commit_write may
  					// succeed
  __u32			garbage;	// used for flags during garbage collection
};

struct ngffs_info {
  struct ngffs_block *blocklist;
  __u32 offset;			// offset from start of flash where flash fs starts
  __u32 blocks;			// number of erase blocks used for fs
  __u32 erasesize;		// total size in bytes of erase blocks
  __u32 size;			// total size in bytes of fs
  __u32 max_entry_id;		// highest ino in fs
  struct mtd_info *mtd;
  struct list_head dirs;		
  struct list_head files;
  rwlock_t global_lock;		// locks the entire filesystem
  struct super_block *sb;	// superblock
};

#define NGFFS_INFO(x)	((struct ngffs_info *)x->s_fs_info)

#define INODE_DIR(x)	(((struct dir_list_entry *)(x->u.generic_ip))->chunk)
#define INODE_FILE(x)	(((struct file_list_entry *)(x->u.generic_ip))->chunk)
#define INODE_FILE_OFS(x) (((struct file_list_entry *)(x->u.generic_ip))->ofs)
#define INODE_DIR_OFS(x) (((struct dir_list_entry *)(x->u.generic_ip))->ofs)

#define GENERIC_CHUNK(x)	((struct generic_chunk *)x)
#define DIR_CHUNK(x)	((struct dir_chunk *)x)
#define FILE_CHUNK(x)	((struct file_chunk *)x)
#define DATA_CHUNK(x)	((struct data_chunk *)x)
#define HEADER_CHUNK(x)	((struct header_chunk *)x)

#define NGFFS_SUPER_MAGIC	0x476d4373

#define NGFFS_NUMCHUNKS(x)	( (x/NGFFS_MAXCSIZE) + ((x%NGFFS_MAXCSIZE)?1:0) )


#define NGFFS_GARBAGE_NONE	0
#define NGFFS_GARBAGE_EMPTY	1
#define NGFFS_GARBAGE_ACTIVATE	2
#define NGFFS_GARBAGE_INVALIDATE 3
#define NGFFS_GARBAGE_IGNORE	4 
#define NGFFS_GARBAGE_COLLECT 5


inline static void ngffs_copy_from_meta(struct inode *inode,struct ngffs_metadata *meta) {
  inode->i_mode=meta->mode;
  inode->i_nlink=meta->link_count;
  inode->i_uid=meta->uid;
  inode->i_gid=meta->gid;
  inode->i_mtime.tv_sec=meta->mtime;
  inode->i_atime.tv_sec=meta->mtime;
  inode->i_ctime.tv_sec=meta->ctime;
  inode->i_mtime.tv_nsec=0;
  inode->i_atime.tv_nsec=0;
  inode->i_ctime.tv_nsec=0;
}

inline static void ngffs_copy_to_meta(struct inode *inode,struct ngffs_metadata *meta) {
  meta->mode=inode->i_mode;
  meta->link_count=inode->i_nlink;
  meta->uid=inode->i_uid;
  meta->gid=inode->i_gid;
  meta->mtime=inode->i_mtime.tv_sec;
  meta->ctime=inode->i_ctime.tv_sec;
}

